# -*- coding: utf-8 -*-

# GLiMR Python Toolbox
# Dan Bean      Static snapshot Feb 2015

# Please obtain current version

import arcpy
from arcpy import env
from arcpy.sa import *
import os
import sys
import time
import traceback
import math
import _winreg
import numpy as np         # will have to check availability if put in-line
from operator import itemgetter, attrgetter


class Toolbox(object):
    def __init__(self):
        self.label = u'Toolbox'
        self.alias = u'GLiMR Toolbox'
        self.tools = [GLiMR_Batch, GLiMR_Single, GLiMR_SummarizeStats, GLiMR_Landmarks,
                      GLiMR_MorphoJ]


class GLiMR_env(object):
    def __init__(self):
        '''Expect GLiMR\GLiMR.env.txt on C: or D: drive'''
        self.dir_list = None
        GLiMR_env_dir = self.get_reg('GLiMR_path')                              # get path from registry

        if GLiMR_env_dir:
            arcpy.AddMessage('Setting GLiMR environment path to ' + GLiMR_env_dir)
            GLiMR_env_file = os.path.join(GLiMR_env_dir, 'GLiMR.env.txt')
        else:
            if arcpy.Exists(r'c:\GLiMR\GLiMR.env.txt'):                         # fallback if registry access problems
                GLiMR_env_file = r'c:\GLiMR\GLiMR.env.txt'
            elif arcpy.Exists(r'd:\GLiMR\GLiMR.env.txt'):                       # fallback if registry access problems
                GLiMR_env_file = r'd:\GLiMR\GLiMR.env.txt'

        if arcpy.Exists(GLiMR_env_file):
            self.dir_list = self.get_env(GLiMR_env_file)
            arcpy.AddMessage('Found GLiMR environment named ' + GLiMR_env_file)
            return
        else:
            arcpy.AddError('\n\nCannot find GLiMR.env.txt environment file. Looking for ' + GLiMR_env_file +
                           'If you have moved the GLiMR.env.txt file, you can reset the GLiMR registry path'
                           '\nby placing the files Set_GLiMR_path.bat and GLiMR_set_registry.py in the directory'
                           '\ncontaining GLiMR.env.txt and then running Set_GLiMR_path.bat\n')
            arcpy.GetMessages()



    def get_env(self, env_file):
        '''
        Reads GLiMR environment file and returns directory list
        If GLiMR environment file does not exist, it creates one
        list order: xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir
        '''

        try:

            if not arcpy.Exists(env_file):

                xyz_dir = r'C:\GLiMR\GLiMR_XYZ'
                gdb_dir = r'C:\GLiMR\GLiMR_GDB'
                mxd_dir = r'C:\GLiMR\GLiMR_MXD'
                image_dir = r'C:\GLiMR\GLiMR_Images'
                template_dir = r'C:\GLiMR\GLiMR_Templates'
                gdb_edit_dir = r'C:\GLiMR\GLiMR_GDB_Edits'
                output_dir = r'C:\GLiMR\GLiMR_Output'

                outfile = open(env_file, 'w')

                outfile.write(xyz_dir + '\n')
                outfile.write(gdb_dir + '\n')
                outfile.write(mxd_dir + '\n')
                outfile.write(image_dir + '\n')
                outfile.write(template_dir + '\n')
                outfile.write(gdb_edit_dir + '\n')
                outfile.write(output_dir + '\n')
                outfile.write('\n\nThis is the GLiMR environment file.\nThe first five lines contain the following '
                              'directories\nOne directory path per line in the following order:'
                              '\n\nXYZ\nGDB\nMXD\nImages\nTemplates\nGDB_Edits\nOutput\n')
                outfile.write(
                    '\nThe GDB_Edits directory contains persistent data such as manually digitized scar feature '
                    'classes\nand Re_entrants that may have had edits to Left, Right, Base Type attributes.'
                    '\n\nThe standard location for this file is C:\GLiMR')

                outfile.close()

            else:
                infile = open(env_file, 'r')
                xyz_dir = infile.readline().strip('\n')
                gdb_dir = infile.readline().strip('\n')
                mxd_dir = infile.readline().strip('\n')
                image_dir = infile.readline().strip('\n')
                template_dir = infile.readline().strip('\n')
                gdb_edit_dir = infile.readline().strip('\n')
                output_dir = infile.readline().strip('\n')

            return ([xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir])

        except arcpy.ExecuteError:
            arcpy.AddMessage(arcpy.GetMessages())
        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

    def set_reg(self, name, value):
        '''Set registry value in CURRENT_USER\SOFTWARE\PSAL\GLiMR
           create key if necessary '''

        REG_PATH = r"SOFTWARE\PSAL\GLiMR"
        try:
            _winreg.CreateKey(_winreg.HKEY_CURRENT_USER, REG_PATH)
            registry_key = _winreg.OpenKey(_winreg.HKEY_CURRENT_USER, REG_PATH, 0,
                                           _winreg.KEY_WRITE)
            _winreg.SetValueEx(registry_key, name, 0, _winreg.REG_SZ, value)
            _winreg.CloseKey(registry_key)
            return True
        except WindowsError:
            return False

    def get_reg(self, name):
        '''Get registry value in CURRENT_USER\SOFTWARE\PSAL\GLiMR'''

        REG_PATH = r"SOFTWARE\PSAL\GLiMR"
        try:
            registry_key = _winreg.OpenKey(_winreg.HKEY_CURRENT_USER, REG_PATH, 0,
                                           _winreg.KEY_READ)
            value, regtype = _winreg.QueryValueEx(registry_key, name)
            _winreg.CloseKey(registry_key)
            return value
        except WindowsError:
            print 'Was unable to set registry value in HKEY_CURRENT_USER\PSAL\GLiMR'
            return None



# Tool implementation code

class GLiMR_Batch(object):
    """C:\GLiMR\Python\Toolbox.tbx\GLiMR_Batch"""

    def __init__(self):
        self.label = u'GLiMR Batch mode'
        self.description = u'''Run all scan files in selected Scan File (*.xyz) directory.
        Creates GDBs, feature classes, MXD files and images.   Summary Access database in Geodatabase directory'''

        self.canRunInBackground = False

        g_env = GLiMR_env()  # populate directory list from GLiMR.env.txt file for default locations
        self.dir_list = g_env.dir_list

        if self.dir_list:
            self.xyz_dir = self.dir_list[0]
            self.gdb_dir = self.dir_list[1]
            self.mxd_dir = self.dir_list[2]
            self.image_dir = self.dir_list[3]
            self.template_dir = self.dir_list[4]
            self.gdb_edit_dir = self.dir_list[5]
            self.output_dir = self.dir_list[6]
        else:
            self.xyz_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.gdb_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.mxd_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.image_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.template_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.gdb_edit_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.output_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'


    def getParameterInfo(self):

        # Source Scan File Directory
        param_1 = arcpy.Parameter()
        param_1.name = u'xyz_dir'
        param_1.displayName = u'Scan File Directory  (*.xyz)'
        param_1.parameterType = 'Required'
        param_1.direction = 'Input'
        param_1.datatype = u'DEFolder'
        param_1.value = self.xyz_dir

        # Geodatabase Directory
        param_2 = arcpy.Parameter()
        param_2.name = u'gdb_dir'
        param_2.displayName = u'Geodatabase Directory  (*.gdb)'
        param_2.parameterType = 'Required'
        param_2.direction = 'Input'
        param_2.datatype = u'DEFolder'
        param_2.value = self.gdb_dir

        # Persistent Geodatabase Edit Directory
        param_3 = arcpy.Parameter()
        param_3.name = u'gdb_edit_dir'
        param_3.displayName = u'Persistent Geodatabase Edit Directory'
        param_3.parameterType = 'Required'
        param_3.direction = 'Input'
        param_3.datatype = u'DEFolder'
        param_3.value = self.gdb_edit_dir

        # ArcMap MXD Directory
        param_4 = arcpy.Parameter()
        param_4.name = u'mxd_dir'
        param_4.displayName = u'ArcMap (*.mxd) File Directory'
        param_4.parameterType = 'Required'
        param_4.direction = 'Input'
        param_4.datatype = u'DEFolder'
        param_4.value = self.mxd_dir

        # Image Directory
        param_5 = arcpy.Parameter()
        param_5.name = u'image_dir'
        param_5.displayName = u'Image File Directory'
        param_5.parameterType = 'Required'
        param_5.direction = 'Input'
        param_5.datatype = u'DEFolder'
        param_5.value = self.image_dir

        # Template MXD Directory
        param_6 = arcpy.Parameter()
        param_6.name = u'template_dir'
        param_6.displayName = u'Template (*.mxd) File Directory'
        param_6.parameterType = 'Required'
        param_6.direction = 'Input'
        param_6.datatype = u'DEFolder'
        param_6.value = self.template_dir

        # GLiMR Output Directory
        param_7 = arcpy.Parameter()
        param_7.name = u'output_dir'
        param_7.displayName = u'GLiMR Output Directory'
        param_7.parameterType = 'Required'
        param_7.direction = 'Input'
        param_7.datatype = u'DEFolder'
        param_7.value = self.output_dir

        # GLiMR Verbose flag
        param_8 = arcpy.Parameter()
        param_8.name = u'verbose_flag'
        param_8.displayName = u'GLiMR Verbose flag'
        param_8.parameterType = 'Optional'
        param_8.direction = 'Input'
        param_8.datatype = u'GPBoolean'
        param_8.value = False

        # GLiMR Compute flake flag
        param_9 = arcpy.Parameter()
        param_9.name = u'compute_flake_flag'
        param_9.displayName = u'GLiMR Compute Flake flag'
        param_9.parameterType = 'Optional'
        param_9.direction = 'Input'
        param_9.datatype = u'GPBoolean'
        param_9.value = True

        # GLiMR Make images flag
        param_10 = arcpy.Parameter()
        param_10.name = u'make_images_flag'
        param_10.displayName = u'GLiMR Make images flag'
        param_10.parameterType = 'Optional'
        param_10.direction = 'Input'
        param_10.datatype = u'GPBoolean'
        param_10.value = True

        return [param_1, param_2, param_3, param_4, param_5, param_6, param_7, param_8, param_9, param_10]

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        if parameters[0].value:  # xyz_dir
            if parameters[0].valueAsText.find('GDB') > -1:
                parameters[0].setWarningMessage(
                    'You appear to be selecting a GDB directory instead of an XYZ directory.')

        if parameters[1].value:  # gdb_dir
            if parameters[1].valueAsText.find('XYZ') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an XYZ directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Edit') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an Edit directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Image') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an Image directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Output') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an Output directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Template') > -1:
                parameters[1].setWarningMessage('You appear to be selecting a Template directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('MXD') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an MXD directory instead of a GDB directory.')

            if parameters[1].value.value == parameters[2].value.value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')

        if parameters[2].value:  # gdb_edit_dir
            if parameters[1].value == parameters[2].value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')
            if parameters[2].valueAsText.find('GLiMR_GDB_Edits') == -1:
                parameters[2].setWarningMessage(
                    "Do not change casually!  GLiMR depends on the correlation between GDB and GDB_Edit geodatabases. Expects 'GLiMR_GDB_Edits'")
            if parameters[1].value.value == parameters[2].value.value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')

        if parameters[3].value:  # mxd_dir
            if parameters[3].valueAsText.find('XYZ') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an XYZ directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Edit') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an Edit directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Image') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an Image directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Output') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an Output directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('GDB') > -1:
                parameters[3].setWarningMessage('You appear to be selecting a GDB directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Template') > -1:
                parameters[3].setWarningMessage('You appear to be selecting a Template directory instead of an MXD directory.')

        if parameters[4].value:  # image_dir
            if parameters[4].valueAsText.find('XYZ') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an XYZ directory instead of an Image directory.')
            if parameters[4].valueAsText.find('Edit') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an Edit directory instead of an Image directory.')
            if parameters[4].valueAsText.find('MXD') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an MXD directory instead of an Image directory.')
            if parameters[4].valueAsText.find('Output') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an Output directory instead of an Image directory.')
            if parameters[4].valueAsText.find('Template') > -1:
                parameters[4].setWarningMessage('You appear to be selecting n Template directory instead of an Image directory.')
            if parameters[4].valueAsText.find('GDB') > -1:
                parameters[4].setWarningMessage('You appear to be selecting a GDB directory instead of an Image directory.')

        if parameters[5].value:  # template_dir
            if parameters[5].valueAsText.find('XYZ') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an XYZ directory instead of a Template directory.')
            if parameters[5].valueAsText.find('Edit') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an Edit directory instead of a Template directory.')
            if parameters[5].valueAsText.find('MXD') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an MXD directory instead of a Template directory.')
            if parameters[5].valueAsText.find('Output') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an Output directory instead of a Template directory.')
            if parameters[5].valueAsText.find('Image') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an Image directory instead of a Template directory.')
            if parameters[5].valueAsText.find('GDB') > -1:
                parameters[5].setWarningMessage('You appear to be selecting a GDB directory instead of a Template directory.')

        if parameters[6].value:  # output_dir
            if parameters[6].valueAsText.find('XYZ') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an XYZ directory instead of na Output directory.')
            if parameters[6].valueAsText.find('Edit') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an Edit directory instead of an Output directory.')
            if parameters[6].valueAsText.find('MXD') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an MXD directory instead of an Output directory.')
            if parameters[6].valueAsText.find('Template') > -1:
                parameters[6].setWarningMessage('You appear to be selecting a Template directory instead of an Output directory.')
            if parameters[6].valueAsText.find('Image') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an Image directory instead of an Output directory.')
            if parameters[6].valueAsText.find('GDB') > -1:
                parameters[6].setWarningMessage('You appear to be selecting a GDB directory instead of an Output directory.')

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        messages.AddMessage("\n\nGLiMR batch mode\n")
        messages.AddMessage('   Scan Dir = ' + parameters[0].valueAsText)
        messages.AddMessage('   Geodatabase Dir = ' + parameters[1].valueAsText)
        messages.AddMessage('   Geodatabase Edit Dir = ' + parameters[2].valueAsText)
        messages.AddMessage('   MXD Dir = ' + parameters[3].valueAsText)
        messages.AddMessage('   Image Dir = ' + parameters[4].valueAsText)
        messages.AddMessage('   Template Dir = ' + parameters[5].valueAsText)
        messages.AddMessage('   Output Dir = ' + parameters[6].valueAsText)
        messages.AddMessage('\nProcessing started')

        if (sys.maxsize > 2 ** 32):
            is64bit = True
            messages.AddMessage(
                'Using the 64 bit version of Python  Until Spline with Barriers runs in 64 bit, use 32 bit version')
            return
        else:
            is64bit = False
            messages.AddMessage('Using the 32 bit version of Python\n')

        verbose = parameters[7].value               # additional reports to ArcGIS message stream
        compute_flake_flag = parameters[8].value
        make_image_flag = parameters[9].value


        # Set directories from tool parameters

        xyz_dir = parameters[0].valueAsText
        gdb_dir = parameters[1].valueAsText
        gdb_edit_dir = parameters[2].valueAsText
        mxd_dir = parameters[3].valueAsText
        image_dir = parameters[4].valueAsText
        template_dir = parameters[5].valueAsText
        output_dir = parameters[6].valueAsText

        try:

            p = GLiMR(xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir, verbose)  # instance

            # The following statements do all of the work

            p.get_arcgis_licenses()  # always necessary

            p.run_xyz_files(compute_flake_flag, make_image_flag)
            #p.run_transfer_all_stats()
            # p.assemble_procrustes_fc()  delete this once checked moved to process point
            arcpy.GetMessages()

            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')

            messages.AddMessage('\nScript checked in licenses')
            arcpy.GetMessages()

        except arcpy.ExecuteError:
            arcpy.AddMessage('In execute error exception code')
            arcpy.GetMessages()

            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)


        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')


class GLiMR_Single(object):
    """C:\GLiMR\Python\Toolbox.tbx\GLiMR_Single"""

    def __init__(self):
        self.label = u'GLiMR Process single scan file'
        self.description = u'''Run selected scan file.
        Creates GDBs, feature classes, MXD files and images.   Does not run statistics summary'''

        self.canRunInBackground = False

        g_env = GLiMR_env()  # populate directory list from GLiMR.env.txt file for default locations
        self.dir_list = g_env.dir_list

        if self.dir_list:
            self.xyz_dir = self.dir_list[0]
            self.gdb_dir = self.dir_list[1]
            self.mxd_dir = self.dir_list[2]
            self.image_dir = self.dir_list[3]
            self.template_dir = self.dir_list[4]
            self.gdb_edit_dir = self.dir_list[5]
            self.output_dir = self.dir_list[6]
        else:
            self.xyz_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.gdb_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.mxd_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.image_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.template_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.gdb_edit_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.output_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'


    def getParameterInfo(self):

        # Source Scan File
        param_1 = arcpy.Parameter()
        param_1.name = u'xyz_file'
        param_1.displayName = u'Scan File (*.xyz)'
        param_1.parameterType = 'Required'
        param_1.direction = 'Input'
        param_1.datatype = u'DEFile'
        param_1.filter.list = ['xyz']

        # Geodatabase Directory
        param_2 = arcpy.Parameter()
        param_2.name = u'gdb_dir'
        param_2.displayName = u'Geodatabase Directory  (*.gdb)'
        param_2.parameterType = 'Required'
        param_2.direction = 'Input'
        param_2.datatype = u'DEFolder'
        param_2.value = self.gdb_dir

        # Persistent Geodatabase Edit Directory
        param_3 = arcpy.Parameter()
        param_3.name = u'gdb_edit_dir'
        param_3.displayName = u'Persistent Geodatabase Edit Directory'
        param_3.parameterType = 'Required'
        param_3.direction = 'Input'
        param_3.datatype = u'DEFolder'
        param_3.value = self.gdb_edit_dir

        # ArcMap MXD Directory
        param_4 = arcpy.Parameter()
        param_4.name = u'mxd_dir'
        param_4.displayName = u'ArcMap (*.mxd) File Directory'
        param_4.parameterType = 'Required'
        param_4.direction = 'Input'
        param_4.datatype = u'DEFolder'
        param_4.value = self.mxd_dir

        # Image Directory
        param_5 = arcpy.Parameter()
        param_5.name = u'image_dir'
        param_5.displayName = u'Image File Directory'
        param_5.parameterType = 'Required'
        param_5.direction = 'Input'
        param_5.datatype = u'DEFolder'
        param_5.value = self.image_dir

        # Template MXD Directory
        param_6 = arcpy.Parameter()
        param_6.name = u'template_dir'
        param_6.displayName = u'Template (*.mxd) File Directory'
        param_6.parameterType = 'Required'
        param_6.direction = 'Input'
        param_6.datatype = u'DEFolder'
        param_6.value = self.template_dir

        # GLiMR Output Directory
        param_7 = arcpy.Parameter()
        param_7.name = u'output_dir'
        param_7.displayName = u'GLiMR Output Directory'
        param_7.parameterType = 'Required'
        param_7.direction = 'Input'
        param_7.datatype = u'DEFolder'
        param_7.value = self.output_dir

        # GLiMR Verbose flag
        param_8 = arcpy.Parameter()
        param_8.name = u'verbose_flag'
        param_8.displayName = u'GLiMR Verbose flag'
        param_8.parameterType = 'Optional'
        param_8.direction = 'Input'
        param_8.datatype = u'GPBoolean'
        param_8.value = False

        # GLiMR Compute flake flag
        param_9 = arcpy.Parameter()
        param_9.name = u'compute_flake_flag'
        param_9.displayName = u'GLiMR Compute Flake flag'
        param_9.parameterType = 'Optional'
        param_9.direction = 'Input'
        param_9.datatype = u'GPBoolean'
        param_9.value = True

        # GLiMR Make images flag
        param_10 = arcpy.Parameter()
        param_10.name = u'make_images_flag'
        param_10.displayName = u'GLiMR Make images flag'
        param_10.parameterType = 'Optional'
        param_10.direction = 'Input'
        param_10.datatype = u'GPBoolean'
        param_10.value = True

        return [param_1, param_2, param_3, param_4, param_5, param_6, param_7, param_8, param_9, param_10]

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        if parameters[1].value:  # gdb_dir
            if parameters[1].valueAsText.find('XYZ') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an XYZ directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Edit') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an Edit directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Image') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an Image directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Output') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an Output directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('Template') > -1:
                parameters[1].setWarningMessage('You appear to be selecting a Template directory instead of a GDB directory.')
            if parameters[1].valueAsText.find('MXD') > -1:
                parameters[1].setWarningMessage('You appear to be selecting an MXD directory instead of a GDB directory.')

            if parameters[1].value.value == parameters[2].value.value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')

        if parameters[2].value:  # gdb_edit_dir
            if parameters[1].value == parameters[2].value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')
            if parameters[2].valueAsText.find('GLiMR_GDB_Edits') == -1:
                parameters[2].setWarningMessage(
                    "Do not change casually!  GLiMR depends on the correlation between GDB and GDB_Edit geodatabases. Expects 'GLiMR_GDB_Edits'")
            if parameters[1].value.value == parameters[2].value.value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')

        if parameters[3].value:  # mxd_dir
            if parameters[3].valueAsText.find('XYZ') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an XYZ directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Edit') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an Edit directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Image') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an Image directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Output') > -1:
                parameters[3].setWarningMessage('You appear to be selecting an Output directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('GDB') > -1:
                parameters[3].setWarningMessage('You appear to be selecting a GDB directory instead of an MXD directory.')
            if parameters[3].valueAsText.find('Template') > -1:
                parameters[3].setWarningMessage('You appear to be selecting a Template directory instead of an MXD directory.')

        if parameters[4].value:  # image_dir
            if parameters[4].valueAsText.find('XYZ') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an XYZ directory instead of an Image directory.')
            if parameters[4].valueAsText.find('Edit') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an Edit directory instead of an Image directory.')
            if parameters[4].valueAsText.find('MXD') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an MXD directory instead of an Image directory.')
            if parameters[4].valueAsText.find('Output') > -1:
                parameters[4].setWarningMessage('You appear to be selecting an Output directory instead of an Image directory.')
            if parameters[4].valueAsText.find('Template') > -1:
                parameters[4].setWarningMessage('You appear to be selecting n Template directory instead of an Image directory.')
            if parameters[4].valueAsText.find('GDB') > -1:
                parameters[4].setWarningMessage('You appear to be selecting a GDB directory instead of an Image directory.')

        if parameters[5].value:  # template_dir
            if parameters[5].valueAsText.find('XYZ') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an XYZ directory instead of a Template directory.')
            if parameters[5].valueAsText.find('Edit') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an Edit directory instead of a Template directory.')
            if parameters[5].valueAsText.find('MXD') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an MXD directory instead of a Template directory.')
            if parameters[5].valueAsText.find('Output') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an Output directory instead of a Template directory.')
            if parameters[5].valueAsText.find('Image') > -1:
                parameters[5].setWarningMessage('You appear to be selecting an Image directory instead of a Template directory.')
            if parameters[5].valueAsText.find('GDB') > -1:
                parameters[5].setWarningMessage('You appear to be selecting a GDB directory instead of a Template directory.')

        if parameters[6].value:  # output_dir
            if parameters[6].valueAsText.find('XYZ') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an XYZ directory instead of na Output directory.')
            if parameters[6].valueAsText.find('Edit') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an Edit directory instead of an Output directory.')
            if parameters[6].valueAsText.find('MXD') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an MXD directory instead of an Output directory.')
            if parameters[6].valueAsText.find('Template') > -1:
                parameters[6].setWarningMessage('You appear to be selecting a Template directory instead of an Output directory.')
            if parameters[6].valueAsText.find('Image') > -1:
                parameters[6].setWarningMessage('You appear to be selecting an Image directory instead of an Output directory.')
            if parameters[6].valueAsText.find('GDB') > -1:
                parameters[6].setWarningMessage('You appear to be selecting a GDB directory instead of an Output directory.')
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        messages.AddMessage('\nGLiMR: Processing single scan file\n')
        messages.AddMessage('   Scan File = ' + parameters[0].valueAsText)
        messages.AddMessage('   Geodatabase Dir = ' + parameters[1].valueAsText)
        messages.AddMessage('   Geodatabase Edit Dir = ' + parameters[2].valueAsText)
        messages.AddMessage('   MXD Dir = ' + parameters[3].valueAsText)
        messages.AddMessage('   Image Dir = ' + parameters[4].valueAsText)
        messages.AddMessage('   Template Dir = ' + parameters[5].valueAsText)
        messages.AddMessage('   Output Dir = ' + parameters[6].valueAsText)
        messages.AddMessage('\nProcessing started')

        if (sys.maxsize > 2 ** 32):
            is64bit = True
            messages.AddMessage(
                'Using the 64 bit version of Python  Until Spline with Barriers runs in 64 bit, use 32 bit version')
            return
        else:
            is64bit = False
            messages.AddMessage('Using the 32 bit version of Python\n')

        verbose = parameters[7].value               # additional reports to ArcGIS message stream
        compute_flake_flag = parameters[8].value
        make_image_flag = parameters[9].value


        # Set directories from tool parameters

        xyz_file_full = parameters[0].valueAsText
        xyz_dir = os.path.dirname(xyz_file_full)
        xyz_file = os.path.basename(xyz_file_full)
        gdb_dir = parameters[1].valueAsText
        gdb_edit_dir = parameters[2].valueAsText
        mxd_dir = parameters[3].valueAsText
        image_dir = parameters[4].valueAsText
        template_dir = parameters[5].valueAsText
        output_dir = parameters[6].valueAsText

        try:
            p = GLiMR(xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir, verbose)

            # The following statements do all of the work
            p.get_arcgis_licenses()  # always necessary

            p.process_point(xyz_file, compute_flake_flag, make_image_flag)

            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')
            messages.AddMessage(
                '\nSummary statistics databases were not updated. You may manually update them with the '
                'Summarize Statistics tool.\n\nScript checked in licenses')


        except arcpy.ExecuteError:
            print arcpy.GetMessages()

        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')


class GLiMR_SummarizeStats(object):
    """C:\GLiMR\Python\Toolbox.tbx\GLiMR_SummarizeStats"""

    def __init__(self):
        self.label = u'GLiMR Summarize Statistics'
        self.description = u'''Consolidates Outline, Re_entrant, Flake and Procrustes statistics into one database.   Resulting summary Access database in Geodatabase directory'''
        self.canRunInBackground = False

        g_env = GLiMR_env()
        self.dir_list = g_env.dir_list

        if self.dir_list:
            self.gdb_dir = self.dir_list[1]
        else:
            self.gdb_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'


    def getParameterInfo(self):

        # Geodatabase Directory
        param_1 = arcpy.Parameter()
        param_1.name = u'gdb_dir'
        param_1.displayName = u'Geodatabase Directory  (*.gdb)'
        param_1.parameterType = 'Required'
        param_1.direction = 'Input'
        param_1.datatype = u'DEFolder'
        param_1.value = self.gdb_dir

        return [param_1]

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        if parameters[0].value:  # gdb_dir
            if parameters[0].valueAsText.find('XYZ') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an XYZ directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Edit') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Edit directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Image') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Image directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Output') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Output directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Template') > -1:
                parameters[0].setWarningMessage('You appear to be selecting a Template directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('MXD') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an MXD directory instead of a GDB directory.')

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        messages.AddMessage("GLiMR Summarize Statistics\n")
        messages.AddMessage('   Geodatabase Dir = ' + parameters[0].valueAsText)

        messages.AddMessage('\nProcessing started')

        if (sys.maxsize > 2 ** 32):
            is64bit = True
            messages.AddMessage(
                'Using the 64 bit version of Python  Until Spline with Barriers runs in 64 bit, use 32 bit version')
            return
        else:
            is64bit = False
            messages.AddMessage('Using the 32 bit version of Python\n')



        xyz_dir = self.dir_list[0]
        mxd_dir = self.dir_list[2]
        image_dir = self.dir_list[3]
        template_dir = self.dir_list[4]
        gdb_edit_dir = self.dir_list[5]
        output_dir = self.dir_list[6]

        verbose = False  # adds additional reports to ArcGIS message stream

        gdb_dir = parameters[0].valueAsText  # Set directory from tool parameter

        try:

            p = GLiMR(xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir, verbose)  # instance

            # The following statements do all of the work

            p.get_arcgis_licenses()  # always necessary

            p.run_transfer_all_stats()

            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')
            messages.AddMessage('\nScript checked in licenses')



        except arcpy.ExecuteError:
            print arcpy.GetMessages()

        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')


class GLiMR_Landmarks(object):
    """C:\GLiMR\Python\Toolbox.tbx\GLiMR_Landmarks"""

    def __init__(self):
        self.label = u'GLiMR Compute Stemmed Point Landmarks'
        self.description = u'''Compute landmarks for all geodatabases in Geodatabases directory.
        Creates landmark (Procrustes) feature classes.'''

        self.canRunInBackground = False
        arcpy.env.overwriteOutput = True

        g_env = GLiMR_env()  # populate directory list from GLiMR.env.txt file for default locations
        self.dir_list = g_env.dir_list

        if self.dir_list:
            self.gdb_dir = self.dir_list[1]
        else:
            self.gdb_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'


    def getParameterInfo(self):

        # Geodatabase Directory
        param_1 = arcpy.Parameter()
        param_1.name = u'gdb_dir'
        param_1.displayName = u'Geodatabase Directory  (*.gdb)'
        param_1.parameterType = 'Required'
        param_1.direction = 'Input'
        param_1.datatype = u'DEFolder'
        param_1.value = self.gdb_dir

        # Number of Rings
        param_2 = arcpy.Parameter()
        param_2.name = u'rings'
        param_2.displayName = u'Number of Rings'
        param_2.parameterType = 'Required'
        param_2.direction = 'Input'
        param_2.datatype = u'GPLong'
        param_2.value = 1

        # Percent Outward Shift
        param_3 = arcpy.Parameter()
        param_3.name = u'pct_outward_shift'
        param_3.displayName = u'Percent Outward Shift'
        param_3.parameterType = 'Required'
        param_3.direction = 'Input'
        param_3.datatype = u'GPDouble'
        param_3.value = 0.0

        # Blade_rows
        param_4 = arcpy.Parameter()
        param_4.name = u'blade_rows'
        param_4.displayName = u'Number of blade rows'
        param_4.parameterType = 'Required'
        param_4.direction = 'Input'
        param_4.datatype = u'GPLong'
        param_4.value = 5

        # Blade_cols
        param_5 = arcpy.Parameter()
        param_5.name = u'blade_cols'
        param_5.displayName = u'Number of blade columns'
        param_5.parameterType = 'Required'
        param_5.direction = 'Input'
        param_5.datatype = u'GPLong'
        param_5.value = 5

        # Blade top percent setback
        param_6 = arcpy.Parameter()
        param_6.name = u'blade_top_pct_setback'
        param_6.displayName = u'Blade top percent setback'
        param_6.parameterType = 'Required'
        param_6.direction = 'Input'
        param_6.datatype = u'GPDouble'
        param_6.value = 15.0

        # Blade base percent setback
        param_7 = arcpy.Parameter()
        param_7.name = u'blade_base_pct_setback'
        param_7.displayName = u'Blade base percent setback'
        param_7.parameterType = 'Required'
        param_7.direction = 'Input'
        param_7.datatype = u'GPDouble'
        param_7.value = 10.0

        # Haft_rows
        param_8 = arcpy.Parameter()
        param_8.name = u'haft_rows'
        param_8.displayName = u'Number of haft rows'
        param_8.parameterType = 'Required'
        param_8.direction = 'Input'
        param_8.datatype = u'GPLong'
        param_8.value = 4

        # Haft_cols
        param_9 = arcpy.Parameter()
        param_9.name = u'haft_cols'
        param_9.displayName = u'Number of haft columns'
        param_9.parameterType = 'Required'
        param_9.direction = 'Input'
        param_9.datatype = u'GPLong'
        param_9.value = 5

        # Haft top percent setback
        param_10 = arcpy.Parameter()
        param_10.name = u'haft_top_pct_setback'
        param_10.displayName = u'Haft top percent setback'
        param_10.parameterType = 'Required'
        param_10.direction = 'Input'
        param_10.datatype = u'GPDouble'
        param_10.value = 10.0

        # Haft base percent setback
        param_11 = arcpy.Parameter()
        param_11.name = u'haft_base_pct_setback'
        param_11.displayName = u'Haft base percent setback'
        param_11.parameterType = 'Required'
        param_11.direction = 'Input'
        param_11.datatype = u'GPDouble'
        param_11.value = 22.0

        return [param_1, param_2, param_3, param_4, param_5, param_6, param_7, param_8, param_9, param_10, param_11]

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        if parameters[0].value:  # gdb_dir
            if parameters[0].valueAsText.find('XYZ') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an XYZ directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Edit') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Edit directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Image') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Image directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Output') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Output directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Template') > -1:
                parameters[0].setWarningMessage('You appear to be selecting a Template directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('MXD') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an MXD directory instead of a GDB directory.')

        if parameters[1].value:                                 # rings
            if parameters[1].value < 1 or parameters[1].value > 5:
                parameters[1].setErrorMessage('The number of rings must be between 1 and 5')

        if parameters[2].value:                                 # percent outward shift
            if parameters[2].value < -100 or parameters[2].value > 100:
                parameters[2].setErrorMessage(
                    'Please enter a number between 0 and 100.    O does not shift internal ring location.  100 shifts it to the point outline')

            if parameters[2].value < 0 and parameters[2].value > -100:  # negative percent outward shift
                parameters[2].setWarningMessage('Using a negative number shifts points inward.  You clever dog!')

        if parameters[3].value:
            if parameters[3].value < 1 or parameters[3].value > 10:                        # blade_rows
                parameters[3].setErrorMessage('The number of blade rows must be between 1 and 10')

        if parameters[4].value:
            if parameters[4].value < 1 or parameters[4].value > 10:                        # blade_cols
                parameters[4].setErrorMessage('The number of blade columns must be between 1 and 10')

        if parameters[5].value:                                 # blade_top_pct_setback
            if parameters[5].value < 0. or parameters[5].value > 100.:
                parameters[5].setErrorMessage('The top blade setback needs to be between 0. and 100.\n'
                                              'Reasonable expected values range between 5. and 40.')
            if parameters[5].value < 5.:
                parameters[5].setWarningMessage('Small values of top blade setback force many columns into a small width near the point.\n'
                                              'Reasonable expected values range between 5. and 40.')

        if parameters[6].value:                                 # blade_base_pct_setback
            if parameters[6].value < 0. or parameters[6].value > 100.:
                parameters[6].setErrorMessage('The top blade setback needs to be between 0. and 100.\n'
                                              'Reasonable expected values range between 0. and 40.')

        if parameters[7].value:
            if parameters[7].value < 1 or parameters[7].value > 10:                        # haft_rows
                parameters[7].setErrorMessage('The number of haft columns must be between 1 and 10')

        if parameters[8].value:
            if parameters[8].value < 1 or parameters[8].value > 10:                        # haft_cols
                parameters[8].setErrorMessage('The number of haft columns must be between 1 and 10')

        if parameters[7].value:                                 # haft_top_pct_setback
            if parameters[9].value < 0. or parameters[9].value > 100.:
                parameters[9].setErrorMessage('The top haft setback needs to be between 0. and 100.\n'
                                              'Reasonable expected values range between 5. and 40.')

        if parameters[10].value:                                 # haft_base_pct_setback
            if parameters[10].value < 0. or parameters[10].value > 100.:
                parameters[10].setErrorMessage('The top haft setback needs to be between 0. and 100.\n'
                                              'Reasonable expected values range between 0. and 40.')
            if parameters[10].value < 5.:
                parameters[10].setWarningMessage('Small values of base haft setback force many columns into a small width near the haft base.\n'
                                              'Reasonable expected values range between 5. and 40.')

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        messages.AddMessage('GLiMR Computing all landmarks\n')
        messages.AddMessage('   Geodatabase Dir = ' + parameters[0].valueAsText)

        messages.AddMessage('\nProcessing started')

        if (sys.maxsize > 2 ** 32):
            is64bit = True
            messages.AddMessage(
                'Using the 64 bit version of Python  Until Spline with Barriers runs in 64 bit, use 32 bit version')
            return
        else:
            is64bit = False
            messages.AddMessage('Using the 32 bit version of Python\n')

        arcpy.GetMessages()

        xyz_dir = self.dir_list[0]
        gdb_dir = self.dir_list[1]
        mxd_dir = self.dir_list[2]
        image_dir = self.dir_list[3]
        template_dir = self.dir_list[4]
        gdb_edit_dir = self.dir_list[5]
        output_dir = self.dir_list[6]

        verbose = False  # verbose outputs to ArcGIS message stream

        gdb_dir = parameters[0].valueAsText

        try:

            p = GLiMR(xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir, verbose)  # instance

            # The following statements do all of the work

            p.get_arcgis_licenses()  # always necessary

            p.compute_landmarks(parameters[1].value, parameters[2].value, parameters[3].value, parameters[4].
                              value, parameters[5].value, parameters[6].value, parameters[7].value,
                              parameters[8].value, parameters[9].value, parameters[10].value)

            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')
            messages.AddMessage('\nScript checked in licenses')
            arcpy.GetMessages()

        except arcpy.ExecuteError:
            print arcpy.GetMessages()

        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')

class GLiMR_MorphoJ(object):
    """C:\GLiMR\Python\Toolbox.tbx\GGLiMR_MorphoJ"""

    def __init__(self):
        self.label = u'GLiMR Write MorphoJ file'
        self.description = u'''Write MorphoJ input text file for PRE-EXISTING Procrustes Landmarks.'''

        self.canRunInBackground = False

        g_env = GLiMR_env()  # populate directory list from GLiMR.env.txt file for default locations
        self.dir_list = g_env.dir_list

        if self.dir_list:
            self.gdb_dir = self.dir_list[1]
            self.output_dir = self.dir_list[6]
        else:
            self.gdb_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'
            self.output_dir = '\n\n*** Missing GLiMR environment file GLiMR.env.txt  ***  Please fix and re run\n'

    def getParameterInfo(self):

        # Geodatabase Directory
        param_1 = arcpy.Parameter()
        param_1.name = u'gdb_dir'
        param_1.displayName = u'Geodatabase Directory  (*.gdb)'
        param_1.parameterType = 'Required'
        param_1.direction = 'Input'
        param_1.datatype = u'DEFolder'
        param_1.value = self.gdb_dir

        # Output file for Topo-based landmarks
        param_2 = arcpy.Parameter()
        param_2.name = u'Procrustes_Ext_Topo_filename'
        param_2.displayName = u'Procrustes 2D landmarks using Topo Z value'
        param_2.parameterType = 'Required'
        param_2.direction = 'Output'
        param_2.datatype = u'DEFile'
        param_2.value = self.output_dir + '\\ProcrustesExt2DTopo.txt'
        param_2.filter.list = ['txt']

        # Output file for Smoothed-based landmarks
        param_3 = arcpy.Parameter()
        param_3.name = u'Procrustes_Ext_Smoothed_filename'
        param_3.displayName = u'Procrustes 2D landmarks using Smoothed Z value'
        param_3.parameterType = 'Required'
        param_3.direction = 'Output'
        param_3.datatype = u'DEFile'
        param_3.value = self.output_dir + '\\ProcrustesExt2DSmooth.txt'
        param_3.filter.list = ['txt']

        # Output file for Topo-based landmarks
        param_4 = arcpy.Parameter()
        param_4.name = u'Procrustes_Int_Topo_filename'
        param_4.displayName = u'Procrustes 3D landmarks using Topo Z value'
        param_4.parameterType = 'Required'
        param_4.direction = 'Output'
        param_4.datatype = u'DEFile'
        param_4.value = self.output_dir + '\\ProcrustesInt3DTopo.txt'
        param_4.filter.list = ['txt']

        # Output file for Topo-based landmarks
        param_5 = arcpy.Parameter()
        param_5.name = u'Procrustes_Int_Smoothed_filename'
        param_5.displayName = u'Procrustes 3D landmarks using Smoothed Z value'
        param_5.parameterType = 'Required'
        param_5.direction = 'Output'
        param_5.datatype = u'DEFile'
        param_5.value = self.output_dir + '\\ProcrustesInt3DSmooth.txt'
        param_5.filter.list = ['txt']


        return [param_1, param_2, param_3, param_4, param_5]

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        if parameters[0].value:  # gdb_dir
            if parameters[0].valueAsText.find('XYZ') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an XYZ directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Edit') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Edit directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Image') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Image directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Output') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an Output directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('Template') > -1:
                parameters[0].setWarningMessage('You appear to be selecting a Template directory instead of a GDB directory.')
            if parameters[0].valueAsText.find('MXD') > -1:
                parameters[0].setWarningMessage('You appear to be selecting an MXD directory instead of a GDB directory.')

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        messages.AddMessage('\nGLiMR: Writing four Procrustes landmark files for MorphoJ\n')
        messages.AddMessage('   Geodatabase Dir = ' + parameters[0].valueAsText)
        messages.AddMessage('   Topo Landmark file name = ' + parameters[1].valueAsText)
        messages.AddMessage('   Smoothed Landmark file name ' + parameters[2].valueAsText)
        messages.AddMessage('\nPoint count is the number of points found in the current artifact.'
                            '\nBeware of different landmark counts...\n')

        arcpy.GetMessages()

        verbose = False  # verbose output to ArcGIS message stream

        # Set variables from tool parameters
        output_dir = parameters[0].valueAsText
        Procrustes_Ext_Topo_filename = parameters[1].valueAsText
        Procrustes_Ext_Smooth_filename = parameters[2].valueAsText
        Procrustes_Int_Topo_filename = parameters[3].valueAsText
        Procrustes_Int_Smooth_filename = parameters[4].valueAsText

        try:
            arcpy.env.workspace = output_dir + '\\a_GLiMR_Summary.mdb'

            if not  arcpy.Exists('ProcrustesExteriorSummary'):
                arcpy.AddError('Cannot find ProcrustesExteriorSummary table in ' + arcpy.env.workspace)
                arcpy.AddMessage('Be sure you have previously created landmarks AND run Summarize Statistics')
                sys.exit(-1)

            if not  arcpy.Exists('ProcrustesInteriorSummary'):
                arcpy.AddError('Cannot find ProcrustesInteriorSummary table in ' + arcpy.env.workspace)
                exit()

            if os.path.exists(Procrustes_Ext_Topo_filename):
                arcpy.Delete_management(Procrustes_Ext_Topo_filename)

            if os.path.exists(Procrustes_Ext_Smooth_filename):
                arcpy.Delete_management(Procrustes_Ext_Smooth_filename)

            if os.path.exists(Procrustes_Int_Topo_filename):
                arcpy.Delete_management(Procrustes_Int_Topo_filename)

            if os.path.exists(Procrustes_Int_Smooth_filename):
                arcpy.Delete_management(Procrustes_Int_Smooth_filename)

            arcpy.AddMessage('Deleted previous Procrustes text files from GLiMR Output directory\n')

            
            # External Procrustes files
            
            arcpy.AddMessage('External Procrustes Points\n')

            ft = open(Procrustes_Ext_Topo_filename, mode = 'a')
            fs = open(Procrustes_Ext_Smooth_filename, mode = 'a')

            ptName = 'NotDefined'
            landmark_count = 0
            pt_count = 0
            pcn = 0
            last_gkey = 'None'

            pcr_ExtSum = arcpy.env.workspace + '\\ProcrustesExteriorSummary'

            with arcpy.da.SearchCursor(pcr_ExtSum, ['GKey','ProcrustesPtNum'],
                                       sql_clause=(None, 'ORDER BY GKey, ProcrustesPtNum')) as cursor:
                for row in cursor:
                    pcn += 1
                    gkey = row[0]
                    if gkey != last_gkey:                # first point for new gkey
                        if last_gkey != 'None':
                            landmark_count = pcn
                            arcpy.AddMessage('External Landmark count for ' + last_gkey + ' is ' + str(landmark_count))

                        last_gkey = gkey
                        pcn = 0

                # get last row
                arcpy.AddMessage('External Landmark count for ' + gkey + ' is ' + str(pcn) + '    (last record)')

            del cursor
            del row


            cursor = sorted(arcpy.da.SearchCursor(pcr_ExtSum, ['GKey', 'ProcrustesPtNum',
                                                            'X', 'Y', 'Z', 'Label']))

            print >> ft, 'LM3=' + str(landmark_count)    # first line of file
            print >> fs, 'LM3=' + str(landmark_count)

            for row in cursor:
                gkey = row[0]
                if row[2] is not None:
                    x = str('%0.4f'%row[2])
                else:
                    x = '0'

                if row[3] is not None:
                    y = str('%0.4f'%row[3])
                else:
                    y = '0'

                if row[4] is not None:
                    zt = str('%0.4f'%row[4])
                else:
                    zt = '0'

                if row[5] is not None:              # duplicating z topo
                    zs = str('%0.4f'%row[4])
                else:
                    zs = '0'

                if ptName =='NotDefined':         # first record
                    ptName = row[0]

                if row[0] == ptName:            # same point
                    st = x + ' ' + y + ' ' + zt
                    ss = x + ' ' + y + ' ' + zs

                    ft.writelines(st + '\n')
                    fs.writelines(ss + '\n')
                    pt_count += 1

                else:                           # next point, print tail and head info
                    print >> ft, 'ID=' + str(ptName)        # name of previous point
                    print >> fs, 'ID=' + str(ptName)    # name of previous point
                    print >> ft, 'LM3=' + str(landmark_count)
                    print >> fs, 'LM3=' + str(landmark_count)

                    ptName = row[0]             # next point name

                    st = x + ' ' + y + ' ' + zt
                    ss = x + ' ' + y + ' ' + zs
                    ft.writelines(st + '\n')
                    fs.writelines(ss + '\n')
                    pt_count += 1

                    # print >> ft, st
                    # print >> fs, ss

            del cursor
            del row

            print >> ft, 'ID=' + str(ptName)  + '\n'  # final point tail info
            print >> fs, 'ID=' + str(ptName)  + '\n'

            ft.close()
            fs.close()



            # Internal Procrustes files
            
            arcpy.AddMessage('\n\nInternal Points\n')

            ft = open(Procrustes_Int_Topo_filename, mode = 'a')
            fs = open(Procrustes_Int_Smooth_filename, mode = 'a')

            ptName = 'NotDefined'
            landmark_count = 0
            pt_count = 0
            pcn = 0
            last_gkey = 'None'

            pcr_Int_Sum = arcpy.env.workspace + '\\ProcrustesInteriorSummary'

            with arcpy.da.SearchCursor(pcr_Int_Sum, ['GKey','ProcrustesPtNum'],
                                       sql_clause=(None, 'ORDER BY GKey, ProcrustesPtNum')) as cursor:
                for row in cursor:
                    pcn += 1
                    gkey = row[0]
                    if gkey != last_gkey:                # first point for new gkey
                        if last_gkey != 'None':
                            landmark_count = pcn
                            arcpy.AddMessage('Internal Landmark count for ' + last_gkey + ' is ' + str(landmark_count))

                        last_gkey = gkey
                        pcn = 0

                # get last row
                arcpy.AddMessage('Internal Landmark count for ' + gkey + ' is ' + str(pcn) + '    (last record)')
            del cursor
            del row


            cursor = sorted(arcpy.da.SearchCursor(pcr_Int_Sum,  ['GKey', 'ProcrustesPtNum',
                                                             'X', 'Y', 'Z_Topo', 'Z_Smooth', 'Label']))
            print >> ft, 'LM3=' + str(landmark_count)    # first line of file
            print >> fs, 'LM3=' + str(landmark_count)

            for row in cursor:
                if row[2] is not None:
                    x = str('%0.4f'%row[2])
                else:
                    x = '0'

                if row[3] is not None:
                    y = str('%0.4f'%row[3])
                else:
                    y = '0'

                if row[4] is not None:
                    zt = str('%0.4f'%row[4])
                else:
                    zt = '0'

                if row[5] is not None:
                    zs = str('%0.4f'%row[5])
                else:
                    zs = '0'

                if ptName =='NotDefined':         # first record
                    ptName = row[0]

                if row[0] == ptName:            # same point
                    st = x + ' ' + y + ' ' + zt
                    ss = x + ' ' + y + ' ' + zs

                    ft.writelines(st + '\n')
                    fs.writelines(ss + '\n')
                    pt_count += 1

                else:                           # next point, print tail and head info
                    print >> ft, 'ID=' + str(ptName)        # name of previous point
                    print >> fs, 'ID=' + str(ptName)    # name of previous point
                    print >> ft, 'LM3=' + str(landmark_count)
                    print >> fs, 'LM3=' + str(landmark_count)

                    ptName = row[0]             # next point name

                    st = x + ' ' + y + ' ' + zt
                    ss = x + ' ' + y + ' ' + zs
                    ft.writelines(st + '\n')
                    fs.writelines(ss + '\n')
                    pt_count += 1

            del cursor
            del row

            print >> ft, 'ID=' + str(ptName)  + '\n'  # final point tail info
            print >> fs, 'ID=' + str(ptName)  + '\n'

            ft.close()
            fs.close()


            arcpy.AddMessage('\nFour MorphJ Procrustes landmark files written')


        except SystemExit:
            pass

        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            ft.close()
            fs.close()



class GLiMR_CheckEnvironment(object):
    """C:\GLiMR\Python\Toolbox.tbx\GLiMR_CheckEnvironment"""

    def __init__(self):
        self.label = u'GLiMR Check environment and version'
        self.description = u'''Check environment and version.
        Checks ArcGIS Desktop version, Spatial Analyst and 3D Analyst licenses'''

        import time
        self.canRunInBackground = False

    def getParameterInfo(self):
        return True


    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        '''
        if parameters[2].value:  # gdb_edit_dir
            if parameters[1].value == parameters[2].value:
                parameters[1].setErrorMessage('The GDB and GDB_Edit directory may not the the same.')
            if parameters[2].valueAsText.find('GLiMR_GDB_Edits') == -1:
                parameters[2].setWarningMessage(
                    "Do not change casually!  GLiMR depends on the correlation between GDB and GDB_Edit geodatabases. Expects 'GLiMR_GDB_Edits'")
            if parameters[1].value.value == parameters[2].value.value:
                parameters[1].setErrorMessage('Error in Check.')
        '''

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        arcpy.AddMessage('\n\nGLiMR Environment check\n')

        statusMsg = 'ArcGIS Desktop and extensions appear ready to use.'

        d = arcpy.GetInstallInfo()       # get installation and version info
        for key, value in d.iteritems():
            arcpy.AddMessage("{:<13} : {}".format(key, value))

        FullVersion = d.get("Version")
        v = FullVersion.split(".")

        if v[0] != '10':
            arcpy.AddMessage('This version of GLiMR requires ArcGIS 10.2 and this is version ' + FullVersion)
            arcpy.AddMessage('\n*** GLiMR will not run as configured ***\n')
            statusMsg = 'Version error.  GLiMR will not run as configured.'

        if v[0] == '10' and v[1] < '2':
            arcpy.AddMessage('This version of GLiMR requires ArcGIS 10.2 and this is version ' + FullVersion)
            arcpy.AddMessage('\n*** GLiMR will not run as configured ***\n')
            statusMsg = 'Version error.  GLiMR will not run as configured. ArcGIS 10.2 or later is required.'

        if v[0] == '10' and v[1] == '2':
            arcpy.AddMessage('\n\nThe correct version of ArcGIS Desktop is installed for GLiMR:  ' + FullVersion)


        if (sys.maxsize > 2 ** 32):     # check 32 bit or 64 bit python
            messages.AddMessage(
                'Using the 64 bit version of Python  Until Spline with Barriers runs in 64 bit, use 32 bit version')
            statusMsg = statusMsg + '\nPlease configure to use 32-bit version of Python,' \
                                    ' rather than the 64-bit version currently configured.'
        else:
            messages.AddMessage('Using the correct (32 bit) version of Python.')


        try:                            # Check ArcGIS extension licenses
            if arcpy.CheckExtension('3D') == 'Available':
                arcpy.CheckOutExtension('3D')
                messages.AddMessage('The 3D Analyst extension was successfully checked out.')
            else:
                raise LicenseError
        except LicenseError:
            messages.AddMessage('3D Analyst license is unavailable')
        except:
             messages.AddMessage(arcpy.GetMessages(2))

        try:
            if arcpy.CheckExtension('Spatial') == 'Available':
                arcpy.CheckOutExtension('Spatial')
                messages.AddMessage('The Spatial Analyst extension was successfully checked out.')

            else:
                raise LicenseError
        except LicenseError:
             messages.AddMessage( 'Spatial Analyst license is unavailable')
        except:
             messages.AddMessage( arcpy.GetMessages(2))

        try:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')
            messages.AddMessage('Licenses checked back in...\n')

            messages.AddMessage(statusMsg + '\n\n')

        except arcpy.ExecuteError:
            print arcpy.GetMessages()

        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')
            # time.sleep(10)

class GLiMR_Copy_Scar(object):
    """C:\GLiMR\Python\Toolbox.tbx\GGLiMR_Copy_Scar"""

    def __init__(self):
        self.label = u'GLiMR Copy Scar feature class within Artifact dataset'
        self.description = u'''Copy Scar feature class.
        Copies scar feature class within Artifact dataset and sets up topology and topology rules'''

        self.canRunInBackground = False

    def getParameterInfo(self):
        arcpy.env.workspace = r'c:\GLiMR\GLiMR_Edit_GDB'

        # Input Scar polygon feature class
        param_1 = arcpy.Parameter()
        param_1.name = u'Input Scar feature class'
        param_1.displayName = u'Input Scar feature class'
        param_1.parameterType = 'Required'
        param_1.direction = 'Input'
        param_1.datatype = u'DEFeatureClass'
        param_1.filter.list = ['Polygon']

        # Input Scar polygon feature class
        param_2 = arcpy.Parameter()
        param_2.name = u'Output Scar feature class'
        param_2.displayName = u'Output Scar feature class'
        param_2.parameterType = 'Required'
        param_2.direction = 'Output'
        param_2.datatype = u'DEFeatureClass'

        return [param_1, param_2]

    def isLicensed(self):
        return True

    def updateParameters(self, parameters):
        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateParameters()

    def updateMessages(self, parameters):

        Scar_input = parameters[0].valueAsText
        Scar_output = parameters[1].valueAsText

        if parameters[0].value:  # input scar feature class
            if parameters[0].value.value == parameters[1].value.value:
                parameters[0].setErrorMessage('The input and output file names may not the the same.')

            if Scar_input.find('Artifact') == -1 :  # input not in Artifact feature dataset
                parameters[0].setErrorMessage('The input scar feature class must be within the Artifact feature dataset')

            elif Scar_input.find('Artifact') == -1 and Scar_output.find('Artifact') == -1:
                parameters[0].setErrorMessage('Both input and output scar feature class must be within the same Artifact feature dataset')

        if parameters[1].value:  # input scar feature class

            if Scar_output.find('Artifact') == -1:  # output not in Artifact feature dataset
                parameters[1].setErrorMessage('The output scar feature class must be within the same Artifact feature dataset as the input (due to topology)')

        validator = getattr(self, 'ToolValidator', None)
        if validator:
            return validator(parameters).updateMessages()

    def execute(self, parameters, messages):

        messages.AddMessage('\nGLiMR: Copying scar feature class within artifact dataset\n')
        messages.AddMessage('   Input Scar feature class = ' + parameters[0].valueAsText)
        messages.AddMessage('   Output Scar feature class = ' + parameters[1].valueAsText)
        # messages.AddMessage('   Geodatabase Edit Dir = ' + parameters[2].valueAsText)


        if (sys.maxsize > 2 ** 32):
            is64bit = True
            messages.AddMessage(
                'Using the 64 bit version of Python  Until Spline with Barriers runs in 64 bit, use 32 bit version')
            return
        else:
            is64bit = False
            messages.AddMessage('Using the 32 bit version of Python\n')

        arcpy.GetMessages()

        verbose = False  # verbose output to ArcGIS message stream


        # Set directories from tool parameters



        try:
            # arcpy.FeatureClassToFeatureClass_conversion("buildings_point", "C:/output/output.gdb", "buildings_point")
            Scar_input = parameters[0].valueAsText
            Scar_output = parameters[1].valueAsText

            fc = os.path.basename(Scar_output)
            fds = os.path.dirname(Scar_output)

            parts = fds.split('.gdb')
            full_gdb = parts[0] + '.gdb'

            messages.AddMessage('fds: ' + fds)
            messages.AddMessage('gdb: ' + full_gdb)
            messages.AddMessage('fc: ' + fc)

            # copy_scar_fc(fc,)
            messages.AddMessage('\nFeature class copied and topology updated.')

            arcpy.GetMessages()

        except:
            # Get the traceback object
            tb = sys.exc_info()[2]
            tbinfo = traceback.format_tb(tb)[0]
            # Concatenate error information into message string
            pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                .format(tbinfo, str(sys.exc_info()[1]))
            msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
            # Return python error messages for script tool or Python Window
            arcpy.AddError(pymsg)
            arcpy.AddError(msgs)

        finally:
            arcpy.CheckInExtension('3D')
            arcpy.CheckInExtension('Spatial')

def copy_scar_fc(self, infc, outfc, edit_gdb):
    # Copy scar fc 'infc' to 'outfc' within feature dataset Artifact in edit gdb
    # and add outfc to topology and set up rules
    # infc and outfc are short names of feature classes
    # set up so can be called without GLiMR object

    if arcpy.Exists(edit_gdb + '\\Artifact\\' + infc):
        self.add_fc_to_artifact_fds(infc, outfc, edit_gdb + '\\Artifact')
    else:
        arcpy.AddMessage(infc + ' does not exist in ' + edit_gdb + '\\Artifact')


def delete_scar_fc(self, fc, edit_gdb):
    # Remove scar fc from topology  then delete it from feature dataset Artifact in edit gdb
    # fc is short name of feature class
    # set up so can be called without GLiMR object

    if arcpy.Exists(edit_gdb + '\\Artifact\\' + fc):

        try:
                # Must remove from topology, then delete in edit_gdb,
                # expect it is in topology as well, so just look there first
                artifact_topology = edit_gdb + '\\Artifact\\Scar_topology'

                if arcpy.Exists(artifact_topology):     # if topology present remove fc from topology
                    d = arcpy.Describe(artifact_topology)
                if len(d.featureClassNames) > 0:
                    if fc in d.featureClassNames:        # Is fc in Scar_topology?
                        arcpy.RemoveFeatureClassFromTopology_management(artifact_topology, fc)
                        # self.topology example is C:\GLiMR\GLiMR_GDB_Edits\G9943_1_1.gdb\Artifact\Scar_topology
                self.delete_preexisting(edit_gdb + '\\Artifact\\' +  fc)
                return 1
        except:
            arcpy.AddMessage('*** Exception encountered in delete_scar_fc.  Processing terminated.')
            exit()

    else:
        arcpy.AddMessage(fc + ' does not exist in ' +  edit_gdb + '\\Artifact')
        return 0


def rename_scar_fc(self, oldfc, newfc, edit_gdb):
    # Rename scar fc 'oldfc' to 'newfc' within feature dataset Artifact in edit gdb
    # and add newfc to topology and set up rules
    # set up so can be called without GLiMR object

    if arcpy.Exists(edit_gdb + '\\Artifact\\' + fc):
        self.copy_scar_fc(fc)
        self.delete_scar_fc(fc)
    else:
        arcpy.AddMessage(fc + ' does not exist in ' +  edit_gdb + '\\Artifact  feature dataset' )


class GLiMR(object):
    ''' Primary object representing a lithic projectile point.

    An instance has data attributes of geodatabase names and directories for both persistent and user data.
    GLiMR methods are all of the functions to create and analyze projectile points.
    GLiMR data attributes are created for certain key shared metrics for outline and haft properties.
    '''

    def __init__(self, xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir, verbose):
        self.xyz_dir = xyz_dir
        self.gdb_dir = gdb_dir
        self.mxd_dir = mxd_dir
        self.image_dir = image_dir
        self.template_dir = template_dir
        self.gdb_edit_dir = gdb_edit_dir
        self.output_dir = output_dir
        self.pt_name = ''
        self.gkey = ''
        self.sr = arcpy.SpatialReference(32601)     # WGS 1984 UTN Zone 1N    (arbitrary choice)
        self.verbose = verbose
        self.version = 1.0
        self.haft_collar_E_pnt = (0,0)
        self.haft_collar_W_pnt = (0,0)
        self.zRange = 0.
        # self.logfile = open(self.output_dir + '\\' + 'Logfile.txt', 'w')

        arcpy.env.workspace = xyz_dir  # Set workspace location
        arcpy.env.overwriteOutput = True  # NOTE !

        arcpy.AddMessage('GLiMR   Version ' + str(self.version) + '\nGIS-Based Lithic Morphometric Research\n')


    def run_xyz_files(self, compute_flake_flag=True, make_image_flag=True):
        '''For each xyz file in xyz_dir, create and populate geodatabase for that point.

        All GDB and GDB_Edit directory names are derived from the xyz_file name.
        '''

        xyz_files = arcpy.ListFiles('*.xyz')        # list of files to process

        if len(xyz_files) == 0:
            # no files in directory, nothing to do
            arcpy.AddMessage('Nothing to do. No .xyz files found in ' + arcpy.env.workspace)
            exit()                                  # DEBUG: throw exception instead  (revisit)

        arcpy.AddMessage('Found ' + str(len(xyz_files)) + ' file(s) to process in ' + self.xyz_dir + '\n')
        arcpy.env.workspace = self.gdb_dir

        for xyz_file in xyz_files:
            arcpy.AddMessage('Processing ' + xyz_file)
            self.process_point(xyz_file, compute_flake_flag, make_image_flag)
        del xyz_files
        # self.logfile.close()
        return True


    def process_point(self, xyz_file, compute_flake_flag=True, make_image_flag=True):
        ''' Process a single xyz file for geometry and flakes

        Sequentially
        - Determines component names
        - Builds geometry
        - Fetches persistent data
        - Create flake scars (if compute_flake_flag == True)
        - Create MXD files and images  (if make_image_flag == True)
        '''

        arcpy.env.workspace = self.gdb_dir              # Workspace at GDB directory level...

        root_pt_name = os.path.splitext(xyz_file)[0]
        self.parse_pt_name(root_pt_name)                # parse file name to set self.gkey of cat_face_ver  e.g. 123_2_1

        pt_gdb = 'G' + self.gkey + '.gdb'               # example: G123_2_1.gdb
        final_gdb = self.gdb_dir + '\\' + pt_gdb        # example: G123_2_1.gdb is final geodatabase with path
        edit_gdb = self.gdb_edit_dir + '\\' + pt_gdb    # example: G123_2_1.gdb is edit geodatabase with path

        self.xyz_file = xyz_file                        # example: G123_2_1.xyz contains X Y Z point triplets
        self.pt_name = root_pt_name                     # example: G123_2_1 is the current point name
        self.pt_gdb = pt_gdb                            # example: G123_2_1.gdb is geodatabase for the current point
        self.final_gdb = final_gdb                      # example: <path>\G123_2_1.GDB is geodatabase with full path
        self.edit_gdb = edit_gdb                        # example: <path>\G123_2_1.GDB is geodatabase with full path
        self.artifact_dataset = edit_gdb + '\\Artifact'                  # full path to artifact dataset in edit gdb
        self.artifact_topology = edit_gdb + '\\Artifact\Scar_topology'   # full path for associated topology

        if self.verbose:
            arcpy.AddMessage('     ' + xyz_file + ' being read.  Analysis started for this point.\n')

        arcpy.AddMessage('   Building geometry...')
        self.build_geometry(xyz_file)
        self.load_edit_fc()

        if compute_flake_flag:
            arcpy.AddMessage('   Creating and aggregating flake scars...')
            self.create_auto_flake_scars()
        else:
            arcpy.AddMessage('   Skipping creating and aggregating flake scars...')

        if make_image_flag:
            arcpy.AddMessage('   Creating MXD and images...')
            self.create_mxd_and_images()
        else:
            arcpy.AddMessage('   Skipping creating MXD and images...')

        return True

    def parse_pt_name(self, ptname):
        ''' Parses point name from xyz_file root to determine an set unique identifier gkey  (self.gkey).

         gkey is concatenated <catalog_face_version> in the form 123_2_1

         The point name is the XYZ file name stripped of '.xyz'.
         Expects point names in form of pptp123, pptp123_2, pptp123_1_2 or xxxxxpptp123
         If 'pptp' is embedded, it ignores any characters before 'pptp'
         defaults version and face to 1 if not they are not explicitly specified

         If unsuccessful parsing, gkey is set to the point name (ptname)
        '''

        self.gkey = ''
        face_str = ''
        version_str = ''

        s = ptname.lower()
        i = s.find('pptp') + 4
        pt_str = s[i:]                          # strips 'pptp' and everything before

        if i > 3:                               # 'pptp' or 'PPTP' is found in file name
            tokens = pt_str.split('_')
            if len(tokens) == 3:                # possibly Catalog number in form pptp123_F1_V1
                catalog_string = tokens[0]
                if catalog_string.isdigit():
                    catalog_str = str(int(catalog_string))

                    face_string = tokens[1]     # possible Face as 'F1' or '1'

                    if len(face_string) == 0:   # no face supplied, default to 1
                        face_str = '1'          # This could be double underscore case
                        version_str = '1'

                    elif len(face_string) == 1:
                        if face_string.isdigit():
                            face_str = face_string
                        else:
                            arcpy.AddMessage('Cannot determine face number for ' + ptname +
                                             ', expected a a single digit or F1. Used original file name.')
                            self.gkey = ptname

                    elif len(face_string) == 2:
                        face_prefix = tokens[1][0]
                        face_string = tokens[1][1:]
                        if face_prefix == 'f' and face_string.isdigit():  # 'F1' style
                            face_str = face_string
                        else:
                            arcpy.AddMessage('Cannot determine face number for ' + ptname +
                                             ', expected a a single digit or F1. Used original file name.')
                            self.gkey = ptname

                    else:  # face string too long: bail
                        arcpy.AddMessage('Cannot determine face number for ' + ptname +
                                             ', expected a a single digit or F1. Used original file name.')
                        self.gkey = ptname

                    if self.gkey == '':  # catalog and face found OK
                        version_string = tokens[2]  # possible Version as 'V1' or '1'
                        if len(version_string) == 0:
                            version_str = '1'

                        elif len(version_string) == 1 and version_string.isdigit():
                            version_str = version_string

                        elif len(version_string) == 2:
                            version_prefix = tokens[2][0]
                            version_string = tokens[2][1:]
                            if version_prefix == 'v' and version_string.isdigit():  # 'F1' style
                                version_str = version_string
                            else:
                                version_str = '1'

                        else:
                            arcpy.AddMessage('Cannot determine version number for ' + ptname +
                                             ', expected a a single digit or F1. Used original file name.')
                            self.gkey = ptname

                else:
                    if ptname[0] == 'G' and ptname[1].isdigit():
                        self.gkey = ptname[1:]
                        arcpy.AddMessage('Point name started with G, assumed in testing_loop and set gkey to ' +
                                         self.gkey)
                    else:
                        arcpy.AddMessage('Cannot determine catalog number for ' + ptname)
                        self.gkey = ptname

                if self.gkey == '':
                    self.gkey = catalog_str + '_' + face_str + '_' + version_str

            elif len(tokens) == 2:  # possibly in form pptp123_F1
                catalog_string = tokens[0]  # strip 'pptp' from first token
                if catalog_string.isdigit():
                    catalog_str = catalog_string
                else:
                    arcpy.AddMessage('Cannot determine catalog number for ' + ptname)
                    self.gkey = ptname

                face_string = tokens[1]  # possible Face as 'F1' or '1'

                if len(face_string) == 0:  # no face supplied, default to 1
                    face_str = '1'
                    version_str = '1'

                elif len(face_string) > 3:  # face string too long: bail
                    arcpy.AddMessage('Cannot determine face number for ' + ptname)
                    self.gkey = ptname

                else:  # determine face and version
                    if face_string.isdigit():  # just a digit
                        face_str = face_string
                    else:
                        face_prefix = tokens[1][0]
                        face_string = tokens[1][1:]
                        if face_prefix == 'f' and face_string.isdigit():  # 'F1' style
                            face_str = face_string
                        else:
                            face_str = '1'
                            # defaults to version = 1

                if self.gkey == '':
                    self.gkey = catalog_str + '_' + face_str + '_1'


            elif len(tokens) == 1:  # possibly in form pptp123
                catalog_string = tokens[0]  # strip 'pptp' from first token
                if catalog_string.isdigit():
                    catalog_str = catalog_string
                else:
                    arcpy.AddMessage('Cannot determine catalog number for ' + ptname)
                    self.gkey = ptname

                if self.gkey == '':
                    self.gkey = catalog_str + '_1_1'


            elif len(tokens) == 0:  #
                self.gkey = ptname

            else:  # more than three tokens (extra underscores)
                self.gkey = ptname

        else:  # non-pptp name
            # arcpy.AddMessage(' DEBUG:  Dealing with ptname = ' + ptname)
            if ptname[0] == 'G' and ptname[1].isdigit():  # G123_2_1 style name
                self.gkey = ptname[1:]
            else:  # non standard name
                self.gkey = ptname

        if len(self.gkey) > 9:
            arcpy.AddMessage('Suspect a point name parse problem on ' +  self.gkey)
        return True


    def load_edit_fc(self):
        ''' Copies persistent data from GDB_Edit to GDB.

        Copies Haft_Collar and any Scar feature classes fro Edit_GDB to GDB
        Haft_Collar is a feature class
        Scar* feature classes are contained in the Artifacts feature dataset (for topology edit reasons)
        '''

        if arcpy.Exists(self.edit_gdb):
            tmp_workspace = arcpy.env.workspace
            arcpy.env.workspace = self.edit_gdb

            # if Haft_Collar exists in current (final_gdb) and is missing from edit_gdb, copy it
            fc = 'Haft_Collar'
            if not arcpy.Exists(fc):
                if arcpy.Exists(self.final_gdb + '\\' + fc):
                    arcpy.CopyFeatures_management(self.final_gdb + '\\' + fc, fc)

            # copy any scar feature classes
            fcList = arcpy.ListFeatureClasses('Scar*', 'All', 'Artifact')

            for fc in fcList:
                arcpy.CopyFeatures_management(fc, self.final_gdb + '\\' + fc)
                arcpy.AddMessage('   Copied edited scar feature class ' + fc + ' to ' + tmp_workspace)
            arcpy.env.workspace = tmp_workspace
        return True


    def build_geometry(self, xyz_file):
        ''' Reads ASCII xyz scan file and creates basic geometry feature classes and rasters.

        - Creates geodatabases
        - Reads ASCII file into 3D points
        - Shifts points in XY to centroid of (0,0)
        - Shifts points in Z such that lowest value will be z = 0    (BasePts feature class)
        - Creates TIN and related feature classes
        - Creates re_entrants
        - Determines haft points
        - Determines various quadrant and ellipse key points
        - Creates Smoothed and Residual surface rasters
        '''

        # Set current workspace to work GDB for this Point
        arcpy.env.workspace = self.final_gdb

        if self.is_geometry_complete():             # looks for pre-existing fc and grids
            self.fetch_persistent_data()
            return True                             # already complete, skips rest of geometry rebuild

        # Create new geodatabase
        self.create_gdb(self.gdb_dir, self.pt_gdb)
        arcpy.env.workspace = self.final_gdb        # redundant set required

        # Load the ASCII point cloud and make OrigPts feature class
        if self.verbose:
            arcpy.AddMessage('     Loading XYZ points to create feature class OrigPts')

        # add check for nonsense input file
        input_full_fn = self.xyz_dir + '\\' + xyz_file
        if not self.check_input_file(input_full_fn):
            arcpy.AddMessage(xyz_file + ' in ' + self.xyz_dir + ' may be corrupt. Not processed.')
            return False

        arcpy.ASCII3DToFeatureClass_3d(input_full_fn, 'XYZ', 'OrigPts', 'POINT', '1', self.sr, '', '', 'DECIMAL_POINT')

        arcpy.CopyFeatures_management('OrigPts', 'in_memory\\OrigPts')

        # Compute centroid of original point outline and shift points to (0,0)
        self.create_outline('in_memory\\OrigPts', 'tmpTIN', 'in_memory\\tmpTIN_Tri', 'in_memory\\OrigOutline')
        arcpy.Delete_management('tmpTIN')
        self.shift_points_in_xy('in_memory\\OrigPts', 'PtsXYCentered', self.centroid_x, self.centroid_y)
        # PtsXYCentered target name should NOT include path
        arcpy.Delete_management('in_memory\\OrigPts')

        # Adjust points in Z so Min Z = 0

        self.rotate_points('in_memory\\PtsXYCentered')
        self.shift_points_in_z('in_memory\\Pts_Flattened', 'BasePts')

        # Create point outline feature class
        self.create_outline('BasePts', 'BaseTIN', 'in_memory\\TIN_Tri', 'Outline')
        self.create_TIN_edge('BaseTIN', 'TIN_edge')

        # Housekeeping
        # arcpy.Delete_management('BaseTIN')
        # 

        # Compute point outline properties
        self.compute_outline_properties('Outline')

        # Housekeeping
        arcpy.CopyFeatures_management('in_memory\\TIN_Tri', 'TIN_Tri')
        arcpy.Delete_management('in_memory\\TIN_Tri')


        # Create re_entrants
        self.create_re_entrants('Outline', 'Re_entrants')
        self.create_quadrant_points('Outline')
        self.create_reentrant_quadrant_points('Re_entrants')
        self.create_reentrant_ellipse_pts('Re_entrants')
        self.draw_belts()
        self.define_haft()


        #Create topo, smoothed and residual rasters from topoext and outline
        self.create_residual('TopoExt', 'Outline')

        arcpy.Delete_management('in_memory')  # DEBUG added after independent reentrant testing

        # geometry complete can check for residual presence to determine if previous run successful
        return True

    def check_input_file(self, filename):
        #  Sanity check on ASCII XYZ input file    #DEBUG yet to do
        return True

    def compute_landmarks(self, internal_rings, pct_outward_shift, blade_rows, blade_columns, blade_top_pct_setback,
                        blade_base_pct_setback, haft_rows, haft_columns, haft_top_pct_setback, haft_base_pct_setback):
        ''' Compute landmarks for Procrustes analysis.  Iterates through all GDBs in gdb_dir.

        - Creates radial edge and ring landmarks
        - Creates rectangular haft and blade landmarks
        - Creates Composite Procrustes_Interior and Procrustes_Exterior feature classes for stemmed points

        This routine is designed to be run after basic processing builds geometry
        '''

        arcpy.env.workspace = self.gdb_dir                      # Workspace at directory level...
        workspaces = arcpy.ListFiles('*.gdb')                   # list of files to process

        if len(workspaces) == 0:
            arcpy.AddMessage('No geodatabases found in : ' + arcpy.env.workspace)
            arcpy.GetMessages()
        arcpy.AddMessage('   ' + str(len(workspaces)) + ' artifacts to process\n')

        for workspace in workspaces:
            try:
                if (workspace != 'a_GLiMR_Summary.gdb'):

                    # reconstruct GLiMR name environment pt_name and GDBs for this point
                    root_pt_name = workspace.split('.')[0]      # expect std gdb names 'G123_1_1'
                    if root_pt_name.startswith('G'):
                        self.gkey = root_pt_name[1:]
                    else:
                        arcpy.AddMessage('\n   Oddly named geodatabase found: ' + workspace +
                                         '  ...Skipped.  Please examine your Geodatabase directory...\n' +
                                         '  workspace does not start with G.  Not processing ' + workspace)
                        continue

                    pt_gdb = workspace                          # example: G123_2_1.gdb
                    final_gdb = self.gdb_dir + '\\' + pt_gdb    # example: G123_2_1.gdb is final geodatabase with path
                    self.pt_name = root_pt_name                 # example: G123_2_1 is the current point name
                    self.pt_gdb = pt_gdb                        # example: G123_2_1.gdb is geodatabase
                    self.final_gdb = final_gdb                  # example: <path>\G123_2_1.GDB has full path
                    self.edit_gdb = self.gdb_edit_dir + '\\' + pt_gdb

                    if not self.rebuild_GLiMR_member_variables():
                        arcpy.AddMessage('Skipping landmarks for ' + self.final_gdb)
                        continue

                    arcpy.AddMessage('   Computing landmark point feature classes for ' + root_pt_name)
                    arcpy.env.workspace = self.final_gdb

                    # Compute radial edge and internal Procrustes landmarks
                    edgeFC = 'Outline'

                    arcpy.AddMessage('     Computing radial landmarks')
                    self.create_procrustes_pts_radial(internal_rings, edgeFC, pct_outward_shift)

                    # Compute rectangular Procrustes landmarks for haft and blade  (Stemmed points only right now)
                    ''' default values set in ArcGIS tool
                    internal_rings = 2
                    pct_outward_shift = 0.0

                    blade_rows = 5
                    blade_columns = 5
                    blade_top_pct_setback = 15.0
                    blade_base_pct_setback = 10.0

                    haft_rows = 4
                    haft_columns = 5
                    haft_top_pct_setback = 10.0
                    haft_base_pct_setback = 22.0
                    '''

                    arcpy.AddMessage('     Computing blade rectangular landmarks')
                    self.create_procrustes_rect('Blade', blade_rows, blade_columns,blade_top_pct_setback,
                                                blade_base_pct_setback, edgeFC)

                    arcpy.AddMessage('     Computing haft rectangular landmarks')
                    self.create_procrustes_rect('Haft', haft_rows, haft_columns,haft_top_pct_setback,
                                                haft_base_pct_setback, edgeFC)

                    # Build composite Procrustes_Exterior and Procrustes_Interior feature classes
                    arcpy.AddMessage('     Assembling Procrustes_Interior and Procrustes_Exterior landmarks\n')
                    self.assemble_procrustes_fc()

                    self.delete_preexisting('Blade_HzLines')        # housekeeping of Procrustes construction lines
                    self.delete_preexisting('Haft_HzLines')
                    self.delete_preexisting('Haft_VertLines')

            except arcpy.ExecuteError:
                print arcpy.GetMessages()

            except:
                # Get the traceback object
                tb = sys.exc_info()[2]
                tbinfo = traceback.format_tb(tb)[0]
                # Concatenate error information into message string
                pymsg = 'PYTHON ERRORS:\nTraceback info:\n{0}\nError Info:\n{1}' \
                    .format(tbinfo, str(sys.exc_info()[1]))
                msgs = 'ArcPy ERRORS:\n {0}\n'.format(arcpy.GetMessages(2))
                # Return python error messages for script tool or Python Window
                arcpy.AddError(pymsg)
                arcpy.AddError(msgs)



    def get_outline_extreme_pt(self, direction):
        ''' Query Outline_quad_pnts feature class for extreme vertices ("N, 'S', 'E', 'W').

        Returns [X, Y].   Do not call it prior to creating 'Outline_quad_pnts'
        '''


        if arcpy.Exists('Outline_quad_pnts'):
            with arcpy.da.SearchCursor('Outline_quad_pnts', ['Ref', 'POINT_X', 'POINT_Y']) as cursor:
                for row in cursor:
                    if row[0] == direction:
                        return [row[1], row[2]]
            del cursor
            del row
        else:
            arcpy.AddMessage('Error in get_outline_extreme_pt: Outline_quad_pnts '
                             'does not exist or odd direction requested')
            exit()
        return True

    def create_pc_pnt_fc(self, fcName):
        ''' Create Procrustes point feature class with appropriate fields. '''
        self.delete_preexisting(fcName)
        arcpy.CreateFeatureclass_management(self.final_gdb, fcName, 'POINT', has_z= 'ENABLED',
                                            spatial_reference= self.sr)
        arcpy.AddField_management(fcName, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(fcName, 'Type_', 'TEXT', '', '', 25)    # reentrant label Left, Right, Base, Procrustes
        arcpy.AddField_management(fcName, 'Ref', 'TEXT', '', '', 10)     # point reference name T, B, R, L
        arcpy.AddField_management(fcName, 'Label', 'TEXT', '', '', 50)
        arcpy.AddField_management(fcName, 'ProcrustesPtNum', 'LONG')
        return True


    def create_procrustes_rect(self, area_name, rows, columns, top_pct_setback, base_pct_setback, outlineFC):
        ''' Compute rectangular array of Procrustes points for blade or haft

        Creates rectangular points at the intersection of horizontal lines and 'Outline',
        then interpolates one or more columns between those edge points.

        Expected area_name is 'Blade' or 'Haft'
        Values for rows, columns, and setbacks specified separately for blade and haft
        top and base pct_setback is percent of height that the upper and lower bounds will be shrunk
        '''

        if self.verbose:
            arcpy.AddMessage('     Computing rectangular Procrustes ' + area_name + ' points for ' + self.pt_name)

        if self.haft_collar_E_pnt == (0,0):
            return False

        row_x_col = '_' + str(rows) + 'x' + str(columns + 2)                # restore columns count

        edgePtFC = 'Procrustes_' + area_name + '_Edge_Pts_' + row_x_col
        insidePtFC = 'Procrustes_' + area_name + '_Inside_Pts_' + row_x_col
        tipFC = 'Procrustes_tmp_Tip'                                        # debug put this in memory
        finalTipFC = 'Procrustes_Blade_Tip_Pnt'
        baseFC = 'Procrustes_Haft_Base_Pnt'

        columns -= 2                                                 # compensate for two end points (col = col - 2)

        self.delete_preexisting(edgePtFC)
        self.delete_preexisting(insidePtFC)
        self.delete_preexisting(tipFC)
        self.delete_preexisting(baseFC)

        if not arcpy.Exists(outlineFC):
            arcpy.AddMessage('Outline feature class not found in ' + self.final_gdb)
            exit()

        if not (arcpy.Exists('Topo') and arcpy.Exists('Smoothed7mm')):
            arcpy.AddMessage('Topo or Smoothed7mm feature classes not found in ' + self.final_gdb)
            exit()

        midhaft_y = (self.haft_collar_W_pnt[1] + self.haft_collar_E_pnt[1] ) /2.    # center point y of haft collar
        haft_top_y = midhaft_y
        blade_base_y = midhaft_y

        # create Procrustes_Blade_Tip_Pt fc with north quad point from outline
        if area_name == 'Blade':

            self.create_pc_pnt_fc(tipFC)

            with arcpy.da.InsertCursor(tipFC, ['SHAPE@XY', 'Type_', 'GKey', 'ProcrustesPtNum', 'Label']) as cursor:
                cursor.insertRow([self.get_outline_extreme_pt('N'), 'Tip', self.gkey, -2, 'Blade tip'])  # insert point at tip
            del cursor

            arcpy.AddXY_management(tipFC)
            arcpy.AlterField_management(tipFC, 'POINT_X', 'X', 'X')
            arcpy.AlterField_management(tipFC, 'POINT_Y', 'Y', 'Y')
            arcpy.AddSurfaceInformation_3d(tipFC, 'Topo', 'Z')
            arcpy.AlterField_management(tipFC, 'Z', 'Z_Topo', 'Z_Topo')
            arcpy.AddSurfaceInformation_3d(tipFC, 'Smoothed7mm', 'Z')
            arcpy.AlterField_management(tipFC, 'Z', 'Z_Smooth', 'Z_Smooth')
            arcpy.DeleteField_management(tipFC, 'POINT_Z')

            arcpy.CopyFeatures_management(tipFC, finalTipFC)       # tipFC exists here, but gone later
            arcpy.Delete_management(tipFC)

            # create horizontal line feature class  Blade_HzLines'
            blade_Hz = self.create_pc_blade_haft_hz_lines('Blade', blade_base_y, self.outline_max_y, rows, columns,
                                                           top_pct_setback, base_pct_setback, outlineFC)
            self.intersect_procrustes(blade_Hz, outlineFC, 'Blade_Pts', rows, columns, area_name)

        elif area_name == 'Haft':
            # create Procrustes_Haft_Base_Pt fc with south quad point from outline

            self.create_pc_pnt_fc(baseFC)
            with arcpy.da.InsertCursor(baseFC, ['SHAPE@XY', 'Type_', 'GKey', 'ProcrustesPtNum', 'Label']) as cursor:
                cursor.insertRow([self.get_outline_extreme_pt('S'), 'Base', self.gkey, -3,  'Haft base'])  # insert point at base
            del cursor

            arcpy.AddXY_management(baseFC)
            arcpy.AlterField_management(baseFC, 'POINT_X', 'X', 'X')
            arcpy.AlterField_management(baseFC, 'POINT_Y', 'Y', 'Y')

            arcpy.AddSurfaceInformation_3d(baseFC, 'Topo', 'Z')
            arcpy.AlterField_management(baseFC, 'Z', 'Z_Topo', 'Z_Topo')
            arcpy.AddSurfaceInformation_3d(baseFC, 'Smoothed7mm', 'Z')
            arcpy.AlterField_management(baseFC, 'Z', 'Z_Smooth', 'Z_Smooth')
            arcpy.DeleteField_management(baseFC, 'POINT_Z')

            # create horizontal line feature class  Haft_HzLines'
            haft_Hz = self.create_pc_blade_haft_hz_lines('Haft', self.outline_min_y, haft_top_y, rows, columns,
                                                         top_pct_setback, base_pct_setback, outlineFC)
            self.intersect_procrustes(haft_Hz, outlineFC, 'Haft_Pts', rows, columns, area_name)

            x_left = self.x_left
            x_right = self.x_right
            haft_Vert = self.create_pc_blade_haft_vert_lines('Haft', x_left, x_right, rows, columns,0., 0., outlineFC)
            self.intersect_procrustes(haft_Vert, outlineFC, 'Haft_Basal_Pts', rows, columns, area_name)

        else:
            arcpy.AddMessage( 'Bad area_name given in Procrustes: ' + area_name)
            arcpy.GetMessages()
        return True

    def intersect_procrustes(self, lines, outlineFC, pointFC, rows, columns, area_name):
        ''' Intersect haft or blade horizontal or vertical lines with outline fc.

        - Yields point feature class with filled attributes and Z values
        - Result name in the form of <'Procrustes_' + pointFC + row_x_col>   e.g Procrustes_Haft_4x5
        '''

        row_x_col = '_' + str(rows) + 'x' + str(columns + 2)        # restore columns count

        fid_name = 'FID_' + lines       # e.g 'FID_Blade_HzLines'
        tmpFCname = 'in_memory\\PCR_tmp'

        arcpy.Intersect_analysis([outlineFC, lines], 'in_memory\\MutliEdgePts', 'ONLY_FID', 0.002, 'POINT')
        arcpy.MultipartToSinglepart_management('in_memory\\MutliEdgePts', tmpFCname)
        arcpy.Delete_management('in_memory\\MutliEdgePts')

        arcpy.AddField_management(tmpFCname, 'Type_', 'TEXT', '', '', 20)
        arcpy.AddField_management(tmpFCname, 'ProcrustesPtNum', 'LONG')
        arcpy.AddField_management(tmpFCname, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(tmpFCname, 'Label', 'TEXT', '', '', 50)

        # now have edge point fc with attribute fields

        cursor = arcpy.da.UpdateCursor('in_memory\\PCR_tmp', ['Type_', 'GKey',
                                        'ProcrustesPtNum', 'OID@', 'Label', 'SHAPE@X'])
        for row in cursor:
            row[0] = area_name + '_Edge' + row_x_col
            row[1] = self.gkey
            row[2] = row[3]                  # TO DO  Procrustes point number   (actually filled later)
            row[4] = 'Edge_' + str(row[3])
            cursor.updateRow(row)
        del cursor

        ptEdgeFC = pointFC + '_Edge' + row_x_col
        arcpy.CopyFeatures_management(tmpFCname, ptEdgeFC)

        if lines != 'Haft_VertLines':            # skip internals point for basal haft points
            # compute internal points

            cursor = sorted(arcpy.da.SearchCursor(ptEdgeFC, ['SHAPE@Y', 'SHAPE@X', 'OID@']))     # walks through edge points

            # appends internal points to existing edge points   (relying on apparent sort of ptEdgeFC)
            cursor2 = arcpy.da.InsertCursor(tmpFCname, ['SHAPE@X', 'SHAPE@Y', 'Type_', 'GKey', 'ProcrustesPtNum', 'Label'])

            count = 0
            for row in cursor:
                count += 1

                if count % 2 == 1:
                    x0 = row[1]
                    y0 = row[0]
                else:
                    x1 = row[1]
                    y1 = row[0]

                    xMin = min(x0, x1)

                    if count == 2 and area_name == 'Haft':
                        x_left = x0
                        y_left = y0
                        x_right = x1
                        y_right = y1

                        self.x_left = x_left
                        self.x_right = x_right

                        # arcpy.AddMessage('x0, y0, x1, y1: ' + str(x0) + ', ' + str(y0) + ', ' + str(x1) + ', ' + str(y1))
                        # arcpy.GetMessages()

                    for i in range(columns):
                        col_num = i + 1

                        # now have two endpoints to interpolate between (x0,y0 and x1,y1)  y is constant
                        x_step = abs(x1 - x0) / (columns + 1)

                        x = xMin + (col_num * x_step)  # before outward shift

                        int_pt_num = (col_num * 10000) + count / 2                    # every other count is a newline
                        label = 'Int_' + str(col_num) + '_' + str(count/2)            # count/2 is row number
                        type_str = area_name + '_Internal' + row_x_col
                        cursor2.insertRow([x, y1, type_str, self.gkey, int_pt_num, label])


            del cursor
            del cursor2

        arcpy.AddXY_management(tmpFCname)
        arcpy.DeleteField_management(tmpFCname, ['FID_Outline', 'ORIG_FID', fid_name])
        arcpy.AlterField_management(tmpFCname, 'POINT_X', 'X', 'X')
        arcpy.AlterField_management(tmpFCname, 'POINT_Y', 'Y', 'Y')
        arcpy.AlterField_management(tmpFCname, 'POINT_Z', 'Z', 'Z')

        arcpy.AddSurfaceInformation_3d(tmpFCname, 'Topo', 'Z')
        arcpy.AlterField_management(tmpFCname, 'Z', 'Z_Topo', 'Z_Topo')

        arcpy.AddSurfaceInformation_3d(tmpFCname, 'Smoothed7mm', 'Z')
        arcpy.AlterField_management(tmpFCname,'Z', 'Z_Smooth', 'Z_Smooth')

        ptCombinedFC = 'Procrustes_' + pointFC + row_x_col
        arcpy.CopyFeatures_management(tmpFCname, ptCombinedFC)
        arcpy.Delete_management(ptEdgeFC)
        return True


    def create_pc_blade_haft_hz_lines(self, area_name, min_y, max_y, rows, columns, top_pct_setback, base_pct_setback, outlineFC):
        # area_name = 'Blade' or 'Haft'

        # find top and base y dimension for hz lines

        top = max_y - ((max_y - min_y ) * top_pct_setback / 100.)
        base = min_y + ((max_y - min_y ) * base_pct_setback / 100.)
        step = (top - base) / (rows - 1)

        # arcpy.AddMessage(self.gkey + ' ' +  str(max_y) + '   ' +  str(min_y)+ '   ' +  str(top)+ '  ' +  str(base) + '  setback ' + str(base_pct_setback))


        # Compute outline points

        # create vertex list for each horizontal line
        left_extreme = -100.
        right_extreme = 100.

        # polyline_vertex_list is a list of features and coordinate pairs for equally spaced hz lines
        polyline_vertex_list = []

        for i in range(rows):
            y = base + (i * step)
            polyline_vertex_list.append([[left_extreme, y], [right_extreme, y]])

        # features is a list that will hold each of the Polyline objects
        features = []

        for feature in polyline_vertex_list:
            # Create a Polyline object based on the array of points
            # Append to the list of Polyline objects
            features.append(
                arcpy.Polyline(
                    arcpy.Array([arcpy.Point(*coords) for coords in feature])))

        # Persist a copy of the Polyline objects using CopyFeatures
        linetarget = area_name + '_HzLines'
        arcpy.CopyFeatures_management(features, self.final_gdb + '\\' + linetarget)
        return linetarget

    def create_pc_blade_haft_vert_lines(self, area_name, min_x, max_x, rows, columns, left_pct_setback, right_pct_setback, outlineFC):
        # area_name = 'Haft' only

        columns = 3  # hard wired for now
        left_pct_setback = 0.
        right_pct_setback = 0.

        # find left and base x dimension for vertical lines

        xright = max_x - ((max_x - min_x ) * left_pct_setback / 100.)
        xleft = min_x + ((max_x - min_x ) * right_pct_setback / 100.)
        xstep = abs(xleft - xright) / float((columns + 1))

        # create vertex list for each vertical line
        top_extreme = 0.
        base_extreme = -100.

        # polyline_vertex_list is a list of features and coordinate pairs for equally spaced hz lines
        polyline_vertex_list = []

        for i in range(columns):
            x = xleft + ((i+1) * xstep)     # skip the endpoints and draw the internal verticals
            polyline_vertex_list.append([[x, top_extreme], [x, base_extreme]])

        # features is a list that will hold each of the Polyline objects
        features = []

        for feature in polyline_vertex_list:
            # Create a Polyline object based on the array of points
            # Append to the list of Polyline objects
            features.append(
                arcpy.Polyline(
                    arcpy.Array([arcpy.Point(*coords) for coords in feature])))

        # Persist a copy of the Polyline objects using CopyFeatures
        linetarget = area_name + '_VertLines'
        arcpy.CopyFeatures_management(features, self.final_gdb + '\\' + linetarget)
        return linetarget


    def create_procrustes_pts_radial(self, internal_rings, outlineFC, pct_outward_shift):
        # Create Procrustes points. Create points at the intersection of 'Starburst' polygons and 'Outline',
        # then interpolates one or more rings between the centroid and those points.

        if self.verbose:
            arcpy.AddMessage('     Computing radial Procrustes points for ' + self.pt_name)

        self.delete_preexisting('Procrustes_Radial_Edge_Pts')
        self.delete_preexisting('Procrustes_Rings_1')
        self.delete_preexisting('Procrustes_Rings_2')
        self.delete_preexisting('Procrustes_Blade_Tip')

        # internal_rings is the number of rings, not counting the points on the outline or the centroid
        # pct_outward_shift is percent of distance from point to edge, that the point gets shifted outward

        starburst_edge = 'Starburst_edge'  # fc to intersect with edge to use directly as edge points
        starburst_internal = 'Starburst_rings'  # fc to intersect with edge, then use for internal points
        fid_name = 'FID_' + starburst_edge

        if not arcpy.Exists(outlineFC):
            arcpy.AddMessage('Outline feature class not found in ' + self.final_gdb + '  ...Exiting.')
            exit()

        if not (arcpy.Exists('Topo') and arcpy.Exists('Smoothed7mm')):
            arcpy.AddMessage('Topo or Smoothed7mm feature classes not found in ' + self.final_gdb + '  ...Exiting.')
            exit()

        if not arcpy.Exists(starburst_edge):  # Copy Starburst radial line feature from Template GDB
            if arcpy.Exists(self.template_dir + '\\Template.gdb\\' + starburst_edge):
                arcpy.CopyFeatures_management(self.template_dir + '\\Template.gdb\\' + starburst_edge,
                                              self.final_gdb + '\\' + starburst_edge)
            else:
                arcpy.AddMessage('Feature class ' + starburst_edge + '(for Procrustes edges) not found in ' +
                       self.template_dir + '\\Template.gdb')
                arcpy.AddMessage('Error: Please insure ' + starburst_edge + ' and ' + starburst_internal +
                       ' radial line feature classes exists for Procrustes analysis')
                exit()

        if not arcpy.Exists(starburst_internal):  # Copy Starburst radial line feature from Template GDB
            if arcpy.Exists(self.template_dir + '\\Template.gdb\\' + starburst_internal):
                arcpy.CopyFeatures_management(self.template_dir + '\\Template.gdb\\' + starburst_internal,
                                              self.final_gdb + '\\' + starburst_internal)
            else:
                arcpy.AddMessage('Feature class ' + starburst_internal + '(for Procrustes edges) not found in ' +
                       self.template_dir + '\\Template.gdb')
                arcpy.AddMessage('Error: Please insure ' + starburst_internal + ' and ' + starburst_internal +
                       ' radial line feature classes exists for Procrustes analysis')
                exit()

        # Find intersection points between radial lines and outline
        # Note: intersect is picking up the Z value of the 3D perimeter polyline. Probably a good number.

        # Create pcr_Ext_FCPoint feature class

        arcpy.Intersect_analysis([outlineFC, starburst_edge], 'in_memory\\pcr_Ext_FCPts', 'ONLY_FID', 0.002, 'POINT')

        arcpy.MultipartToSinglepart_management('in_memory\\pcr_Ext_FCPts', 'in_memory\\PCR_tmp')
        arcpy.AddField_management('in_memory\\PCR_tmp', 'Type_', 'TEXT', '', '', 20)
        arcpy.AddField_management('in_memory\\PCR_tmp', 'ProcrustesPtNum', 'LONG')
        arcpy.AddField_management('in_memory\\PCR_tmp', 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management('in_memory\\PCR_tmp', 'Label', 'TEXT', '', '', 50)
        arcpy.AddXY_management('in_memory\\PCR_tmp')

        pc_radial_edges = []
        with  arcpy.da.SearchCursor('in_memory\\PCR_tmp', ['OID@', 'POINT_X', fid_name]) as cursor:
            for row in cursor:
                a = [row[0], abs(row[1]), row[2]]  # absolute value of X
                pc_radial_edges.append(a)
        del cursor
        del row

        OIDs = []  # Will be list of good point OIDs  (choosing closest point to origin along radial)
        prev_FID = -1
        for pt in sorted(pc_radial_edges, key=itemgetter(2, 1)):
            if pt[2] == prev_FID:  # nth point on same radial, so skip
                continue

            OIDs.append(pt[0])  # append OID
            prev_FID = pt[2]

        with arcpy.da.UpdateCursor('in_memory\\PCR_tmp', ['Type_', 'GKey', 'ProcrustesPtNum',
                                                          'OID@', 'Label']) as cursor:
            n = 0

            for row in cursor:
                row[4] = 'Radial_' + str(n+1)
                if row[3] in OIDs:
                    n = n + 1
                    row[0] = 'Radial_edge_pt'
                    row[1] = self.gkey
                    row[2] = n
                else:
                    row[2] = 0  # flagged to be deleted
                cursor.updateRow(row)
        del cursor
        del row

        arcpy.DeleteField_management('in_memory\\PCR_tmp', ['FID_Outline', 'ORIG_FID', fid_name])

        arcpy.MakeFeatureLayer_management('in_memory\\PCR_tmp', 'pc_lyr')
        arcpy.SelectLayerByAttribute_management('pc_lyr', 'NEW_SELECTION', "ProcrustesPtNum > 0")
        arcpy.CopyFeatures_management('pc_lyr', 'in_memory\\pc_final')

        arcpy.AlterField_management('in_memory\\pc_final', 'POINT_X', 'X', 'X')
        arcpy.AlterField_management('in_memory\\pc_final', 'POINT_Y', 'Y', 'Y')
        arcpy.AlterField_management('in_memory\\pc_final', 'POINT_Z', 'Z', 'Z')

        arcpy.AddSurfaceInformation_3d('in_memory\\pc_final', 'Topo', 'Z')
        arcpy.AlterField_management('in_memory\\pc_final', 'Z', 'Z_Topo', 'Z_Topo')

        arcpy.AddSurfaceInformation_3d('in_memory\\pc_final', 'Smoothed7mm', 'Z')
        arcpy.AlterField_management('in_memory\\pc_final', 'Z', 'Z_Smooth', 'Z_Smooth')

        arcpy.CopyFeatures_management('in_memory\\pc_final', 'Procrustes_Radial_Edge_Pts')

        # Create sparse pcr_Ext_FCPoint feature class for internal point interpolation

        arcpy.Intersect_analysis([outlineFC, starburst_internal], 'in_memory\\pcr_Ext_FCPtsSparse', 'ONLY_FID', '',
                                 'POINT')
        arcpy.MultipartToSinglepart_management('in_memory\\pcr_Ext_FCPtsSparse', 'in_memory\\PCR_int_tmp')
        arcpy.AddField_management('in_memory\\PCR_int_tmp', 'Type_', 'TEXT', '', '', 20)
        arcpy.AddField_management('in_memory\\PCR_int_tmp', 'ProcrustesPtNum', 'LONG')
        arcpy.AddField_management('in_memory\\PCR_int_tmp', 'GKey', 'TEXT', '', '', 25)
        arcpy.AddXY_management('in_memory\\PCR_int_tmp')

        fid_name = 'FID_' + starburst_internal
        pc_radial_edges = []
        with  arcpy.da.SearchCursor('in_memory\\PCR_int_tmp', ['OID@', 'POINT_X', fid_name]) as cursor:
            for row in cursor:
                a = [row[0], abs(row[1]), row[2]]  # absolute value of X
                pc_radial_edges.append(a)
        del cursor
        del row

        OIDs = []  # Will be list of good point OIDs
        prev_FID = -1
        for pt in sorted(pc_radial_edges, key=itemgetter(2, 1)):
            if pt[2] == prev_FID:  # nth point on same radial, so skip
                continue

            OIDs.append(pt[0])  # append OID
            prev_FID = pt[2]

        with arcpy.da.UpdateCursor('in_memory\\PCR_int_tmp', ['Type_', 'GKey', 'ProcrustesPtNum', 'OID@']) as cursor:
            n = 0
            for row in cursor:
                if row[3] in OIDs:
                    n = n + 1
                    row[0] = 'pcr_Ext_FC'
                    row[1] = self.gkey
                    row[2] = n
                else:
                    row[2] = 0  # flagged to be deleted
                cursor.updateRow(row)
        del cursor
        del row

        arcpy.DeleteField_management('in_memory\\PCR_int_tmp', ['FID_Outline', 'ORIG_FID', fid_name])

        arcpy.MakeFeatureLayer_management('in_memory\\PCR_int_tmp', 'pc_int_lyr')
        arcpy.SelectLayerByAttribute_management('pc_int_lyr', 'NEW_SELECTION', "ProcrustesPtNum > 0")
        arcpy.CopyFeatures_management('pc_int_lyr', 'in_memory\\pc_int_final')

        # Compute internal points

        ringFC = self.final_gdb + '\\Procrustes_Radial_Rings_' + str(internal_rings)
        self.delete_preexisting(ringFC)

        arcpy.CreateFeatureclass_management(self.final_gdb, 'Procrustes_Radial_Rings_' + str(internal_rings), 'POINT',  has_z= 'ENABLED', spatial_reference= self.sr)

        arcpy.AddField_management(ringFC, 'Type_', 'TEXT', '', '', 20)
        arcpy.AddField_management(ringFC, 'ProcrustesPtNum', 'LONG')
        arcpy.AddField_management(ringFC, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(ringFC, 'Label', 'TEXT', '', '', 50)

        # insert rings

        with arcpy.da.SearchCursor('in_memory\\pc_int_final', 'Shape@XY') as cursor:
            with  arcpy.da.InsertCursor(ringFC, ['SHAPE@X', 'SHAPE@Y', 'Type_', 'GKey', 'ProcrustesPtNum',
                                                 'Label']) as cursor2:
                cursor2.insertRow([0., 0., 'Centroid', self.gkey, 0,  'RadialCentroid'])  # insert point at centroid

                count = 0
                for row in cursor:
                    count += 1
                    x0 = row[0][0]
                    y0 = row[0][1]

                    for i in range(internal_rings):
                        ring_num = i + 1

                        # now have two endpoints to interpolate between (0,0 and x0,y0)
                        x_step = x0 / (internal_rings + 1)
                        y_step = y0 / (internal_rings + 1)

                        x = ring_num * x_step  # before outward shift
                        y = ring_num * y_step

                        xdist = x0 - x
                        ydist = y0 - y

                        x = x + (xdist * pct_outward_shift / 100.)  # after outward shift
                        y = y + (ydist * pct_outward_shift / 100.)

                        int_pt_num = (ring_num * 10000) + count

                        type_name = 'RadialInternal_' + str(internal_rings)
                        cursor2.insertRow([x, y, type_name, self.gkey, int_pt_num, 'Radial'])
        del cursor
        del cursor2
        del row

        arcpy.AddXY_management(ringFC)
        arcpy.AddSurfaceInformation_3d(ringFC, 'Topo', 'Z')
        arcpy.AlterField_management(ringFC, 'Z', 'Z_Topo', 'Z_Topo')
        arcpy.AddSurfaceInformation_3d(ringFC, 'Smoothed7mm', 'Z')
        arcpy.AlterField_management(ringFC, 'Z', 'Z_Smooth', 'Z_Smooth')

        arcpy.AlterField_management(ringFC, 'POINT_X', 'X', 'X')
        arcpy.AlterField_management(ringFC, 'POINT_Y', 'Y', 'Y')

        if self.verbose:
            arcpy.AddMessage('     Finished Procrustes for ' + self.pt_name)
        return True


    def create_reentrant_ellipse_pts(self, inputRE_FC):
        # Create ellipse pts for Base, Left and Right reentrants
        # inputRE_FC is the Re_entrant feature class

        if not arcpy.Exists(inputRE_FC):
            arcpy.AddMessage(inputRE_FC + ' feature class not found in ' + self.final_gdb)
            exit()

        if self.verbose:
            arcpy.AddMessage('     Processing Ellipses for ' + inputRE_FC)

        # draw reentrant ellipse polyline using TableToEllipse
        ellipseFC = 'RE_Ellipse'
        self.delete_preexisting(ellipseFC)
        arcpy.TableToEllipse_management(inputRE_FC, ellipseFC, 'CENTROID_X', 'CENTROID_Y', 'MAJORAXIS',
                                        'MINORAXIS', '', 'ORIENTATION', 'DEGREES', 'Type_', self.sr)
        # draw ellipse using TableToEllipse
        # add left, right and base reentrants into new RE_EllipseFC
        # add X_top, Y_top, X_right, Y_right, X_base, Y_base, X_left, Y_left fields for each ellipse
        # populate  X_top, Y_top, X_right, Y_right, X_base, Y_base, X_left, Y_left points for each ellipse

        # inputRE_FC is input feature class without path

        arcpy.env.workspace = self.final_gdb

        outFC = ellipseFC + '_pnts'

        if arcpy.Exists(outFC):
            arcpy.Delete_management(outFC)

        arcpy.CreateFeatureclass_management(self.final_gdb, outFC, 'POINT',  has_z= 'ENABLED', spatial_reference= self.sr)
        nameFID = ellipseFC + '_FID'

        if not self.is_field_present(outFC, nameFID):
            # arcpy.AddField_management(outFC, nameFID, 'LONG')
            arcpy.AddField_management(outFC, 'GKey', 'TEXT', '', '', 25)
            arcpy.AddField_management(outFC, 'Type_', 'TEXT', '', '', 15)    # reentrant label Left, Right, Base
            arcpy.AddField_management(outFC, 'Ref', 'TEXT', '', '', 10)     # point reference name T, B, R, L
            arcpy.AddField_management(outFC, 'Label', 'TEXT', '', '', 50)

        if not self.is_field_present(ellipseFC, 'MinorTop_X'):
            arcpy.AddField_management(ellipseFC, 'GKey', 'TEXT', '', '', 25)
            arcpy.AddField_management(ellipseFC, 'MinorTop_X', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MinorTop_Y', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MajorTop_X', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MajorTop_Y', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MinorBase_X', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MinorBase_Y', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MajorBase_X', 'DOUBLE')
            arcpy.AddField_management(ellipseFC, 'MajorBase_Y', 'DOUBLE')

        edit = arcpy.da.Editor(self.final_gdb)
        edit.startEditing(False, False)

        fields = ['OID@', 'CENTROID_X', 'CENTROID_Y', 'MAJORAXIS', 'MINORAXIS', 'ORIENTATION', 'MinorTop_X',
                  'MinorTop_Y', 'MajorTop_X', 'MajorTop_Y', 'MinorBase_X', 'MinorBase_Y', 'MajorBase_X',
                  'MajorBase_Y', 'Type_', 'GKey']

        with arcpy.da.UpdateCursor(ellipseFC, fields) as cursor:
            for row in cursor:
                re_type = row[14]

                majorR = (row[3] / 2.)  # MAJORAXIS /2
                minorR = (row[4] / 2.)  # MINORAXIS /2

                majordX = math.sin(math.radians(row[5])) * majorR  # Orientation is a bearing
                majordY = math.cos(math.radians(row[5])) * majorR
                minordX = math.cos(math.radians(row[5])) * minorR  # Minor axis 90 deg to major hence trig swap
                minordY = math.sin(math.radians(row[5])) * minorR

                x0 = row[1]  # centroids
                y0 = row[2]

                re_majorTop = [x0 + majordX, y0 + majordY]
                re_majorBase = [x0 - majordX, y0 - majordY]
                re_minorTop = [x0 + minordX, y0 - minordY]
                re_minorBase = [x0 - minordX, y0 + minordY]

                row[6] = re_minorTop[0]
                row[7] = re_minorTop[1]
                row[8] = re_majorTop[0]
                row[9] = re_majorTop[1]
                row[10] = re_minorBase[0]                   # transpose left labels in this section, not in assemble
                row[11] = re_minorBase[1]
                row[12] = re_majorBase[0]
                row[13] = re_majorBase[1]
                row[15] = self.gkey

                cursor.updateRow(row)

                with arcpy.da.InsertCursor(outFC, ['SHAPE@XY', 'Ref', 'Type_', 'GKey', 'Label']) as in_cursor:
                    in_cursor.insertRow([re_majorTop, 'MajorTop', re_type, self.gkey, 'Ellipse'])
                    in_cursor.insertRow([re_majorBase, 'MajorBase', re_type, self.gkey, 'Ellipse'])
                    in_cursor.insertRow([re_minorTop, 'MinorTop', re_type, self.gkey, 'Ellipse'])
                    in_cursor.insertRow([re_minorBase, 'MinorBase', re_type, self.gkey, 'Ellipse'])

            edit.stopEditing(True)
        del cursor
        del in_cursor
        del row

        arcpy.AddXY_management(outFC)
        return True


    def assemble_procrustes_fc(self):

        if self.verbose:
            arcpy.AddMessage('     Assembling Procrustes Exterior feature class for ' + inputRE_FC)

        arcpy.env.workspace = self.final_gdb
        radial_edge_fc = 'Procrustes_Radial_Edge_Pts'
        rect_blade_fc = 'Procrustes_Blade_Pts_5x5'
        rect_haft_fc = 'Procrustes_Haft_Pts_4x5'

        # check that required parts are present

        if not (arcpy.Exists(radial_edge_fc) and arcpy.Exists(rect_blade_fc) and arcpy.Exists(rect_haft_fc)):
            arcpy.AddMessage('    Procrustes Interior and Exterior assembly aborted due to undefined haft')
            return False

        pcr_Ext_FC = 'Procrustes_Exterior'

        # Create Procrustes External Points FC with z = 0

        self.create_pc_pnt_fc(pcr_Ext_FC)
        arcpy.AddField_management(pcr_Ext_FC, 'X', 'DOUBLE')
        arcpy.AddField_management(pcr_Ext_FC, 'Y', 'DOUBLE')
        arcpy.AddField_management(pcr_Ext_FC, 'Z', 'DOUBLE')

        pcrNum = 1
        in_cursor = arcpy.da.InsertCursor(pcr_Ext_FC, ['SHAPE@XY', 'GKey', 'Type_', 'Ref',
                                                       'Label', 'X', 'Y', 'Z', 'ProcrustesPtNum'])

        # left and right haft points from Haft_Collar fc
        z = 0
        fieldlistCollar = ['GKey', 'Type_',  'X_left', 'Y_left', 'X_right', 'Y_right']
        collar_cursor  = arcpy.da.SearchCursor('Haft_Collar', fieldlistCollar)

        for row in collar_cursor:
            rgkey = row[0]
            rtype = row[1]
            rref = 'Left'
            rlabel = 'Left haft collar'
            x = row[2]
            y = row[3]
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z, pcrNum])

            pcrNum += 1
            rref = 'Right'
            rlabel = 'Right haft collar'
            x = row[4]
            y = row[5]
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z, pcrNum])
            
        del fieldlistCollar
        del collar_cursor

        '''
        # blade tip point
        fieldlistQuad = ['GKey', 'Type_', 'Ref', 'Point_X', 'Point_Y', 'Point_Z']

        whereclause = """Ref = 'N' OR Ref = 'X_Mid' OR Ref = 'Y_Mid' """

        pnts_Outline_quad_cursor  = arcpy.da.SearchCursor('Outline_quad_pnts', fieldlistQuad, whereclause)
        for row in pnts_Outline_quad_cursor:
            rgkey = row[0]
            rtype = row[1]
            rref = row[2]
            rlabel = rtype +  ' ' + rref + ' point'
            x = row[3]
            y = row[4]
            z = row[5]

            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z])

        del pnts_Outline_quad_cursor
        '''
        # reentrant quadrant points
        begin_pcr = pcrNum + 1
        fieldlistQuad = ['GKey', 'Type_', 'Ref', 'Point_X', 'Point_Y', 'Point_Z']
        whereclause = """Type_ = 'Right'"""


        pnts_Ellipse_cursor  = arcpy.da.SearchCursor('RE_Ellipse_pnts', fieldlistQuad, whereclause)
        for row in pnts_Ellipse_cursor:
            rgkey = row[0]
            rtype = row[1]
            rref = row[2]
            rlabel = rtype + ' re-entrant ellipse point: ' + rref
            x = row[3]
            y = row[4]
            z = row[5]
            if rref == 'MajorTop':
                pcrNumRight = begin_pcr     # DEBUG:  double check labeling sequence
            elif rref == 'MinorBase':
                pcrNumRight = begin_pcr + 1
            elif rref == 'MinorTop':
                pcrNumRight = begin_pcr + 2
            elif rref == 'MajorBase':
                pcrNumRight = begin_pcr + 3

            pcrNum += 1
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z, pcrNumRight])
        del pnts_Ellipse_cursor

        begin_pcr = pcrNum + 1
        whereclause = """Type_ = 'Left'"""

        # re_entrant ellipse points
        pnts_Ellipse_cursor  = arcpy.da.SearchCursor('RE_Ellipse_pnts', fieldlistQuad, whereclause)
        for row in pnts_Ellipse_cursor:
            rgkey = row[0]
            rtype = row[1]
            rref = row[2]

            if rref == 'MajorTop':      # swap ellipse for Left side  (stemmed point only)
                rref = 'MajorBase'
            elif rref == 'MajorBase':
                rref = 'MajorTop'

            if rref == 'MinorTop':
                rref = 'MinorBase'
            elif rref == 'MinorBase':
                rref = 'MinorTop'

            rlabel = rtype + ' re-entrant ellipse point: ' + rref
            x = row[3]
            y = row[4]
            z = row[5]

            if rref == 'MajorTop':
                pcrNumLeft = begin_pcr
            elif rref == 'MinorBase':
                pcrNumLeft = begin_pcr + 1
            elif rref == 'MinorTop':
                pcrNumLeft = begin_pcr + 2
            elif rref == 'MajorBase':
                pcrNumLeft = begin_pcr + 3

            pcrNum += 1
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z, pcrNumLeft])
        del pnts_Ellipse_cursor

        # reentrant centroid points
        fieldlistOutline = ['GKey', 'CENTROID_X', 'CENTROID_Y']

        pnts_Outline_cursor  = arcpy.da.SearchCursor('Outline', fieldlistOutline)
        for row in pnts_Outline_cursor:
            rgkey = row[0]
            rtype = 'Outline'
            rref = 'Centroid'
            rlabel = rtype + '  ' + rref
            x = row[1]
            y = row[2]
            z = 0.   # solve for this
            pcrNum += 1
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z, pcrNum])
        del pnts_Outline_cursor

        # radial edge points points
        fieldlistRadialEdge = ['GKey', 'ProcrustesPtNum', 'Type_', 'Label', 'X', 'Y', 'Z_Topo', 'Z_Smooth']

        pnts_RadialEdge_cursor  = sorted(arcpy.da.SearchCursor(radial_edge_fc, fieldlistRadialEdge))

        for row in pnts_RadialEdge_cursor:
            rgkey = row[0]
            rtype = row[2]
            rref = str(row[1])
            rlabel = rtype + ' point number: ' + str(row[1])
            x = row[4]
            y = row[5]
            z = 0.      # row[6] is actual z
            pcrNum += 1
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z, pcrNum])
        del pnts_RadialEdge_cursor
        del in_cursor
        del pcr_Ext_FC

        # Create Procrustes Internal Points FC with z values
        if self.verbose:
            arcpy.AddMessage('     Assembling Procrustes Interior feature class for ' + inputRE_FC)

        pcr_Int_FC = 'Procrustes_Interior'

        pcrNum = 0

        # Create Procrustes Interior Points FC with z = 0')
        self.create_pc_pnt_fc(pcr_Int_FC)
        arcpy.AddField_management(pcr_Int_FC, 'X', 'DOUBLE')
        arcpy.AddField_management(pcr_Int_FC, 'Y', 'DOUBLE')
        arcpy.AddField_management(pcr_Int_FC, 'Z_Topo', 'DOUBLE')
        arcpy.AddField_management(pcr_Int_FC, 'Z_Smooth', 'DOUBLE')

        in_cursor = arcpy.da.InsertCursor(pcr_Int_FC, ['SHAPE@XY', 'GKey', 'Type_', 'Ref',
                                                       'Label', 'X', 'Y', 'Z_Topo', 'Z_Smooth', 'ProcrustesPtNum'])

        # radial edge points points
        fieldlistRadialEdge = ['GKey', 'ProcrustesPtNum', 'Type_', 'Label', 'X', 'Y', 'Z_Topo', 'Z_Smooth']

        pnts_RadialEdge_cursor  = sorted(arcpy.da.SearchCursor(radial_edge_fc, fieldlistRadialEdge))
        for row in pnts_RadialEdge_cursor:
            rgkey = row[0]
            rtype = row[2]
            rref = str(row[1])
            rlabel = row[3]
            x = row[4]
            y = row[5]
            z_topo = row[6]
            z_smooth = row[7]

            pcrNum += 1
            if pcrNum != row[1]:
                arcpy.AddMessage('PC Number problem in assemble radial points Interior')
                arcpy.GetMessages()
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z_topo, z_smooth, row[1]])
        del pnts_RadialEdge_cursor

        # Blade internal points
        fieldlistBlade = ['GKey', 'ProcrustesPtNum', 'Type_', 'Label', 'X', 'Y', 'Z_Topo', 'Z_Smooth']
        whereclause = """Type_ like 'Blade_Internal%'"""                     # file gdb wants % not *

        pnts_Blade_cursor  = sorted(arcpy.da.SearchCursor(rect_blade_fc, fieldlistBlade, whereclause))
        for row in pnts_Blade_cursor:
            rgkey = row[0]
            rtype = row[2]
            rref = str(row[1])
            rlabel = 'Blade_' + row[3]
            x = row[4]
            y = row[5]
            z_topo = row[6]
            z_smooth = row[7]

            pcrNum += 1
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z_topo, z_smooth, pcrNum])
        del pnts_Blade_cursor

        # Haft internal points
        pnts_Haft_cursor = ['GKey', 'ProcrustesPtNum', 'Type_', 'Label', 'X', 'Y', 'Z_Topo', 'Z_Smooth']
        whereclause = """Type_ like 'Haft_Internal%'"""

        pnts_Haft_cursor  = sorted(arcpy.da.SearchCursor(rect_haft_fc, pnts_Haft_cursor, whereclause))
        for row in pnts_Haft_cursor:
            rgkey = row[0]
            rtype = row[2]
            rref = str(row[1])
            rlabel = 'Haft_' + row[3]
            x = row[4]
            y = row[5]
            z_topo = row[6]
            z_smooth = row[7]

            pcrNum += 1
            in_cursor.insertRow([[x, y], rgkey,  rtype, rref, rlabel, x, y, z_topo, z_smooth, pcrNum])
        del pnts_Haft_cursor

        del in_cursor
        return True

    def ellipse_seq(self, majorTop, majorBase, minorTop, minorBase, centroid_X, centroid_Y):
        # find major axis end with largest Y value and number points clockwise
        for pt in [majorTop, majorBase, minorTop, minorBase]:
            maxY = 0.
            maxPt = [0., 0.]
            startPt = 'Undefined'       # DEBUG UNFINISHED

            if majorTop[1] > maxY:
                maxY = majorTop[1]
                maxPt = majorTop
                startPt = 'MajorTop'
            if majorBase[1] > maxY:
                maxy = majorBase[1]
                maxPt = majorBase
                startPt = 'MajorBase'

            # starting at startPt, sort clockwise
            dirRadians = math.atan2(maxPt[0] - centroid_X, maxPt[1] - centroid_Y)
            if self.verbose:
                arcpy.AddMessage('     ' + repr(startPt), repr(dirRadians))       # untested
            exit()

    def define_haft(self):
        if arcpy.Exists('in_memory\\tmp_Haft_Collar'):
            arcpy.FeatureToLine_management('Outline', 'in_memory\\Perimeter')
            arcpy.Merge_management(['in_memory\\Perimeter', 'in_memory\\tmp_Haft_Collar'],
                                   'in_memory\\Blade_Haft_lines')
            arcpy.FeatureToPolygon_management('in_memory\\Blade_Haft_lines', 'in_memory\\Blade_Haft', 0.005)
            arcpy.AddField_management('in_memory\\Blade_Haft', 'Part', 'TEXT', '', '', 20)
            arcpy.AddGeometryAttributes_management('in_memory\\Blade_Haft', 'CENTROID')

            with arcpy.da.UpdateCursor('in_memory\\Blade_Haft', ['Part', 'SHAPE@AREA', 'CENTROID_Y']) as cursor:
                min_centroid_y = 1000000.
                for row in cursor:
                    if row[1] >= 2.:  # eliminate the odd tiny polygon at the ends of the haft collar
                        row[0] = 'Blade'
                        if row[2] < min_centroid_y: min_centroid_y = row[2]
                        cursor.updateRow(row)
            del cursor
            del row

            with arcpy.da.UpdateCursor('in_memory\\Blade_Haft', ['Part', 'SHAPE@AREA', 'CENTROID_Y']) as cursor:
                for row in cursor:
                    if min_centroid_y == row[2]: row[0] = 'Haft'
                    cursor.updateRow(row)
            del cursor
            del row

            arcpy.CopyFeatures_management('in_memory\\Blade_Haft', 'Blade_Haft')
            arcpy.DeleteFeatures_management('in_memory\\Blade_Haft')
            arcpy.DeleteFeatures_management('in_memory\\Blade_Haft_lines')
            arcpy.DeleteFeatures_management('in_memory\\Perimeter')
            # arcpy.DeleteFeatures_management('in_memory\\tmp_Haft_Collar')
        return True


    def pre_check_scar_fc(self, inFlakePolyFC):
        '''Check to be sure Id field is present in flake scar boundary polygon feature class'''

        arcpy.env.workspace = self.final_gdb

        fields = arcpy.ListFields(inFlakePolyFC)  # add and populate Id field if missing
        if not ('Id' in fields):
            arcpy.AddField_management(inFlakePolyFC, 'Id', 'LONG')
        if not ('GKey' in fields):
            arcpy.AddField_management(inFlakePolyFC, 'GKey', 'TEXT', '', '', 25)
        if not ('Version' in fields):
            arcpy.AddField_management(inFlakePolyFC, 'Version', 'TEXT', '', '', 10)
        if not ('ScarSource' in fields):
            arcpy.AddField_management(inFlakePolyFC, 'ScarSource', 'TEXT', '', '', 50)

        with arcpy.da.UpdateCursor(inFlakePolyFC, ['OBJECTID', 'Id', 'GKey', 'Version', 'ScarSource']) as cursor:
            for row in cursor:
                row[1] = row[0]
                row[2] = self.gkey
                row[3] = self.version
                row[4] = inFlakePolyFC
                cursor.updateRow(row)
        del cursor
        del row
        return True

    def run_flake_analysis_all_GDB(self):       # DEBUG: not called
        '''Compute flake aggregate properties
        Process every flake feature class in GDB directory'''

        arcpy.env.workspace = self.gdb_dir

        workspaces = arcpy.ListWorkspaces('*', 'FileGDB')

        if len(workspaces) == 0:
            # nothing to do
            arcpy.AddMessage('Nothing to do. No GDB directories found in ' + arcpy.env.workspace)
            return False

        arcpy.AddMessage(str(len(workspaces)) + ' geodatabases to process for flake scar properties')

        # Process file GDBs in GDB_Dir

        for workspace in workspaces:
            if (workspace != self.gdb_dir + '\\' + 'a_GLiMR_Summary.gdb'):
                self.final_gdb = workspace
                # arcpy.AddMessage('Current workspace: ' + workspace + '\n')

                if not arcpy.Exists(self.final_gdb + '\\' + 'TIN_Tri'):
                    arcpy.AddMessage('self.final_gdb = ' + self.final_gdb)
                    arcpy.AddMessage('TIN_Tri does not exist...   Exiting from runFlakeAnalysis...')
                    arcpy.GetMessages()
                    exit()

                if self.visited(self.final_gdb + '\\' + 'TIN_Tri', 'FID_TIN_Tri_1'):
                    arcpy.AddMessage('Been here before...  Reprocessing already processed attributes...   '
                                     'Exiting from runFlakeAnalysis...')
                    exit()

                self.process_flakes()  # called when reprocessing all flakes in all geodatabases
        return True


    def process_flakes(self):
        '''Process all Basin* features in one geodatabase'''

        arcpy.env.workspace = self.final_gdb

        if not self.visited('TIN_Tri', 'Z_Min'):
            # Compute basic Z attributes for TIN_Tri
            arcpy.AddZInformation_3d('TIN_Tri', 'Z_Min; Z_Max; Z_Mean', 'NO_FILTER')
            # Z values calculated before Identity, due to non-triangles after on flake edges
            # Z Min will be of orig triangle before split at flake edge
            

        arcpy.CopyFeatures_management('TIN_Tri', 'in_memory\\TIN_Tri')
        

        # Loop through each flake scar 'ScarPoly' polygon feature class in list

        scar_polygons = arcpy.ListFeatureClasses('Scar*')
        domain_polygons = arcpy.ListFeatureClasses('SubDomain*')
        poly_list = scar_polygons + domain_polygons

        for inFlakePolyFC in poly_list:
            if not arcpy.Exists(inFlakePolyFC):
                arcpy.AddMessage(inFlakePolyFC + ' feature class does not exist and was skipped '
                                                 'for Flake Scar processing.')
                continue

            if self.verbose:
                arcpy.AddMessage('     Analyzing polygon aggregate data for ' + inFlakePolyFC)

            self.pre_check_scar_fc(inFlakePolyFC)

            # Create layer file to only process cells Shape_Area >= 0.005 (two cells or greater (avoids nulls)
            inFlakePolyLyr = inFlakePolyFC + '_lyr'
            arcpy.MakeFeatureLayer_management(inFlakePolyFC, inFlakePolyLyr, 'Shape_Area >= 0.005')

            memoryTIN = 'in_memory\\TIN_TriID'
            memoryBasin = 'in_memory\\' + inFlakePolyLyr

            arcpy.CopyFeatures_management(inFlakePolyLyr, memoryBasin)
            

            if self.verbose:
                arcpy.AddMessage('     Creating aggregate attribute fields')

            # Add basin FID to each triangle. Results:  FID_TIN_Tri = OID triangle  FID_Basin = OID basin
            inFeatureClass = 'in_memory\\TIN_Tri'
            idFeatureClass = memoryBasin
            outFeatureClass = memoryTIN
            arcpy.Identity_analysis(inFeatureClass, idFeatureClass, outFeatureClass, 'ONLY_FID')
            

            # Add fields Aspect_wx, Aspect_wy, Z_wgt, Slope_wgt to TIN_TriID
            inFeatureClass = memoryTIN
            arcpy.AddField_management(inFeatureClass, 'Aspect_wx', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Aspect_wy', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Z_wgt', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Slope_wgt', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'SArea', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'A_Slope_Sqrd', 'DOUBLE')

            arcpy.CopyFeatures_management(memoryTIN, 'TIN_TriID_Mid_' + inFlakePolyFC)


            # Compute Aspect_wx, Aspect_wy, Z_wgt, Slope_wgt in TIN_TriID
            if self.verbose:
                arcpy.AddMessage('     Compute slope and aspect statistics')

            fields = ('Aspect', 'Slope_Deg', 'Z_Mean', 'Aspect_wx', 'Aspect_wy', 'Slope_wgt', 'Z_wgt',
                      'SHAPE@AREA', 'A_Slope_Sqrd')

            with arcpy.da.UpdateCursor(inFeatureClass, fields) as cursor:
                for row in cursor:
                    aspect = row[0]
                    slope = row[1]
                    zmean = row[2]
                    area = row[7]

                    # sin and cos swapped to compensate for +X vs +Y origins
                    row[3] = area * math.sin(aspect * math.pi / 180.0)
                    row[4] = area * math.cos(aspect * math.pi / 180.0)
                    row[5] = area * slope
                    row[6] = area * zmean
                    row[8] = area * slope * slope  # area * slope squared

                    cursor.updateRow(row)
            del cursor
            del row

            # see if you can merge
            with arcpy.da.UpdateCursor(inFeatureClass, ['SArea', 'Slope_Deg', 'SHAPE@AREA']) as cursor:
                for row in cursor:
                    row[0] = row[2] / math.cos(math.radians(row[1]))  # SurfaceArea=ShapeArea/Cos(slope)
                    cursor.updateRow(row)
            del cursor
            del row

            arcpy.CopyFeatures_management(memoryTIN, 'TIN_TriID_' + inFlakePolyFC)

            # Create centroid point features
            if self.verbose:
                arcpy.AddMessage('     Compute centroid attributes')

            inFeatureClass = memoryBasin
            outFeatureClass = 'in_memory\\BasinCentroids'

            arcpy.FeatureToPoint_management(inFeatureClass, outFeatureClass, 'CENTROID')

            # Add x,y coordinate to BasinCentroids attributes
            inFeatureClass = 'in_memory\\BasinCentroids'
            arcpy.AddXY_management(inFeatureClass)
            

            # Make perimeter line from Outline
            inFeatureClass = 'Outline'
            arcpy.FeatureToLine_management(inFeatureClass, 'in_memory\\OutlineLine')

            # Find nearest edge from each TIN_Tri
            inFeatureClass = 'in_memory\\TIN_Tri'
            nearFeatureClass = 'in_memory\\OutlineLine'
            arcpy.Near_analysis(inFeatureClass, nearFeatureClass, '', 'LOCATION', 'ANGLE')
            

            # Convert angle to azimuth                               #Move?######
            # inFeatureClass = 'in_memory\\TIN_Tri'
            with arcpy.da.UpdateCursor(inFeatureClass, ['NEAR_ANGLE']) as cursor:
                for row in cursor:
                    row[0] = (450 - row[0]) % 360.  # Convert theta (0 = +X)ccw to bearing (0 = +Y) cw
                    cursor.updateRow(row)
            del cursor
            del row

            arcpy.AddGeometryAttributes_management(inFeatureClass, 'CENTROID')
            arcpy.DeleteField_management(inFeatureClass, 'CENTROID_Z')
            arcpy.CopyFeatures_management(inFeatureClass, 'TIN_Tri_Near')


            # Find nearest edge from centroid for flakes  (basins)
            inFeatureClass = 'in_memory\\BasinCentroids'
            nearFeatureClass = 'in_memory\\OutlineLine'
            arcpy.Near_analysis(inFeatureClass, nearFeatureClass, '', 'LOCATION', 'ANGLE')
            

            # Convert angle to azimuth                               #Move?######
            inFeatureClass = 'in_memory\\BasinCentroids'
            with arcpy.da.UpdateCursor(inFeatureClass, ['NEAR_ANGLE']) as cursor:
                for row in cursor:
                    row[0] = (450 - row[0]) % 360.  # Convert theta (0 = +X)ccw to bearing (0 = +Y) cw
                    cursor.updateRow(row)
            del cursor
            del row

            # Append nearest edge data to basin attribute table

            if self.verbose:
                arcpy.AddMessage('     Compute edge proximity data')

            inFeatureClass = memoryBasin
            joinField = 'OBJECTID'
            joinTable = 'in_memory\\BasinCentroids'
            joinField2 = 'ORIG_FID'
            fieldList = ['POINT_X', 'POINT_Y', 'NEAR_DIST', 'NEAR_ANGLE', 'NEAR_X', 'NEAR_Y']

            arcpy.JoinField_management(inFeatureClass, joinField, joinTable, joinField2, fieldList)
            

            # Compute statistics aggregated by basin   !! TIN_Tri_After has to be physical table (not in_memory) !!
            if self.verbose:
                arcpy.AddMessage('     Aggregate statistics by polygon')

            flakeFIDfield = 'FID_' + inFlakePolyFC + '_lyr'
            arcpy.Statistics_analysis('TIN_TriID_' + inFlakePolyFC, 'Aggregate',
                                      [['Slope_wgt', 'SUM'], ['Aspect_wx', 'MEAN'], ['Aspect_wx', 'STD'],
                                       ['Aspect_wy', 'MEAN'], ['Aspect_wy', 'STD'], ['Z_Min', 'MIN'], ['Z_Max', 'MAX'],
                                       ['Z_wgt', 'SUM'], ['SArea', 'SUM'], ['Shape_Area', 'SUM'],
                                       ['A_Slope_Sqrd', 'SUM']], flakeFIDfield)

            # Minor worry: values for area and centroid are different in the 3rd decimal from those computed by Zonal Geometry by table.

            # Add MEAN_Z, MEAN_Aspect, STD_Aspect fields to Aggregate
            # inFeatureClass = 'Aggregate'
            inFeatureClass = 'Aggregate'
            arcpy.AddField_management(inFeatureClass, 'MEAN_Z', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'MEAN_Aspect', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'STD_Aspect', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'MEAN_Slope_Deg', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'STD_Slope_Deg', 'DOUBLE')
            

            # Compute MEAN_Z, MEAN_Aspect, STD_Aspect for each flake

            inFeatureClass = 'Aggregate'
            with arcpy.da.UpdateCursor(inFeatureClass, ['MEAN_Aspect_wx', 'MEAN_Aspect_wy', 'STD_Aspect_wx',
                                                            'STD_Aspect_wy', 'SUM_Shape_Area', 'MEAN_Aspect',
                                                            'STD_Aspect']) as cursor:
                for row in cursor:
                    # row[6] = row[5] / row[4]                                               # compute z mean
                    row[5] = self.angle2bearing(math.degrees(math.atan2(row[1] / row[4], row[0] / row[4])))
                    # row[5] = (450 - (math.degrees(math.atan2(row[1]/row[4] , row[0]/row[4])))) % 360.  # for aspect mean atan2( y,x) convert to degrees then fix origins +X ccw vx +Y cw
                    row[6] = math.degrees(
                        math.atan2(row[3] / row[4], row[2] / row[4]))  # for aspect StDev atan2( y,x) convert to degrees
                    cursor.updateRow(row)
            del cursor
            del row

            with arcpy.da.UpdateCursor(inFeatureClass, ['SUM_Z_wgt', 'SUM_Shape_Area', 'MEAN_Z',
                                                            'SUM_Slope_wgt', 'MEAN_Slope_Deg', 'SUM_A_Slope_Sqrd',
                                                            'STD_Slope_Deg']) as cursor:
                for row in cursor:
                    row[2] = row[0] / row[1]  # compute z mean
                    row[4] = row[3] / row[1]  # compute slope mean

                    # Weighted standard deviation of slope
                    # variance = ((SUM_A_Slope_Sqrd * SUM_Shape Area) - (SUM_Slope_wgt **2))/ SUM_Shape_Area
                    # variance = ((row[5] * row[1]) - (row[3] **2))/ row[1]

                    variance = ((row[5] * row[1]) - (row[3] ** 2)) / row[1]
                    if variance > 0:
                        row[6] = math.sqrt(((row[5] * row[1]) - (row[3] ** 2)) / row[1])

                    cursor.updateRow(row)
            del cursor
            del row

            # Append Aggregate data to basin attribute table
            inFeatureClass = memoryBasin
            joinField = 'OBJECTID'
            joinTable = 'Aggregate'
            joinField2 = flakeFIDfield  # was 'FID_Basin'
            fieldList = ['FREQUENCY', 'MEAN_Slope_Deg', 'STD_Slope_Deg', 'MEAN_Aspect', 'STD_Aspect',
                         'MIN_Z_Min', 'MAX_Z_Max', 'MEAN_Z', 'SUM_SArea']

            arcpy.JoinField_management(inFeatureClass, joinField, joinTable, joinField2, fieldList)
            

            # Compute Flake orientation using Zonal Geometry ellipse and major/minor ratio
            if self.verbose:
                arcpy.AddMessage('     Compute flake ellipse data')

            inFeatureClass = memoryBasin
            zoneField = 'Id'
            zoneGeometryTable = 'in_memory\\FlakeEllipse'
            zoneCellSize = '0.05'
            outZonalGeometryAsTable = ZonalGeometryAsTable(inFeatureClass, zoneField, zoneGeometryTable, zoneCellSize)

            # Append Flake orientation data to basin attribute table
            inFeatureClass = memoryBasin
            joinField = 'OBJECTID'
            joinTable = 'in_memory\\FlakeEllipse'
            joinField2 = 'OBJECTID'
            fieldList = ['THICKNESS', 'MAJORAXIS', 'MINORAXIS', 'ORIENTATION']

            arcpy.JoinField_management(inFeatureClass, joinField, joinTable, joinField2, fieldList)
            


            # Add Roughness field and compute 3D area / 2d area
            inFeatureClass = memoryBasin
            fieldName = 'RoughnessByArea'
            arcpy.AddField_management(inFeatureClass, fieldName, 'DOUBLE')

            fieldName = 'EllipseRatio'
            arcpy.AddField_management(inFeatureClass, fieldName, 'DOUBLE')
            

            # Correct for slope effect.  Correcting area by cos(slope),
            # so roughness does not increase simply due to greater mean slope of flake.
            # Convert angle to bearing for orientation

            with arcpy.da.UpdateCursor(inFeatureClass,
                                           ['RoughnessByArea', 'MEAN_Slope_Deg', 'SUM_SArea', 'SHAPE@AREA',
                                            'MAJORAXIS', 'MINORAXIS', 'EllipseRatio', 'Orientation']) as cursor:

                for row in cursor:
                    row[6] = row[4] / (row[4] + row[5])  # Ellipse ratio
                    row[7] = self.orientation_angle2bearing(row[7])
                    if (row[1] > 0) and (row[2] > 0) and (row[3] > 0):  # crashes due to nulls without this
                        row[0] = math.cos(row[1] * math.pi / 180) * row[2] / row[3]
                        cursor.updateRow(row)
            del cursor
            del row

            # Add FlakeSize field and set it to 'THICKNESS' value
            inFeatureClass = memoryBasin
            fieldName = 'FlakeSize'

            arcpy.AddField_management(inFeatureClass, fieldName, 'DOUBLE')
            

            # Copy values of poorly named 'THICKNESS' into 'FlakeSize'
            arcpy.CalculateField_management(inFeatureClass, fieldName,
                                            '!THICKNESS! * 2.0',
                                            'PYTHON_9.3')
            

            # Delete old field 'THICKNESS'
            fieldName = 'THICKNESS'
            arcpy.DeleteField_management(inFeatureClass, fieldName)
            

            # Delete field 'gridcode'
            fieldName = 'gridcode'
            arcpy.DeleteField_management(inFeatureClass, fieldName)

            if inFlakePolyFC.startswith('Scar'):
                arcpy.CopyFeatures_management(memoryBasin, 'Flakes_' + inFlakePolyFC)
            elif inFlakePolyFC.startswith('SubDomain'):
                arcpy.CopyFeatures_management(memoryBasin, 'Domain' + inFlakePolyFC.replace('SubDomain', ''))
            else:
                arcpy.AddMessage('Error in polygon aggregation. Asked to process non SubDomain or Scar')

            arcpy.CopyFeatures_management(memoryTIN, 'TIN_Tri_' + inFlakePolyFC)
            

            # Delete intermediate TIN_Tri feature classes

            self.delete_preexisting('TIN_TriID_' + inFlakePolyFC)
            
            self.delete_preexisting('TIN_TriID_Mid_' + inFlakePolyFC)
            
            self.delete_preexisting('Aggregate')
            

            if self.verbose:
                arcpy.AddMessage('     Completed aggregation for polygon ' + inFlakePolyFC + '\n')


        # Clear in_memory workspace
        arcpy.Delete_management('in_memory')
        return True


    def create_auto_flake_scars(self):
        ''' Compute flake scar boundary polygons '''
        arcpy.env.workspace = self.final_gdb

        if self.is_autoflake_complete():
            arcpy.AddMessage('Skipping auto flake generation processing on pre-existing ' + self.pt_gdb)
            return

        try:

            self.compute_flake_basins('Residual7mm', 'zScarAutoFill')  # complete fill
            self.compute_flake_basins_multi('Residual7mm', 'zScarAuto')  # partial fills

            # Create basins multi may not produce a valid result for each zfill value
            # so find one to use as ScarAuto

            if arcpy.Exists('zScarAuto_05'):
                arcpy.CopyFeatures_management('zScarAuto_05', 'ScarAuto')
                
            elif arcpy.Exists('zScarAuto_025'):
                arcpy.CopyFeatures_management('zScarAuto_025', 'ScarAuto')
                
            else:
                arcpy.CopyFeatures_management('zScarAutoFill', 'ScarAuto')
                

            # Add detailed scar fcs to Edit GDB and add topology to enable scar editing
            # leave only ScarAuto in the final GDB

            zScarList = arcpy.ListFeatureClasses('zScar*')
            for zScar in zScarList:
                self.transfer_zscar(zScar)

        except:
            arcpy.AddMessage('Exception encountered in create_auto_flake_scars, processing terminated.')
            exit()

        self.process_flakes()
        return True

    def transfer_zscar(self, zScar):
        ''' Transfer ZScar to Edit GDB and add to topology with rules.
            Delete pre-existing zScars in Edit GDB if necessary
        '''

        try:
            if arcpy.Exists(zScar):                         # Does this zScar exist in final_gdb?
                # If pre-existing zScar is in edit_gdb dataset, must remove from topology, then delete
                # if in edit_gdb, expect it is in topology as well, so just look there first

                if not arcpy.Exists(self.artifact_topology):     # if topology missing create dataset and topology
                    self.create_artifact_feature_dataset(self.edit_gdb)
                d = arcpy.Describe(self.artifact_topology)
                if len(d.featureClassNames) > 0:
                    if zScar in d.featureClassNames:        # Is zScar in Scar_topology?
                        arcpy.RemoveFeatureClassFromTopology_management(self.artifact_topology, zScar)
                        # self.topology example is C:\GLiMR\GLiMR_GDB_Edits\G9943_1_1.gdb\Artifact\Scar_topology
                self.delete_preexisting(self.artifact_dataset + '\\' +  zScar)

                self.add_fc_to_artifact_fds(zScar, zScar, self.artifact_dataset)
                arcpy.Delete_management(zScar)              # delete zScar from final_gdb
                return True

        except:
            arcpy.AddMessage('*** Exception encountered in transfer_zscar.  Processing terminated.')
            exit()

    def create_artifact_feature_dataset(self, full_gdb):
        ''' If dataset does not exist, create new 'Artifact' feature dataset with common spatial reference
        to hold all feature classes and topology.  A feature dataset is used to allow for creating topology and
        topologically-consistent editing of shared scar polygon boundaries.

        Also create Scar_topology '''

        if arcpy.Exists(full_gdb + '\\Artifact'):
            return()
        else:
            # Create feature dataset to all artifact feature classes
            dataset = arcpy.CreateFeatureDataset_management(full_gdb, 'Artifact', self.sr)
            topology = arcpy.CreateTopology_management(dataset, 'Scar_topology', "#")
        return True


    def add_fc_to_artifact_fds(self, infc, outfc, fds):
        ''' Add feature class to feature dataset
        fds in form of <path>\\<name>.gdb\\Artifact, e.g. r"d:\GLiMR\GLiMR_Edits\G100_2_1.gdb\Artifact"  '''

        fclass = arcpy.FeatureClassToFeatureClass_conversion(infc, fds, outfc)  # class contained in fds

        if outfc.find('Scar')>=0:           # if output feature class name contains 'Scar', set up topology

            # topology = r'C:\GLiMR\GLiMR_GDB_Edits\G9943_1_1.gdb\Artifact\Scar_topology'
            # arcpy.AddFeatureClassToTopology_management(topology, fclass, "1", "1")
            arcpy.AddFeatureClassToTopology_management(self.artifact_topology, fclass, "1", "1")
            # Note: Using 'Scar_Topology' instead of self.topology fails with silent exit.  Full path issue?
            arcpy.AddRuleToTopology_management(self.artifact_topology, 'Must Not Overlap (Area)', fclass)
            arcpy.AddRuleToTopology_management(self.artifact_topology, 'Must Not Have Gaps (Area)', fclass)
            arcpy.ValidateTopology_management(self.artifact_topology, 'Full_Extent')
        else:
            arcpy.AddMessage('New feature class '+ outfc + ' does not begin with \'Scar\', so was not added to Topology')
        return True

    def run_transfer_all_stats(self):
        ''' Merge all feature classes into one summary table each for Flake, Outline and Re_entrant   '''
        self.gkey = 'Summarizing'  # to make output stream read correctly.  Gkey not used again (Obsolete?)

        arcpy.AddMessage('----\nCollecting and merging Flake, Outline and Re_entrant data from all geodatabases')
        arcpy.AddMessage('Results will be in geodatabase named \'a_GLiMR_Summary\'')

        # Create 'a_GLiMR_Summary' geodatabase if it does not exist'
        summary_gdb_short = 'a_GLiMR_Summary.gdb'
        summary_mdb_short = 'a_GLiMR_Summary.mdb'
        self.summary_gdb = self.gdb_dir + '\\' + 'a_GLiMR_Summary.gdb'

        arcpy.env.workspace = self.gdb_dir
        workspaces = arcpy.ListWorkspaces(summary_gdb_short, 'FileGDB')  # Create summary FILE geodatabase
        if len(workspaces) == 0:
            arcpy.AddMessage('No summary gdb, creating one...')
            arcpy.CreateFileGDB_management(self.gdb_dir, summary_gdb_short.strip('.gdb'))

        workspaces = arcpy.ListWorkspaces(summary_mdb_short, 'Access')  # Create summary ACCESS geodatabase
        if len(workspaces) == 0:
            arcpy.AddMessage('No summary mdb, creating one...')
            arcpy.CreatePersonalGDB_management(self.gdb_dir, summary_mdb_short)

        workspaces = arcpy.ListWorkspaces('*', 'FileGDB')

        arcpy.env.workspace = self.summary_gdb

        if len(workspaces) > 0:

            # Summarize to the summary file geodatabase
            self.summarize_fc(workspaces, 'Flake', 'FlakeScarSummary')
            self.summarize_fc(workspaces, 'Outline', 'OutlineSummary')
            self.summarize_fc(workspaces, 'Re_entrants', 'ReentrantSummary')
            self.summarize_fc(workspaces, 'Procrustes_Radial', 'ProcrustesRadialSummary')
            self.summarize_fc(workspaces, 'Procrustes_Exterior', 'ProcrustesExteriorSummary')
            self.summarize_fc(workspaces, 'Procrustes_Interior', 'ProcrustesInteriorSummary')

            # self.summarize_fc(workspaces, 'Procrustes_Blade', 'ProcrustesBladeSummary')
            # self.summarize_fc(workspaces, 'Procrustes_Haft', 'ProcrustesHaftSummary')
            # self.summarize_fc(workspaces, 'RE_Ellipse_pnts', 'EllipseSummary')
            # self.summarize_fc(workspaces, 'Domains', 'DomainsSummary')
        else:
            arcpy.AddMessage('No workspaces to summarize')

        # Copy results to the Access geodatabase to make available externally
        arcpy.env.workspace = self.summary_gdb

        self.copy2mdb('FlakeScarSummary', 'Flake', summary_mdb_short)
        self.copy2mdb('OutlineSummary', 'Outline', summary_mdb_short)
        self.copy2mdb('ReentrantSummary', 'Re_entrant', summary_mdb_short)
        self.copy2mdb('ProcrustesRadialSummary', 'Procrustes Radial', summary_mdb_short)
        self.copy2mdb('ProcrustesExteriorSummary', 'Procrustes Exterior', summary_mdb_short)
        self.copy2mdb('ProcrustesInteriorSummary', 'Procrustes Interior', summary_mdb_short)
        self.copy2mdb('ProcrustesBladeSummary', 'Procrustes Blade', summary_mdb_short)
        self.copy2mdb('ProcrustesHaftSummary', 'Procrustes Haft', summary_mdb_short)
        self.copy2mdb('EllipseSummary', 'Ellipse', summary_mdb_short)
        self.copy2mdb('DomainsSummary', 'Domains', summary_mdb_short)

        arcpy.AddMessage('\nSummary complete.  See ' + self.gdb_dir + '\\' + summary_mdb_short)
        return True

    def copy2mdb(self, summary_table_name, nickname, summary_mdb_short):
        # copies table from summary file gdb to summary access mdb
        if arcpy.Exists(summary_table_name):
            arcpy.AddMessage('   Copying ' + nickname + ' summary table to Access database')
            arcpy.FeatureClassToFeatureClass_conversion(summary_table_name, self.gdb_dir + '\\' + summary_mdb_short,
                                                    summary_table_name)

    def summarize_fc(self, workspaces, inFC_prefix, out_summary_table_name):

        arcpy.env.workspace = self.summary_gdb
        # arcpy.env.extent = arcpy.Extent(-10000., -10000.0, 10000.0, 10000.0)

        # Copy current summary <name> to Previous_<name> and delete old Previous_<name> if it exists
        if arcpy.Exists('Previous_' + out_summary_table_name):
            arcpy.Delete_management('Previous_' + out_summary_table_name)

        if arcpy.Exists(out_summary_table_name):
            arcpy.Rename_management(out_summary_table_name, 'Previous_' + out_summary_table_name)

        fullFCs = self.get_fc_list(workspaces, inFC_prefix, out_summary_table_name)
        if len(fullFCs) == 0:
            arcpy.AddMessage('   No merge done for ' + inFC_prefix + ' because there were no feature classes')
            return False

        arcpy.AddMessage('   Starting Merge for ' + str(len(fullFCs)) + ' ' + inFC_prefix + ' feature classes')
        arcpy.AddMessage('   The merge process may run for several minutes...')

        # Set local variables
        workspace = self.summary_gdb
        fc = out_summary_table_name
        geometry_type = "POLYGON"
        template = fullFCs[0]
        has_m = "DISABLED"
        has_z = "ENABLED"
        spatial_reference = self.sr

        arcpy.CreateFeatureclass_management(workspace, fc, geometry_type, template, has_m, has_z, spatial_reference)

        arcpy.Append_management(fullFCs, self.summary_gdb + '\\' + out_summary_table_name, 'NO_TEST')

        arcpy.AddMessage('   Completed merge for ' + inFC_prefix)
        return True

    def get_fc_list(self, workspaces, inFC_prefix, out_summary_table_name):
        '''Return list of Features with full path names for prefix such as 'Flake', excluding summary table    '''


        if self.verbose:
            arcpy.AddMessage('     Collecting ' + inFC_prefix + ' attribute feature class list')

        full_fcList = []

        # DEBUG add handle for no fc present

        for workspace in workspaces:
            arcpy.env.workspace = workspace
            if inFC_prefix in ['Outline', 'Re_entrants', 'Domains']:
                listFCs = arcpy.ListFeatureClasses(inFC_prefix)
            else:
                listFCs = arcpy.ListFeatureClasses(inFC_prefix + '*')

            for currentFC in listFCs:
                fullFC = workspace + '\\' + currentFC
                if (currentFC != out_summary_table_name):
                    full_fcList.append(fullFC)

        if self.verbose:
            arcpy.AddMessage('     Total number of ' + inFC_prefix + ' feature classes: ' + str(len(full_fcList)))

        return full_fcList

    def timestampstring(self):
        ''' Return a formatted time string '''
        if (time.localtime().tm_isdst):  # returns readable string with time zone
            return time.asctime(time.localtime()) + ' PDT : '  # use time.tzname[1] instead if needed
        else:
            return time.asctime(time.localtime()) + ' PST : '  # use time.tzname[0] instead if needed
        return True

    def visited(self, inFC, inField):
        '''See if a programatically created field exists to prevent double-processing'''
        arcpy.env.workspace = self.final_gdb

        fields = arcpy.ListFields(inFC)
        for field in fields:
            if field.name == inField:
                return True
        return False

    def get_arcgis_licenses(self):
        ''' Check out ArcGIS extension licenses'''
        try:
            if arcpy.CheckExtension('3D') == 'Available':
                arcpy.CheckOutExtension('3D')
            else:
                raise LicenseError
        except LicenseError:
            arcpy.AddMessage('3D Analyst license is unavailable')
        except:
            print arcpy.GetMessages(2)

        try:
            if arcpy.CheckExtension('Spatial') == 'Available':
                arcpy.CheckOutExtension('Spatial')
            else:
                raise LicenseError
        except LicenseError:
            arcpy.AddMessage('Spatial Analyst license is unavailable')
        except:
            print arcpy.GetMessages(2)
        return True


    def delete_preexisting(self, thing):
        '''Delete pre-existing object.  If not an XYZ file, check the schema lock first '''

        if arcpy.Exists(thing):
            if not (thing.endswith('.xyz') or thing.endswith('_lyr') or thing.endswith('.GDB')
                                                        or thing.endswith('.gdb')):
                if not (arcpy.TestSchemaLock(thing)):
                    arcpy.AddMessage('Note:  Schema lock can not be acquired on ' + thing +
                                     '  Used by another application.  Close ArcCatalog if open.')
                    # add error handling here
                    # exit()

            arcpy.Delete_management(thing)
        return True

    def create_gdb(self, gdb_dir, gdb_name):
        '''Create file geodatabases.  Delete pre-existing GDB without asking'''

        if not os.path.exists(gdb_dir):
            arcpy.AddError('The Edit GDB directory specified does not exist:  ' + gdb_dir)

        if arcpy.Exists(gdb_dir + '\\' + gdb_name):
            arcpy.AddMessage('    * Deleting pre-existing GDB for ' + self.gkey + ' and recreating it')
            self.delete_preexisting(gdb_dir + '\\' + gdb_name)

        # Execute Create File GDBs
        if self.verbose:
            arcpy.AddMessage('     Creating Geodatabase for ' + gdb_name + ' in ' + gdb_dir)

        arcpy.CreateFileGDB_management(gdb_dir, gdb_name)
        if self.verbose:
            arcpy.AddMessage('     Created Geodatabase for ' + gdb_name + ' in ' + gdb_dir)
        if gdb_dir == self.gdb_edit_dir:            # if this is edit gdb, create Artifact feature dataset
            self.create_artifact_feature_dataset(gdb_dir + '\\' + gdb_name)
        return True


    def create_outline(self, in_point_fc, outTIN, outTIN_Tri, out_outline_poly, max_length=0.5, dim3D=True):
        ''' Create outline polygon feature class, TIN, TIN_Tri from points'''
        # create TIN
        # Delineate TIN  (as separate function using max_length)
        # Use TIN domain to generate outline
        # compute centroid
        # Create TIN_Tri polygon fc

        # arcpy.AddMessage('Inside create outline')

        tin_check = False

        while tin_check == False:
            # iterate max_length in Delineate if delineation eats out TIN core while creating outline
            # arcpy.AddMessage('Before Create Delineated TIN: '  + in_point_fc +',   ' + outTIN +',   ' +  str(max_length))
            self.create_delineated_TIN(in_point_fc, outTIN, max_length)

            # arcpy.AddMessage('After create Delineated')

            arcpy.TinDomain_3d(outTIN, out_outline_poly, 'POLYGON')

            # check Area/Perimeter of Outline: if <1.0 then bad delineation

            with arcpy.da.SearchCursor(out_outline_poly, ['Shape@Area', 'Shape@Length']) as cursor:
                for row in cursor:
                    # Only one row...
                    area = row[0]
                    length = row[1]
            del cursor
            del row

            if self.verbose:
                arcpy.AddMessage('     DELINEATE: ' + ';  Area' + str(area) + ';  Length ' + str(length) + ';  A/L  ' + str(area / length) + ';  m_l ' + str(max_length))

            if (area / length) > 1.0:
                tin_check = True
            else:
                max_length = max_length + 0.5
                if max_length > 3.0:
                    arcpy.AddMessage('ERROR: Point ' + self.gkey + ' fails Delineation.  ' \
                                                        'Inter-point spacing might be unexpectedly large')
                    arcpy.AddMessage('Abandoning further processing')
                    exit()

            # Get X, Y of centroid
            self.compute_centroid(out_outline_poly)

            if dim3D:
                arcpy.TinTriangle_3d(outTIN, outTIN_Tri, 'DEGREE', '1', '', '')
        return True


    def compute_centroid(self, outline_poly):
        '''Compute centroid feature class'''
        if self.verbose:
            arcpy.AddMessage('     Computing centroid feature class')

        out_centroid = outline_poly + '_centroid'

        arcpy.FeatureToPoint_management(outline_poly, out_centroid, 'CENTROID')

        # Set local variables
        properties = 'CENTROID; EXTENT'
        length_unit = ''
        area_unit = ''
        coordinate_system = ''

        # Generate the extent coordinates using Add Geometry Properties tool
        arcpy.AddGeometryAttributes_management(outline_poly, properties, length_unit, area_unit, coordinate_system)
        

        if self.verbose:
            arcpy.AddMessage('     Fetch X,Y centroid values')

        # Get X, Y of centroid
        with arcpy.da.SearchCursor(outline_poly, ['CENTROID_X', 'CENTROID_Y']) as cursor:
            for row in cursor:
                # Only one row...
                centroid_x = row[0]
                centroid_y = row[1]
        del cursor
        del row

        self.centroid_x = centroid_x
        self.centroid_y = centroid_y

        if self.verbose:
            arcpy.AddMessage('     For ' + outline_poly + ':    Centroid X: ' + str(round(centroid_x, 3)) +
                    ',    Centroid Y: ' + str(round(centroid_y, 3)))
        return True


    def create_delineated_TIN(self, in_point_fc, outTIN, max_length):
        '''Create and delineate TIN and trap bad delineation'''
        # arcpy.AddMessage('Inside create delineated TIN')
        # if self.verbose:
            # arcpy.AddMessage('     Create TIN named ' + outTIN + ' from ' + in_point_fc)

        # Process: TIN
        if arcpy.Exists(outTIN):
            arcpy.Delete_management(outTIN)

        s = in_point_fc + ' Shape.Z Mass_Points <None>'       # using 3D points
        # s = in_point_fc + ' grid_code Mass_Points <None>'     # using points from raster
        arcpy.CreateTin_3d(outTIN, self.sr, s, 'DELAUNAY')

        if self.verbose:
            arcpy.AddMessage('     Delineate TIN max_length = ' + str(max_length))

        # Process: Delineate TIN Data Area  (removes exterior spider webs)
        arcpy.DelineateTinDataArea_3d(outTIN, max_length, 'PERIMETER_ONLY')
        return True


    def shift_points_in_xy(self, inPointFC, outPointFC, dX, dY, scale_x=1.0, scale_y=1.0):
        ''' Shift inPointFC in X,Y by dX and dY  and scale by scale_X and scale_y to yield outPointFC '''

        # if self.verbose:
            # arcpy.AddMessage('     Create point feature class with shifted XY values ' + inPointFC + ' and  ' + outPointFC)
            # arcpy.AddMessage('     dX = ' + str(dX) + ',  dY = ' + str(dY))

        if arcpy.Exists(outPointFC):
            self.delete_preexisting(outPointFC)


        # Execute CreateFeatureclass
        arcpy.CreateFeatureclass_management(self.final_gdb, outPointFC, 'POINT',  has_z= 'ENABLED', spatial_reference= self.sr)  #DEBUG
        arcpy.CreateFeatureclass_management('in_memory', outPointFC, 'POINT',  has_z= 'ENABLED', spatial_reference= self.sr)
        # arcpy.AddMessage('     created point feature class with shifted XY values')

        # Read OrigPts and write to PtsXYShift applying XY shift of mean center
        with arcpy.da.SearchCursor(inPointFC, ['SHAPE@XY', 'SHAPE@Z']) as cursor1:
            with arcpy.da.InsertCursor('in_memory\\' + outPointFC, ['SHAPE@']) as cursor2:
                # Get X, Y, Z of OrigPts
                for InPnt in cursor1:
                    x_original, y_original = InPnt[0]
                    z_original = InPnt[1]

                    # Shift in XY and write to outPointFC
                    xyz_new = (scale_x * (x_original - dX), scale_y * (y_original - dY), z_original)
                    cursor2.insertRow([xyz_new])
        del cursor1
        del cursor2
        del InPnt
        # arcpy.AddMessage('     Finished creating point feature class with shifted XY values')
        return True

    def rotate_points(self, inPointFC):
        ''' Rotate points to remove minor tilt of object in the scanner.
            This is not a true rotation, rather an adjustment of each point in Z
            to reflect a first order (planar) trend surface correction.
            The XY positional error introduced is negligible and is reflected by the
            non-zero centroid of the outline of the shifted point cloud. '''

        # Process: Trend
        if self.verbose:
            arcpy.AddMessage('     Creating trend surface')

        arcpy.Trend_3d(inPointFC, 'Shape.Z', 'in_memory\\Trend_XYCentered', '0.05', '1', 'LINEAR', '')
        # first order trend (plane) and cell size = 0.05
        

        # Process: Spline with Barriers on original points
        if self.verbose:
            arcpy.AddMessage('     Creating topo raster on shifted points')

        self.create_surface(inPointFC, 'in_memory\\Topo_XYCentered')

        # Process: Minus  XY centered topo - trend surface (1st order) to yield reoriented surface
        # Note that this is a shift in Z, not X or Y, so it is not a true rotation of the points.
        # At the small angles involved, the error is tiny.

        if self.verbose:
            arcpy.AddMessage('     Creating Topo_XYCentered - Trend_XYCentered raster')

        arcpy.Minus_3d('in_memory\\Topo_XYCentered', 'in_memory\\Trend_XYCentered', 'in_memory\\Topo_Normalized')
        
        arcpy.Delete_management('in_memory\\Topo_XYCentered')
        

        # Process: Extract Values to Points  NOTE: Value in RASTERVALU attribute, not shape.z
        if self.verbose:
            arcpy.AddMessage('     Creating Trend points')

        ExtractValuesToPoints(inPointFC, 'in_memory\\Trend_XYCentered', 'Trend_Pts',
                              'INTERPOLATE', 'VALUE_ONLY')
        
        arcpy.Delete_management('in_memory\\Trend_XYCentered')

        # Process: Extract Values to Points  NOTE: Value in RASTERVALU attribute, not shape.z
        if self.verbose:
            arcpy.AddMessage('     Creating the Pts_Flattened reflecting the amount of '
                             'delta Z adjustment from the trend surface')

        ExtractValuesToPoints(inPointFC, 'in_memory\\Topo_Normalized', 'in_memory\\Pts_Flattened',
                              'INTERPOLATE', 'VALUE_ONLY')
        

        # The Pts_Flattened are the pseudorotated (flattened) points with Z values prior to the bulk z shift
        return True


    def shift_points_in_z(self, inPointFC, outPointFC):
        '''Shift points in Z'''

        # Use Describe to access the extent object
        desc = arcpy.Describe(inPointFC)

        xmin = desc.extent.XMin
        xmax = desc.extent.XMax
        ymin = desc.extent.YMin
        ymax = desc.extent.YMax

        if self.verbose:
            arcpy.AddMessage('\n     Original point cloud extent:\n       XMin: {0}, XMax: {1}, YMin: {2}, YMax: {3}'.format(xmin, xmax, ymin, ymax))

        # Describe object properties extent.ZMin, Extent.ZMax or ZExtent do not return values,
        # instead return None.  So loop through manually...
        zmin = 100000
        zmax = -100000

        with arcpy.da.SearchCursor(inPointFC, 'RASTERVALU') as cursor:
            for row in cursor:
                z = row[0]
                if z < zmin and z > -9999.0:  # -9999 is RASTERVALU no data
                    zmin = z
                if z > zmax:
                    zmax = z
        del cursor
        del row

        self.zRange = zmax - zmin

        if self.verbose:
            arcpy.AddMessage('       Z range: ' + str(zmin) + ', ' + str(zmax) + '\n')

        # Shift Topo_Normalized surface in Z my an amount zmin to make lowest point at elevation zero.
        # New surface is TopoBase

        # Process: Minus  Topo_Normalized - zmin to yield reoriented and shifted surface TopoBase
        if self.verbose:
            arcpy.AddMessage('     Shifting surface vertically by ' + str(-1 * zmin))

        arcpy.Minus_3d('in_memory\\Topo_Normalized', zmin, 'in_memory\\TopoBase')
        
        arcpy.Delete_management('in_memory\\Topo_Normalized')
        

        # Process: Extract Values to Points  NOTE: Value in RASTERVALU attribute, not shape.z
        if self.verbose:
            arcpy.AddMessage('     Creating the points for flattened and shifted surface, called BasePts')

        ExtractValuesToPoints('in_memory\\PtsXYCentered', 'in_memory\\TopoBase', 'in_memory\\BasePtsNoZ',
                              'INTERPOLATE', 'VALUE_ONLY')
        
        arcpy.Delete_management('in_memory\\TopoBase')
        

        if self.verbose:
            arcpy.AddMessage('     Create BasePts point feature class with shifted XY and Z values')

        arcpy.CreateFeatureclass_management('in_memory', 'BasePts', 'POINT',  has_z= 'ENABLED', spatial_reference= self.sr)
        

        # Read BasePtsNoZ and write to BasePts applying Z shift of RASTERVALU to shape.z
        with arcpy.da.SearchCursor('in_memory\\BasePtsNoZ', ['SHAPE@XY', 'RASTERVALU']) as cursor1:
            with arcpy.da.InsertCursor('in_memory\\BasePts', ['SHAPE@']) as cursor2:
                # Get X, Y, Z of OrigPts
                for InPnt in cursor1:
                    x_original, y_original = InPnt[0]
                    z_shift = InPnt[1]

                    # Shift in XY and write to PtsXYCentered
                    if z_shift <> -9999.:
                        xyz_new = (x_original, y_original, z_shift)
                        cursor2.insertRow([xyz_new])

        del cursor1
        del cursor2
        del InPnt

        arcpy.Delete_management('in_memory\\PtsXYCentered')
        arcpy.Delete_management('in_memory\\BasePtsNoZ')

        if self.verbose:
            arcpy.AddMessage('     Extend topo surface')
            arcpy.AddMessage('     Create BasePtsExt point feature class and append extended corners')

        # Insert four corner extension points
        ptz = -(zmax - zmin) / 2  # Elevation of corner points to be added

        # Write ASCII file of shifted points, add four extended point, re-import file
        # (circumvents bounds problem with insert cursor)  ADD EXISTENCE CHECK fails if it exists (revisit)
        extFileName = 'Ext_' + self.xyz_file
        extFileNameandPath = self.xyz_dir + '\\Ext_' + self.xyz_file

        self.delete_preexisting(extFileNameandPath)

        arcpy.FeatureClassZToASCII_3d('in_memory\\BasePts', self.xyz_dir, extFileName, 'XYZ')
        

        f = open(extFileNameandPath, mode='a')

        xyz = str(xmin - 10) + ' ' + str(ymax + 10) + ' ' + str(ptz) + '\n'  # NW
        f.writelines(xyz)

        xyz = str(xmax + 10) + ' ' + str(ymax + 10) + ' ' + str(ptz) + '\n'  # NE
        f.writelines(xyz)

        xyz = str(xmin - 10) + ' ' + str(ymin - 10) + ' ' + str(ptz) + '\n'  # SW
        f.writelines(xyz)

        xyz = str(xmax + 10) + ' ' + str(ymin - 10) + ' ' + str(ptz) + '\n'  # SE
        f.writelines(xyz)
        f.close()

        arcpy.ASCII3DToFeatureClass_3d(extFileNameandPath, 'XYZ', 'BasePtsExt', 'POINT',
                                       '1', self.sr, '', '', 'DECIMAL_POINT')
        

        self.delete_preexisting(extFileNameandPath)

        arcpy.SplineWithBarriers_3d('BasePtsExt', 'SHAPE', '#', 0.05, 'TopoExt')
        

        if self.verbose:
            arcpy.AddMessage('     Rotated surface flat based on first order trend surface')
            arcpy.AddMessage('     Shifting surface vertically by ' + str(-1 * zmin))

        arcpy.CopyFeatures_management('in_memory\\BasePts', 'BasePts')
        
        arcpy.Delete_management('in_memory\\BasePts')
        return True


    def create_TIN_edge(self, inTIN, outFC):
        '''Create TIN edge line feature class from TIN'''
        if self.verbose:
            arcpy.AddMessage('     Create TIN_edge polyline feature class')

        arcpy.TinEdge_3d(inTIN, outFC, 'DATA')  # not used again so not saved in memory
        return True


    def compute_outline_properties(self, outlineFC):
        '''Compute Height, width, h/w ratio, and do field name clean-up'''

        arcpy.DeleteField_management(outlineFC, ['CENTROID_Z'])
        arcpy.AddField_management(outlineFC, 'Height', 'DOUBLE')
        arcpy.AddField_management(outlineFC, 'Width', 'DOUBLE')
        arcpy.AddField_management(outlineFC, 'HW_ratio', 'DOUBLE')
        arcpy.AddField_management(outlineFC, 'Half_Thickness', 'DOUBLE')
        arcpy.AddField_management(outlineFC, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(outlineFC, 'Hull_Area', 'DOUBLE')
        arcpy.AddField_management(outlineFC, 'Hull_Length', 'DOUBLE')

        

        # Compute height and width of outline polygon
        with arcpy.da.UpdateCursor(outlineFC, ['Height', 'Width', 'EXT_MIN_X', 'EXT_MAX_X', 'EXT_MIN_Y', 'EXT_MAX_Y',
                                               'HW_ratio', 'Half_Thickness', 'GKey'])as cursor:
            for row in cursor:
                # Only one row...
                row[0] = row[5] - row[4]
                row[1] = row[3] - row[2]
                overallWidth = row[3] - row[2]
                overallHeight = row[5] - row[4]
                row[6] = overallHeight / overallWidth
                row[7] = self.zRange
                row[8] = self.gkey

                self.outline_height = overallHeight  # set up expecting to be run on 'Outline' FC only
                self.outline_width = overallWidth
                self.outline_min_x = row[2]
                self.outline_max_x = row[3]
                self.outline_min_y = row[4]
                self.outline_max_y = row[5]
                cursor.updateRow(row)
        del cursor
        del row

        if self.verbose:
            arcpy.AddMessage('     Outline:   Height: ' + str(round(overallHeight, 1)) + '; Width = ' +
                    str(round(overallWidth, 1)) + '; H/W = ' + str(round((overallHeight / overallWidth), 3)))
        return True


    def create_re_entrants(self, in_outlineFC, out_Re_entrantFC):
        '''Compute convex hull, reentrants and reentrant properties
        in_outlineFC is outline polygon'''

        # Create convex hull around outline
        if self.verbose:
            arcpy.AddMessage('     Creating Convex Hull')

        outFeatureClass = 'Hull'
        arcpy.MinimumBoundingGeometry_management(in_outlineFC, outFeatureClass, 'CONVEX_HULL', 'NONE')
        
        arcpy.DeleteField_management(outFeatureClass, ['ORIG_FID', 'Half-Thickness'])

        with arcpy.da.SearchCursor(outFeatureClass, ['Shape@Area', 'Shape@Length']) as cursor:
            for row in cursor:
                # Only one row...
                self.hull_area = row[0]
                self.hull_length = row[1]
            del cursor
            del row

        arcpy.DeleteField_management('Hull',
                                     ['Height', 'Width', 'HW_ratio', 'Half_Thickness', 'Hull_Area', 'Hull_length'])

        with arcpy.da.UpdateCursor('Hull', 'GKey') as cursor:
            for row in cursor:
                row[0] = self.gkey
                cursor.updateRow(row)
        del cursor
        del row

        with arcpy.da.UpdateCursor('Outline', ['Hull_Area', 'Hull_Length']) as cursor:
            for row in cursor:
                row[0] = self.hull_area
                row[1] = self.hull_length
                cursor.updateRow(row)
        del cursor
        del row

        # actually create the re_entrants
        reentrantEditFC = self.gdb_edit_dir + '\\' + self.pt_gdb + '\\' + 'Re_entrants'

        if arcpy.Exists(reentrantEditFC):
            self.fetch_persistent_data()
            return

        else:
            #  create new reentrant feature class
            #  self.delete_preexisting('Hull_Outline')
            if self.verbose:
                arcpy.AddMessage('     Creating Re_entrants')

            inFeatures = in_outlineFC
            updateFeatures = 'Hull'
            outFeatureClass = 'Hull_Outline'
            clusterTolerance = 0.05  # adjusting this can change reentrant size
            
            # Execute Symmetrical difference to get single-part polygon of area between hull and outline.
            #         clusterTolerance represents how close is considered equal between outline and hull
            arcpy.SymDiff_analysis(inFeatures, updateFeatures, outFeatureClass, 'ONLY_FID', clusterTolerance)

            # break up re-entrants into individual features and compute centroids
            arcpy.MultipartToSinglepart_management('Hull_Outline', 'in_memory\\Re_entrants')
            
            arcpy.Delete_management('Hull_Outline')

            inFeatures = 'in_memory\\Re_entrants'
            arcpy.DeleteField_management(inFeatures, ['FID_Outline', 'FID_Hull', 'ORIG_FID'])

            # Set local variables
            properties = 'CENTROID'
            length_unit = ''
            area_unit = ''
            coordinate_system = ''

            # Generate the extent coordinates using Add Geometry Properties tool
            arcpy.AddGeometryAttributes_management(inFeatures, properties, length_unit, area_unit, coordinate_system)
            

            # Compute Reentrant orientation using Zonal Geometry ellipse and major/minor ratio
            zoneField = 'OBJECTID'
            zoneGeometryTable = 'in_memory\\REEllipse'
            zoneCellSize = '0.05'
            outZonalGeometryAsTable = ZonalGeometryAsTable('in_memory\\Re_entrants', zoneField,
                                                           zoneGeometryTable, zoneCellSize)
            

            # Append Re-entrant orientation data to basin attribute table
            joinField = 'OBJECTID'
            joinTable = 'in_memory\\REEllipse'
            joinField2 = 'OBJECTID'
            fieldList = ['THICKNESS', 'MAJORAXIS', 'MINORAXIS', 'ORIENTATION']

            arcpy.JoinField_management('in_memory\\Re_entrants', joinField, joinTable, joinField2, fieldList)
            


            # Add Ellipse ratio, re-entrant size and point name to re-entrants
            inFeatureClass = 'in_memory\\Re_entrants'
            arcpy.AddField_management(inFeatureClass, 'GKey', 'TEXT', '', '', 25)
            arcpy.AddField_management(inFeatureClass, 'EllipseRatio', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'RE_Size', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Type_', 'TEXT', '', '', 15)
            arcpy.AddField_management(inFeatureClass, 'Pct_RE_Size', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Pct_Centroid_X', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Pct_Centroid_Y', 'DOUBLE')
            arcpy.AddField_management(inFeatureClass, 'Pct_Hull_Area', 'DOUBLE')

            # update orientation angle
            with arcpy.da.UpdateCursor(inFeatureClass, ['THICKNESS', 'MAJORAXIS', 'MINORAXIS', 'EllipseRatio',
                                                            'RE_Size', 'GKey', 'ORIENTATION', 'Shape@Area',
                                                            'Pct_Hull_Area']) as cursor:
                for row in cursor:
                    row[4] = row[0] * 2.0
                    row[3] = row[1] / (row[1] + row[2])
                    row[5] = self.gkey
                    row[6] = self.orientation_angle2bearing(row[6])
                    row[8] = 100. * row[7] / self.hull_area
                    cursor.updateRow(row)
            del cursor
            del row

            # Delete old field 'THICKNESS'
            fieldName = 'THICKNESS'
            arcpy.DeleteField_management(inFeatureClass, fieldName)
            
            fieldName = 'CENTROID_Z'
            arcpy.DeleteField_management(inFeatureClass, fieldName)
            

            arcpy.CopyFeatures_management('in_memory\\Re_entrants', out_Re_entrantFC)
            arcpy.Delete_management('in_memory\\Re_entrants')

            # Categorize Type for notches

            notches = self.make_notch_array()
            # [OID@', 'Type_', 'RE_Size', 'Pct_Centroid_X', 'Pct_Centroid_Y', 'Pct_RE_Size', 'CENTROID_X', 'CENTROID_Y']

            self.analyze_notches(notches)



            if not arcpy.Exists(self.gdb_edit_dir):
                arcpy.AddMessage('Problem with path ' + self.gdb_edit_dir + '\\' + self.pt_gdb)


            # arcpy.AddMessage('Checking  edit gdb 1')

            if not arcpy.Exists(self.gdb_edit_dir + '\\' + self.pt_gdb):  # create persistent gdb if necessary
                self.create_gdb(self.gdb_edit_dir, self.pt_gdb)
                # arcpy.AddMessage('Created  edit gdb 2: ' + self.edit_gdb)

            if not arcpy.Exists(self.gdb_edit_dir + '\\' + self.pt_gdb + '\\' + 'Re_entrants'):  # copy to persistent gdb
                arcpy.CopyFeatures_management('Re_entrants',
                                              self.gdb_edit_dir + '\\' + self.pt_gdb + '\\' + 'Re_entrants')
        return True


    def make_notch_array(self):
        '''
        Find min and max values, update Reentrants with Pct_* values and return notches array
        '''
        inFeatureClass = 'Re_entrants'

        X_min = 10000.
        Y_min = 10000.
        X_max = -10000.
        Y_max = -10000.
        RE_size_max = 0.

        with arcpy.da.SearchCursor(inFeatureClass, ['OID@', 'Type_', 'RE_Size', 'CENTROID_X', 'CENTROID_Y',
                                                    'MAJORAXIS', 'MINORAXIS', 'EllipseRatio', 'ORIENTATION']) as cursor:
            for row in cursor:

                OID = row[0]
                Type = row[1]
                RE_Size = row[2]
                CENTROID_X = row[3]
                CENTROID_Y = row[4]
                MAJORAXIS = row[5]
                MINORAXIS = row[6]
                EllipseRatio = row[7]
                ORIENTATION = row[8]

                if CENTROID_X > X_max: X_max = CENTROID_X
                if CENTROID_Y > X_max: Y_max = CENTROID_Y
                if CENTROID_X < X_max: X_min = CENTROID_X
                if CENTROID_Y < Y_max: Y_min = CENTROID_Y
                if RE_Size > RE_size_max: RE_size_max = RE_Size

        # Now have centroid and size min and max
        del cursor
        del row

        self.normalize_notches(RE_size_max)  # This sets Pct_* values in the Reentrants FC

        notches = []
        with arcpy.da.SearchCursor(inFeatureClass, ['OID@', 'Type_', 'RE_Size', 'Pct_Centroid_X', 'Pct_Centroid_Y',
                                                    'Pct_RE_Size', 'CENTROID_X', 'CENTROID_Y']) as cursor:
            for row in cursor:
                notch = [row[0], row[1], row[2], row[3], row[4], row[5], row[6], row[7]]
                notches.append(notch)
        del cursor
        del row
        del notch

        return notches

    def analyze_notches(self, notches):
        ''' Determine notch type  '''

        # notches
        # [OID@', 'Type_', 'RE_Size', 'Pct_Centroid_X', 'Pct_Centroid_Y', 'Pct_RE_Size', 'CENTROID_X', 'CENTROID_Y']

        # First pass to set initial Type

        bases = []
        rights = []
        lefts = []

        for notch in notches:

            OID = notch[0]
            notch[1] = 'Undefined'
            RE_Size = notch[2]
            Pct_Centroid_X = notch[3]
            Pct_Centroid_Y = notch[4]
            Pct_RE_Size = notch[5]

            # Basal notches   Set Type
            if (Pct_Centroid_X > 44 and Pct_Centroid_X < 56 and Pct_Centroid_Y < 20) or (Pct_Centroid_Y < 20
                                                                                         and Pct_Centroid_X > 36 and Pct_Centroid_X < 64 and Pct_RE_Size > 10 and RE_Size > 0.3 ):
                # or clause was 30-70 on C_X
                notch[1] = 'Base'
                bases.append(notch)

            # Left and right notches   Set Type
            if (Pct_Centroid_Y < 90. and Pct_Centroid_Y > 2. and (Pct_Centroid_X < 34. )
                and (Pct_RE_Size > 40. and RE_Size > 0.4)):
                notch[1] = 'Left'
                lefts.append(notch)

            if (Pct_Centroid_Y < 90. and Pct_Centroid_Y > 2 and (Pct_Centroid_X > 66)
                and (Pct_RE_Size > 40 and RE_Size > 0.4)):
                notch[1] = 'Right'
                rights.append(notch)

        # Compare left and right for similarity
        # Get centroids for L and R
        # Case 1: 1 L and 1 R
        # Case 2: 1 R or L and more than 1 R or L
        # Case 3: Multiple L and R
        # Case 4: Solo

        # Case 1: 1 L and 1 R
        # print 'Before:  Rights: ', len(rights),'   Lefts: ', len(lefts),'   Bases: ', len(bases)

        if len(rights) == 1 and len(lefts) == 1:
            pct_Centroid_X1 = lefts[0][3]
            pct_Centroid_Y1 = lefts[0][4]
            pct_RE_Size1 = lefts[0][5]

            pct_Centroid_X2 = rights[0][3]
            pct_Centroid_Y2 = rights[0][4]
            pct_RE_Size2 = rights[0][5]

            if self.compare_two_normalized_notch_centroids(pct_Centroid_X1, pct_Centroid_Y1, pct_Centroid_X2,
                                                           pct_Centroid_Y2, pct_RE_Size1, pct_RE_Size2, 20., 9., 60.):
                # opened dSize from 40 to 60
                pass
                # print 'LR compare True'
                # print lefts[0]
                # print rights[0]

            else:
                # print 'False for'       # DEBUG 144_2 will be special (compare base incl?
                # print lefts[0]
                # print rights[0]
                # print self.pt_name + ' failed Left -Right comparison'
                # see if one is good...
                if lefts[0][4] > 50. and rights[0][4] < 50:  # compare Pct_Centroid_Y
                    # nuke left
                    lefts[0][1] = 'EdgeL'
                if lefts[0][4] < 50. and rights[0][4] > 50:
                    # nuke right
                    rights[0][1] = 'EdgeR'

        # Case 2A: 1 L and more than 1 R
        if len(rights) > 1 and len(lefts) == 1:  # DEBUG handle case where 'Good one is really bad if all rejected
            # DEBUG if lefts or rights > 3 probably surfboard
            # print self.pt_name + ' has multiple RIGHT reentrants'

            # compare various right centroids to single left
            pct_Centroid_X1 = lefts[0][3]
            pct_Centroid_Y1 = lefts[0][4]
            pct_RE_Size1 = lefts[0][5]

            # print 'Left for compare'
            # print lefts[0]

            j = 0
            for right in rights:
                pct_Centroid_X2 = right[3]
                pct_Centroid_Y2 = right[4]
                pct_RE_Size2 = right[5]

                if not self.compare_two_normalized_notch_centroids(pct_Centroid_X1, pct_Centroid_Y1, pct_Centroid_X2,
                                                                   pct_Centroid_Y2, pct_RE_Size1, pct_RE_Size2, 20., 9.,
                                                                   60.):
                    # this notch will be changed from Right to Right_Reject (eventually edge)
                    # print 'rejecting RIGHT notch '
                    # print right
                    rights[j][1] = 'EdgeR'

                    # else:
                    # print 'accepting RIGHT notch'
                    # print right

                j = j + 1

                # count remaining left and right if solo check centroid Y and kill north ones

        # Case 2B: 1 R and more than 1 L
        if len(rights) == 1 and len(lefts) > 1:  # DEBUG handle case where 'Good one is really bad if all rejected
            # print self.pt_name + ' has multiple LEFT reentrants'
            # compare various left centroids to single right
            pct_Centroid_X2 = rights[0][3]
            pct_Centroid_Y2 = rights[0][4]
            pct_RE_Size2 = rights[0][5]

            # print 'Right for compare'
            # print rights[0]

            j = 0
            for left in lefts:
                pct_Centroid_X1 = left[3]
                pct_Centroid_Y1 = left[4]
                pct_RE_Size1 = left[5]

                if not self.compare_two_normalized_notch_centroids(pct_Centroid_X1, pct_Centroid_Y1, pct_Centroid_X2,
                                                                   pct_Centroid_Y2, pct_RE_Size1, pct_RE_Size2, 20., 9.,
                                                                   60.):
                    # this notch will be changed from Right to Right_Reject (eventually edge)
                    # print 'rejecting LEFT notch '
                    # print left
                    lefts[j][1] = 'EdgeL'

                    # else:
                    # print 'accepting LEFT notch'
                    # print left

            j = j + 1

        # Case 3: Multiple L and R
        if len(rights) > 1 and len(lefts) > 1:
            # print self.pt_name + ' has MULTIPLE RIGHT and LEFT reentrants'
            for i in range(len(rights)):
                rights[i][1] = 'EdgeSR'

            for i in range(len(lefts)):
                lefts[i][1] = 'EdgeSL'

                # Work to do... Surfboard

        # Case 5A: Solo L
        if len(rights) == 0 and len(lefts) == 1:
            # print self.pt_name + ' has SOLO LEFT reentrant'
            lefts[0][1] = 'SoloL'

        # Case 5B: Solo L
        if len(rights) == 1 and len(lefts) == 0:
            # print self.pt_name + ' has SOLO RIGHT reentrant'
            rights[0][1] = 'SoloR'

        # from rights and lefts arrays: update notches array

        self.update_re_entrants_from_notches(notches)
        return True


    def update_re_entrants_from_notches(self, notches):
        # Update the re-entrants feature class from the notches array

        with arcpy.da.UpdateCursor('Re_entrants', ['OID@', 'Type_', 'RE_Size',
                                                   'Pct_Centroid_X', 'Pct_Centroid_Y', 'Pct_RE_Size']) as cursor:
            i = 0
            for row in cursor:
                if row[0] == notches[i][0]:  # Update notches array  back to Re_entrants FC
                    row[1] = notches[i][1]
                    cursor.updateRow(row)
                    i = i + 1
                else:
                    arcpy.AddMessage('Error in analyze notches indexing')
                    exit()
        del cursor
        del row
        return True


    def compare_two_normalized_notch_centroids(self, XL, YL, XR, YR, pctSize1, pctSize2, dx, dy, ds):
        '''Are these two NORMALIZED centroids symmetric pairs? '''

        if abs(YL - YR) < dy:
            # similar Y normalized Y centroids
            if abs((100. - XR) - XL) < dx:
                # similar symmetrical X normalized Y centroids
                if abs(pctSize1 - pctSize2) < ds:
                    # similar pct sizes
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False


    def count_notches(self, notches):       # unused
        '''Counts number of bases, right, left, LR, multi, solo and uncategorized reentrant types currently'''

        bases = 0
        right = 0
        left = 0
        LR = 0
        multi = 0
        solo = 0
        uncategorized = 0
        total = 0

        for notch in notches:

            total = total + 1

            if notch[1] == 'Base':
                bases = bases + 1

            if notch[1] == 'LR':
                LR = LR + 1

            if notch[1] == 'Right':
                right = right + 1

            if notch[1] == 'Left':
                left = left + 1

            if notch[1] == 'Multi':
                multi = multi + 1

            if notch[1] == 'Solo':
                solo = solo + 1

            if notch[1] == 'Undefined':
                uncategorized = uncategorized + 1

        del notch

        return ([bases, right, left, LR, multi, solo, uncategorized, total])

    def count_reentrant_type(self):        # unused
        '''Counts number of bases, right, left, LR, multi, solo and uncategorized reentrant types currently'''

        with arcpy.da.SearchCursor('Re_entrants', ['OID@', 'Type_', 'RE_Size',
                                                   'Pct_Centroid_X', 'Pct_Centroid_Y', 'Pct_RE_Size']) as cursor:
            bases = 0
            right = 0
            left = 0
            LR = 0
            multi = 0
            solo = 0
            uncategorized = 0

            for notch in cursor:

                if notch[1] == 'Base':
                    bases = bases + 1

                if notch[1] == 'LR':
                    LR = LR + 1

                if notch[1] == 'Right':
                    right = right + 1

                if notch[1] == 'Left':
                    left = left + 1

                if notch[1] == 'Multi':
                    multi = multi + 1

                if notch[1] == 'Solo':
                    solo = solo + 1

                if notch[1] == 'Undefined':
                    uncategorized = uncategorized + 1
        del cursor
        del notch

        return ([bases, right, left, LR, multi, solo, uncategorized])

    def normalize_notches(self, RE_size_max):
        '''Compute normalized RE_Size, Centroid_X and Centroid_Y in percent'''

        with arcpy.da.UpdateCursor('Re_entrants', ['RE_Size', 'Pct_RE_Size', 'CENTROID_X', 'CENTROID_Y',
                                                   'Pct_Centroid_X', 'Pct_Centroid_Y']) as cursor:
            for row in cursor:
                row[1] = 100. * self.normalize(0, RE_size_max, row[0])
                row[4] = 100. * self.normalize(self.outline_min_x, self.outline_max_x, row[2])
                row[5] = 100. * self.normalize(self.outline_min_y, self.outline_max_y, row[3])
                cursor.updateRow(row)
        del cursor
        del row
        return True


    def normalize(self, min_x, max_x, x):
        return (x - min_x) / (max_x - min_x)

    def draw_belts(self):
        '''Create line FC with horizontal ines
        1) Haft Collar is left to right haft point (orig from northernmost L and R reentrant vertex
        2) Haft Neck        (connect northernmost vertices of L and R reentrants)
        3) Outline Belt     (min X vertex to max X vertex at avg Y of these two vertices)
        4) Haft Belt        (connects X of innermost vertices of L and R reentrants at avg Y of these two vertices)
        n) Haft Guide       (connects L and R reentrant vertices nearest centroid (not yet)
        '''
        self.delete_preexisting('Collars')

        # Draw belt (max X extent drawn at average Y of two points)

        inFeatureClass = 'Outline_quad_pnts'
        with arcpy.da.SearchCursor(inFeatureClass, ['Ref', 'Shape@XY']) as cursor:
            for row in cursor:
                if row[0] == 'W': W_pnt = row[1]                # W_pnt is Shape@XY e.g. tuple (22,33)
                if row[0] == 'E': E_pnt = row[1]
        del cursor
        del row

        avg_W_pnt = [W_pnt[0], (W_pnt[1] + E_pnt[1]) / 2.]
        avg_E_pnt = [E_pnt[0], (W_pnt[1] + E_pnt[1]) / 2.]

        self.draw_line(avg_W_pnt, avg_E_pnt, 'Collars', 'Outline Belt')

        # Create Haft_Collar fc and draw haft collar line in Collars

        haft_collar_W_pnt = (-2, 0)                             # default haft collar length > 0 or not editable.
        haft_collar_E_pnt = (2, 0)
        haft_neck_W_pnt = (0, 0)
        haft_neck_E_pnt = (0, 0)

        inFeatureClass = "Re_entrants_quad_pnts"
        with arcpy.da.SearchCursor(inFeatureClass, ['Ref', 'Shape@XY', 'Type_']) as cursor:
            for row in cursor:
                # get top (+Y) vertices of L and R reentrants
                if row[0] == 'N' and row[2] == 'Left': haft_collar_W_pnt = row[1]
                if row[0] == 'N' and row[2] == 'Right': haft_collar_E_pnt = row[1]
                # if both are not valid, will get a strange (but editable) haft collar   (Only one Re_entrant case)

                # get inside vertices of L and R reentrants
                if row[0] == 'E' and row[2] == 'Left': haft_neck_W_pnt = row[1]
                if row[0] == 'W' and row[2] == 'Right': haft_neck_E_pnt = row[1]

        del cursor
        del row

        self.haft_collar_W_pnt = haft_collar_W_pnt
        self.haft_collar_E_pnt = haft_collar_E_pnt

        # If haft_collar fc exists in GDB_Edit, update its attributes and copy it here

        persistent_haft_collar = self.edit_gdb + '\\Haft_Collar'

        if arcpy.Exists(persistent_haft_collar):                    # if persistent copy exists, use it
            self.update_haft_collar(persistent_haft_collar)         # insure X,Y attributes in sync with graphics
            arcpy.CopyFeatures_management(persistent_haft_collar, 'Haft_Collar')
        else:
            # Create new Haft_Collar feature class. If undefined, endpoints will be (-2,0) (2,0)
            self.create_haft_collar(self.haft_collar_W_pnt[0], self.haft_collar_W_pnt[1],
                                    self.haft_collar_E_pnt[0], self.haft_collar_E_pnt[1])

        if (self.haft_collar_W_pnt != (-2, 0)) and (self.haft_collar_E_pnt != (2, 0)):
            self.draw_line(self.haft_collar_W_pnt, self.haft_collar_E_pnt, 'Collars', 'Haft Collar')

        if (haft_neck_W_pnt != (0, 0)) and (haft_neck_E_pnt != (0, 0)):
            self.draw_line(haft_neck_W_pnt, haft_neck_E_pnt, 'Collars', 'Haft Neck')

        if (haft_neck_W_pnt != (0, 0)) and (haft_neck_E_pnt != (0, 0)):
            avg_W_pnt = [haft_neck_W_pnt[0], (haft_neck_W_pnt[1] + haft_neck_E_pnt[1]) / 2.]
            avg_E_pnt = [haft_neck_E_pnt[0], (haft_neck_W_pnt[1] + haft_neck_E_pnt[1]) / 2.]

            self.draw_line(avg_W_pnt, avg_E_pnt, 'Collars', 'Haft Belt')
        return True


    def draw_line(self, W_pnt, E_pnt, line_fc, line_label):
        '''Draw a polyline from west point to east point'''

        if (not arcpy.Exists('in_memory\\tmp_Haft_Collar') and line_label == 'Haft Collar'):  # create tmp_haft collar for Blade_haft
            arcpy.CreateFeatureclass_management('in_memory', 'tmp_Haft_Collar', 'POLYLINE',
                                                has_z= 'ENABLED', spatial_reference= self.sr)
            with arcpy.da.InsertCursor('in_memory\\tmp_Haft_Collar', ["SHAPE@"]) as cursor:
                array = arcpy.Array([arcpy.Point(W_pnt[0], W_pnt[1]), arcpy.Point(E_pnt[0], E_pnt[1])])
                polyline = arcpy.Polyline(array)
                cursor.insertRow([polyline])
            del cursor

        if not arcpy.Exists(self.final_gdb + '\\' + line_fc):
            arcpy.CreateFeatureclass_management(self.final_gdb, line_fc, 'POLYLINE',  has_z= 'ENABLED', spatial_reference= self.sr)
            arcpy.AddField_management(self.final_gdb + '\\' + line_fc, 'Collar_Type', 'TEXT', '', '', 20)
            arcpy.AddField_management(self.final_gdb + '\\' + line_fc, 'Inclination', 'DOUBLE')

        with arcpy.da.InsertCursor(self.final_gdb + '\\' + line_fc, ["SHAPE@", 'Inclination', 'Collar_Type']) as cursor:
            array = arcpy.Array([arcpy.Point(W_pnt[0], W_pnt[1]), arcpy.Point(E_pnt[0], E_pnt[1])])
            polyline = arcpy.Polyline(array)
            inclination = math.degrees(math.atan2((E_pnt[1] - W_pnt[1]), (E_pnt[0] - W_pnt[0])))
            cursor.insertRow([polyline, inclination, line_label])
        del cursor
        return True


    def update_haft_collar(self, persistent_haft_collar):
        # Update X, Y attributes to insure they are in sync with graphics in case of manual edit.
        # persistent_haft_collar existence has been checked prior to calling

        #DEBUG think about way to shortcut this if haft collar is unaltered   (no time stamp accessible)

        fields = ['Shape@', 'X_left', 'Y_left', 'X_right', 'Y_right']

        with arcpy.da.UpdateCursor(persistent_haft_collar, fields) as cursor:      # update X,Y attributes
            for row in cursor:
                values = []                 # values will be [X_left, Y_left, X_right, Y_right]
                for part in row[0]:         # needs part in multi-part to run properly
                    for pnt in part:        # walk through vertices.  Expect only 2 points, get x and y of each.
                        values.append(pnt.X)
                        values.append(pnt.Y)

                if len(values) > 4:         # expect 4 values
                    arcpy.AddMessage('Haft Collar in ' + persistent_haft_collar + ' has more than two vertices or '
                                'multiple lines.\n Expecting one line with two vertices.  Please correct and re-run.')

                if values[0] > values[2]:     # if X left >  X_right  swap left and right points
                    xl = values[2]
                    yl = values[3]
                    values[2] = values[0]
                    values[3] = values[1]
                    values[0] = xl
                    values[1] = yl

                for i in range(3):              # fill 'X_left', 'Y_left', 'X_right', 'Y_right' attributes
                    row[i+1] = values[i]

                self.haft_collar_W_pnt = (values[0], values[1])     # (X_left, Y_left)
                self.haft_collar_E_pnt = (values[2], values[3])     # (X_right, Y_right)

                cursor.updateRow(row)
        return True

    def create_haft_collar(self, xl, yl, xr, yr):
        # Create haft collar feature class given endpoints

        fc = 'Haft_Collar'

        arcpy.CreateFeatureclass_management(self.final_gdb, fc, 'POLYLINE', has_z= 'ENABLED',
                                            spatial_reference= self.sr)

        arcpy.AddField_management(fc, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(fc, 'Type_', 'TEXT', '', '', 15)
        arcpy.AddField_management(fc, 'Author', 'TEXT', '', '', 25)
        arcpy.AddField_management(fc, 'X_left', 'DOUBLE')
        arcpy.AddField_management(fc, 'Y_left', 'DOUBLE')
        arcpy.AddField_management(fc, 'X_right', 'DOUBLE')
        arcpy.AddField_management(fc, 'Y_right', 'DOUBLE')

        with arcpy.da.InsertCursor(fc, ['SHAPE@', 'GKey', 'Type_', 'X_left', 'Y_left',
                                        'X_right', 'Y_right', 'Author']) as cursor:
            array = arcpy.Array([arcpy.Point(xl, yl), arcpy.Point(xr, yr)])
            polyline = arcpy.Polyline(array)
            cursor.insertRow([polyline, self.gkey, 'Haft Collar', xl, yl, xr, yr, 'GLiMR'])
        del cursor

        # Copy it to persistent GDB_Edit
        arcpy.CopyFeatures_management('Haft_Collar', self.edit_gdb + '\\Haft_Collar')
        return True


    def create_residual(self, inTopoExtFC, inOutlineFC):
        '''Create Topo, Smoothed and Residual rasters from extended topo raster and outline polygon'''

        # Process: Extract by Mask to clip extended topo to outline
        if self.verbose:
            arcpy.AddMessage('     Create Topo')

        TopoFile = ExtractByMask(inTopoExtFC, inOutlineFC)
        TopoFile.save('Topo')

        '''
        # save grid out as numpy array file for exterior processing
        npa = arcpy.RasterToNumPyArray('Topo')
        np_file = self.output_dir + '\\Topo_NP_' + self.gkey + '.npy'
        outfile = open(np_file, 'w')
        result = numpy.save(outfile, npa)
        outfile.close()
        
        '''
        arcpy.Curvature_3d('Topo', 'Curvature', 1.0, 'Curvature_Dip', 'Curvature_Strike')

        # For sensitivity check set radius to 10 and <= to 100 in next two lines.  Now just runs 7mm
        # 'radius' variable values are in 0.1 mm units
        if self.verbose:
            arcpy.AddMessage('     Create smoothed topo rasters using different smoothing')

        radius = 70                     # originally set up to try different radii
        while radius <= 75:
            mm = str(radius / 10) + 'mm'

            # Process: Focal Statistics      (moving average surfaces)
            # Depends on cell size = 0.05 mm, so if r=100, then d = 10mm  (2 x 100 cells x 0.05 mm)
            # arcpy.AddMessage('Focal Mean over ' + mm + ' circle')
            outFocal = FocalStatistics('TopoExt', 'Circle ' + str(radius) + ' CELL', 'MEAN', 'DATA')
            outFocal.save('FocalC' + mm)

            outExtractByMask = ExtractByMask('FocalC' + mm, 'Outline')
            outExtractByMask.save('Smoothed' + mm)
            
            arcpy.Delete_management('FocalC' + mm)

            arcpy.Contour_3d('Smoothed' + mm, 'Smoothed' + mm + '_Cont', 0.05)

            radius = radius + 10


        # Process: Minus  Topo - Smoothed 7mm to yield residual  (could do others)
        if self.verbose:
            arcpy.AddMessage('     Creating Residual by subtracting Topo - Smoothed7mm')

        arcpy.Minus_3d('Topo', 'Smoothed7mm', 'Residual7mm')
        
        arcpy.Contour_3d('Residual7mm', 'Residual7mm_Cont_05', 0.05)
        
        arcpy.Contour_3d('Topo', 'Topo_Cont_05', 0.05)

        # This completes building basic geometry pieces...
        return True


    def compute_flake_basins(self, inResidualFC, outFlakeBasinFC):

        # Residual_Flow_Direction
        # Flow_Drops
        # Sinks
        # Filled_Residual
        # Flow_Dir_Filled_Residual
        # Flow_Accum_Filled_Residual
        # Filled_Basin_Raster

        if self.verbose:
            arcpy.AddMessage('     Compute flake scar using auto fill')

        arcpy.env.workspace = self.final_gdb

        # Process: Fill
        # arcpy.AddMessage('Flow Fill')
        outFill = Fill(inResidualFC)
        outFill.save('tmp_Filled_Residual_' + inResidualFC)
        

        # Process: Flow Direction
        if self.verbose:
            arcpy.AddMessage('     Flow direction')

        outFlowDir = FlowDirection('tmp_Filled_Residual_' + inResidualFC)
        outFlowDir.save('tmp_Flow_Dir_Filled_Residual_' + inResidualFC)
        

        # Process: Basin
        if self.verbose:
            arcpy.AddMessage('     Compute flake scar basin raster')

        outBasin = Basin('tmp_Flow_Dir_Filled_Residual_' + inResidualFC)
        outBasin.save(inResidualFC + '_Filled_Basin_Raster')
        

        # Boundary Clean to remove skinny slivers
        outBasin = BoundaryClean(inResidualFC + '_Filled_Basin_Raster', 'DESCEND', 'TWO_WAY')
        outBasin.save(inResidualFC + '_Filled_Basin_Raster')
        


        # Process: Raster to Polygon
        if self.verbose:
            arcpy.AddMessage('     Basin Raster to Polygon')

        arcpy.RasterToPolygon_conversion(inResidualFC + '_Filled_Basin_Raster', outFlakeBasinFC, 'SIMPLIFY', 'Value')
        

        inFeatureClass = outFlakeBasinFC
        arcpy.AddField_management(inFeatureClass, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(inFeatureClass, 'Version', 'TEXT', '', '', 10)

        with arcpy.da.UpdateCursor(inFeatureClass, ['GKey', 'Version']) as cursor:
            for row in cursor:
                row[0] = self.gkey
                row[1] = self.version
                cursor.updateRow(row)
        del cursor
        del row

        self.delete_preexisting('tmp_Filled_Residual_' + inResidualFC)
        self.delete_preexisting('tmp_Flow_Dir_Filled_Residual_' + inResidualFC)
        self.delete_preexisting(inResidualFC + '_Filled_Basin_Raster')
        self.delete_preexisting('Fill_Resid*')
        return True

    def compute_flake_basins_multi(self, inResidualFC, outFlakeBasinFC_root):

        # Residual_Flow_Direction
        # Flow_Drops
        # Sinks
        # Filled_Residual
        # Flow_Dir_Filled_Residual
        # Flow_Accum_Filled_Residual
        # Filled_Basin_Raster

        if self.verbose:
            arcpy.AddMessage('     Compute flake scar sequence with varying fill')

        arcpy.env.workspace = self.final_gdb

        for zlimit in [0.025, 0.05, 0.1]:
            try:
                zl_label = '_' + str(zlimit).replace('0.', '')
                outFlakeBasinFC = outFlakeBasinFC_root + zl_label

                # Process: Fill
                if self.verbose:
                    arcpy.AddMessage('     Flow Fill for Fill zlimit = ' + str(zlimit))

                outFill = Fill(inResidualFC, zlimit)
                outFill.save('tmp_Filled_Residual_' + inResidualFC)
                

                # Process: Flow Direction
                if self.verbose:
                    arcpy.AddMessage('     Flow direction')

                outFlowDir = FlowDirection('tmp_Filled_Residual_' + inResidualFC)
                outFlowDir.save('tmp_Flow_Dir_Filled_Residual_' + inResidualFC)
                

                # Process: Basin
                if self.verbose:
                    arcpy.AddMessage('     Compute flake scar basin raster')

                outBasin = Basin('tmp_Flow_Dir_Filled_Residual_' + inResidualFC)
                outBasin.save(inResidualFC + '_Filled_Basin_Raster')
                

                # Boundary Clean to remove skinny slivers
                outBasin = BoundaryClean(inResidualFC + '_Filled_Basin_Raster', 'DESCEND', 'TWO_WAY')
                outBasin.save(inResidualFC + '_Filled_Basin_Raster')
                

                # Process: Raster to Polygon
                if self.verbose:
                    arcpy.AddMessage('     Basin Raster to Polygon')

                arcpy.RasterToPolygon_conversion(inResidualFC + '_Filled_Basin_Raster', outFlakeBasinFC, 'SIMPLIFY',
                                                 'Value')
                

                inFeatureClass = outFlakeBasinFC
                arcpy.AddField_management(inFeatureClass, 'GKey', 'TEXT', '', '', 25)
                arcpy.AddField_management(inFeatureClass, 'Version', 'TEXT', '', '', 10)
                arcpy.AddField_management(inFeatureClass, 'ScarSource', 'TEXT', '', '', 50)

                with arcpy.da.UpdateCursor(inFeatureClass, ['GKey', 'Version', 'ScarSource']) as cursor:
                    for row in cursor:
                        row[0] = self.gkey
                        row[1] = self.version
                        row[2] = 'Auto' + zl_label
                        cursor.updateRow(row)
                del cursor
                del row

            except Exception as e:
                if self.verbose:
                    arcpy.AddMessage('     Exception in compute_flake_basins_multi for zlimit = ' + str(zlimit) + '.  Continuing')
                continue

        self.delete_preexisting('tmp_Filled_Residual_' + inResidualFC)
        self.delete_preexisting('tmp_Flow_Dir_Filled_Residual_' + inResidualFC)
        self.delete_preexisting(inResidualFC + '_Filled_Basin_Raster')
        return True


    def create_surface(self, inPointFC, outRaster):
        try:
            arcpy.SplineWithBarriers_3d(inPointFC, 'Shape', '', '0.05', outRaster, '0')
            
        except:
            arcpy.AddMessage('\nSpline with barrier problem: 64 bit version on Python is calling 32 bit version of Java'
                '\n------> Spline with barrier problem: 64 bit version on Python may be calling 32 bit version of Java'
                '\n------> Run 32 bit version of Python interpreter to correct the problem\n')
            exit(1)
        return True


    def angle2bearing(self, angle):
        ''' Convert angle (+X = 0 and increasing ccw) to bearing (+Y = 0 and increasing cw)  '''
        return (450. - angle) % 360.

    def orientation_angle2bearing(self, angle):
        ''' Convert angle (+X = 0 and increasing ccw) to bearing (+Y = 0 and increasing cw)
            for case where input angle is orientation ( input range between 0 - 180) '''
        bearing = (450. - angle) % 360.
        if bearing >= 180.:
            bearing = bearing - 180
        return bearing

    def rebuild_GLiMR_member_variables(self):
        # set member variables from pre-existing feature classes
        #
        # init handles: xyz_dir, gdb_dir, mxd_dir, image_dir, template_dir, gdb_edit_dir, output_dir, verbose
        #               sr, version
        #
        # local context: final_gdb, edit_gdb, pt_gdb, pt_name, summary_gdb
        # local context: xyz_file
        #
        # not handled currently:
        # centroid_x, centroid_y (Outline_quad_pnts)
        #  x_left, x_right
        # zRange

        previous_workspace = arcpy.env.workspace
        arcpy.env.workspace = self.final_gdb

        if not (arcpy.Exists('Outline') and arcpy.Exists('Hull') ):
            arcpy.AddMessage('Failure to restore GLiMR environment')  # and arcpy.Exists(self.edit_gdb)
            arcpy.AddMessage('Component feature classes not found for ' +  self.gkey + ' exiting...')
            arcpy.AddMessage('You must build the basic geometry by processing the point before subsequent processing')
            arcpy.GetMessages()
            return False

        # Retrieve Outline properties
        with arcpy.da.SearchCursor('Outline', ['Height', 'Width', 'EXT_MIN_X', 'EXT_MAX_X', 'EXT_MIN_Y', 'EXT_MAX_Y',
                                               'HW_ratio', 'Half_Thickness', 'GKey'])as cursor:
            for row in cursor:
                # Only one row...
                overallWidth = row[3] - row[2]
                overallHeight = row[5] - row[4]

                self.outline_height = overallHeight
                self.outline_width = overallWidth
                self.outline_min_x = row[2]
                self.outline_max_x = row[3]
                self.outline_min_y = row[4]
                self.outline_max_y = row[5]
        del cursor
        del row

        # Retrieve Convex Hull properties
        with arcpy.da.SearchCursor('Hull', ['Shape@Area', 'Shape@Length']) as cursor:
            for row in cursor:
                # Only one row...
                self.hull_area = row[0]
                self.hull_length = row[1]
            del cursor
            del row

        # Restore Haft and Reentrant feature classes and set haft point variables
        self.fetch_persistent_data()
        arcpy.env.workspace = previous_workspace

        return True


    def fetch_persistent_data(self):
        ''' Copies persistent version of Re_entrants and Haft_Collar feature classes from GDB_Edits ot current GDB'''
        persistent_haft_collar = self.edit_gdb + '\\Haft_Collar'

        if arcpy.Exists(persistent_haft_collar):                    # if persistent copy exists, use it
            self.change_fieldname(persistent_haft_collar, 'Type', 'Type_')
            self.update_haft_collar(persistent_haft_collar)         # insure X,Y attributes in sync with graphics
            arcpy.CopyFeatures_management(persistent_haft_collar, 'Haft_Collar')

        persistent_reentrant = self.edit_gdb + '\\Re_entrants'

        if arcpy.Exists(persistent_reentrant):                    # if persistent copy exists, use it
            self.change_fieldname(persistent_reentrant, 'Type', 'Type_')
            arcpy.CopyFeatures_management(persistent_reentrant, 'Re_entrants')
        return True

    def change_fieldname(self, fc, oldname, newname):
        if self.is_field_present(fc, oldname):
            arcpy.AlterField_management(fc, oldname, newname, newname)

    def is_geometry_complete(self):
        '''  Check if basic raster and feature classes are already built

             Checks for pre-existing raster files and feature classes
             If using pre-existing,
               Checks for updated Re_entrant and Haft_Collar feature classes
               Reconstructs required GLiMR data attributes from feature classes
        '''
        arcpy.env.workspace = self.final_gdb

        if not arcpy.Exists(self.final_gdb):
            return False

        existing_rasters_set = {r for r in arcpy.ListDatasets()}
        required_rasters_set = {u'Topo', u'Smoothed7mm', u'Residual7mm'}

        existing_fc_set = {f for f in arcpy.ListFeatureClasses()}
        required_fc_set = {u'BasePts', u'Outline', u'Outline_centroid', u'TIN_edge', u'TIN_Tri', u'Hull',
                           u'Re_entrants', u'Outline_quad_pnts', u'Re_entrants_quad_pnts', u'RE_Ellipse',
                           u'RE_Ellipse_pnts', u'Collars',u'Haft_Collar', u'Smoothed7mm_Cont',
                           u'Residual7mm_Cont_05', u'Topo_Cont_05'}

        required_rasters_OK = required_rasters_set.issubset(existing_rasters_set)
        required_fc_OK = required_fc_set.issubset(existing_fc_set)

        if required_rasters_OK and required_fc_OK:
            arcpy.AddMessage('   Using pre-existing feature classes and topo-related rasters')
            #  self.fetch_persistent_data()  # done by rebuild_GLiMR_member_variables()
            self.rebuild_GLiMR_member_variables()
            return True

        else:
            arcpy.AddMessage('    * Some feature classes and topo-related rasters were missing. Recomputing them...')
            return False

    def is_autoflake_complete(self):
        arcpy.env.workspace = self.final_gdb

        if len(arcpy.ListFeatureClasses('zScarAuto*')) > 0:
            return True
        else:
            return False

    def are_flake_attributes_complete(self, workspace):     # not used
        if len(arcpy.ListFeatureClasses('Flakes*')) > 0:
            return True
        return False


    def is_field_present(self, infc, field_name):
        fields = arcpy.ListFields(infc)
        for field in fields:
            if field.name == field_name:
                return True
        return False


    def run_quadrants_all_GDB(self):  # not utilized...  path issues
        '''Compute quadrant points for outline and reentrants
            Process Outline and Re_entrant feature classes in each GDB directory'''

        arcpy.env.workspace = self.gdb_dir

        workspaces = arcpy.ListWorkspaces('*', 'FileGDB')

        if len(workspaces) == 0:
            # nothing to do
            arcpy.AddMessage('Nothing to do. No GDB directories found in ' + arcpy.env.workspace)
            return ()

        if self.verbose:
            arcpy.AddMessage('     ' + str(len(workspaces)) + ' geodatabases to process for quadrant points')

        # Process file GDBs in GDB_Dir

        for workspace in workspaces:
            if (workspace != self.gdb_dir + '\\' + 'a_GLiMR_Summary.gdb'):
                self.final_gdb = workspace
                if self.verbose:
                    arcpy.AddMessage('     Current workspace: ' + workspace + '\n')

                if not arcpy.Exists(self.final_gdb + '\\' + 'Outline'):
                    arcpy.AddMessage('DEBUG: self.final_gdb = ' + self.final_gdb)
                    arcpy.AddMessage('Outline does not exist...   Exiting from Run_quadrants_all_GDB...')
                    exit()

                self.create_quadrant_points('Outline')  # reprocessing all Outlines in all geodatabases
                self.create_reentrant_quadrant_points('Re_entrants')  # reprocessing all Re_entrants in all geodatabases
        return True

    def create_reentrant_quadrant_points(self, infc):
        '''Create polygon attributes and point feature class representing vertices at extreme
        +Y, -Y, +X, -X points along polygon perimeters.'''

        # infc is input feature class without path

        arcpy.env.workspace = self.final_gdb

        if self.verbose:
            arcpy.AddMessage('     Processing ' + infc)

        if not arcpy.Exists(infc):
            arcpy.AddMessage('ERROR: create_quadrant_points was asked to process a '
                             'feature class that does not exist: ' + self.final_gdb + '\\' + infc)

        outFC = infc + '_quad_pnts'

        if arcpy.Exists(outFC):
            arcpy.Delete_management(outFC)

        arcpy.CreateFeatureclass_management(self.final_gdb, outFC, 'POINT',  has_z= 'ENABLED', spatial_reference= self.sr)
        nameFID = infc + '_FID'

        if not self.is_field_present(outFC, nameFID):
            # arcpy.AddField_management(outFC, nameFID, 'LONG')
            arcpy.AddField_management(outFC, 'GKey', 'TEXT', '', '', 25)
            arcpy.AddField_management(outFC, 'Type_', 'TEXT', '', '', 15)    # reentrant label Left, Right, Base
            arcpy.AddField_management(outFC, 'Ref', 'TEXT', '', '', 10)     # point reference name T, B, R, L
            arcpy.AddField_management(outFC, 'Label', 'TEXT', '', '', 50)

        if not self.is_field_present(infc, 'MinX_X'):
            arcpy.AddField_management(infc, 'MinX_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MinX_Y', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxX_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxX_Y', 'DOUBLE')
            arcpy.AddField_management(infc, 'MinY_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MinY_Y', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxY_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxY_Y', 'DOUBLE')

        edit = arcpy.da.Editor(self.final_gdb)
        edit.startEditing(False, False)

        fields = ['OID@', 'SHAPE@', 'MinX_X', 'MinX_Y', 'MaxX_X', 'MaxX_Y',
                  'MinY_X', 'MinY_Y', 'MaxY_X', 'MaxY_Y', 'Type_']

        with arcpy.da.UpdateCursor(infc, fields) as cursor:
            for row in cursor:
                re_type = row[10]

                # Step through each part of the feature
                #
                for part in row[1]:
                    re_MinX = [100000, 100000]
                    re_MaxX = [-100000, 100000]
                    re_MinY = [100000, 100000]
                    re_MaxY = [100000, -100000]

                    for pnt in part:
                        if pnt:
                            if pnt.X < re_MinX[0]: re_MinX = [pnt.X, pnt.Y]
                            if pnt.X > re_MaxX[0]: re_MaxX = [pnt.X, pnt.Y]
                            if pnt.Y < re_MinY[1]: re_MinY = [pnt.X, pnt.Y]
                            if pnt.Y > re_MaxY[1]: re_MaxY = [pnt.X, pnt.Y]
                        else:
                            # If pnt is None, this represents an interior ring
                            #
                            arcpy.AddMessage('This reentrant has an unexpected interior ring:')

                    row[2] = re_MinX[0]
                    row[3] = re_MinX[1]
                    row[4] = re_MaxX[0]
                    row[5] = re_MaxX[1]
                    row[6] = re_MinY[0]
                    row[7] = re_MinY[1]
                    row[8] = re_MaxY[0]
                    row[9] = re_MaxY[1]

                    cursor.updateRow(row)

                    '''
                    # compute mid-points between re_MinX and re_MaxX
                    midXofX = (re_MinX[0] + re_MaxX[0]) / 2.
                    midYofX = (re_MinX[1] + re_MaxX[1]) / 2.
                    re_MidX = [midXofX, midYofX]

                    # compute mid-point between re_MinY and re_MaxY
                    midXofY = (re_MinY[0] + re_MaxY[0]) / 2.
                    midYofY = (re_MinY[1] + re_MaxY[1]) / 2.
                    re_MidY = [midXofY, midYofY]
                    '''

                # shifted left one click: only operate on first part of multi-part Reentrant feature  (G324_1 has 2)
                # DEBUG  G324_1 may have been an edit case (but only haft_collar     Double check...
                with arcpy.da.InsertCursor(outFC, ['SHAPE@XY', 'Ref', 'Type_', 'GKey']) as in_cursor:
                    in_cursor.insertRow([re_MinX, 'W', re_type, self.gkey])
                    in_cursor.insertRow([re_MaxX, 'E', re_type, self.gkey])
                    in_cursor.insertRow([re_MinY, 'S', re_type, self.gkey])
                    in_cursor.insertRow([re_MaxY, 'N', re_type, self.gkey])

        edit.stopEditing(True)
        del cursor
        del in_cursor
        del row

        arcpy.AddXY_management(outFC)
        return True


    def create_quadrant_points(self, infc):
        '''Create polygon attributes and point feature class representing vertices at extreme
        +Y, -Y, +X, -X points along polygon perimeters.'''

        # infc is input feature class without path

        arcpy.env.workspace = self.final_gdb

        if self.verbose:
            arcpy.AddMessage('     Processing ' + infc)

        if not arcpy.Exists(infc):
            arcpy.AddMessage('ERROR: create_quadrant_points was asked to process a feature class '
                             'that does not exist: ' + self.final_gdb + '\\' + infc)

        outFC = infc + '_quad_pnts'

        if arcpy.Exists(outFC):
            arcpy.Delete_management(outFC)

        arcpy.CreateFeatureclass_management(self.final_gdb, outFC, 'POINT', has_z= 'ENABLED', spatial_reference= self.sr)
        nameFID = infc + '_FID'

        if not self.is_field_present(outFC, nameFID):
            # arcpy.AddField_management(outFC, nameFID, 'LONG')     # source fid
            arcpy.AddField_management(outFC, 'GKey', 'TEXT', '', '', 25)
            arcpy.AddField_management(outFC, 'Type_', 'TEXT', '', '', 15)    # reentrant label Left, Right, Base
            arcpy.AddField_management(outFC, 'Ref', 'TEXT', '', '', 10)     # point reference name T, B, R, L
            arcpy.AddField_management(outFC, 'Label', 'TEXT', '', '', 50)

        if not self.is_field_present(infc, 'MinX_X'):
            arcpy.AddField_management(infc, 'MinX_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MinX_Y', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxX_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxX_Y', 'DOUBLE')
            arcpy.AddField_management(infc, 'MinY_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MinY_Y', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxY_X', 'DOUBLE')
            arcpy.AddField_management(infc, 'MaxY_Y', 'DOUBLE')

        edit = arcpy.da.Editor(self.final_gdb)
        edit.startEditing(False, False)

        fields = ['OID@', 'SHAPE@', 'MinX_X', 'MinX_Y', 'MaxX_X', 'MaxX_Y',
                  'MinY_X', 'MinY_Y', 'MaxY_X', 'MaxY_Y']

        with arcpy.da.UpdateCursor(infc, fields) as cursor:
            for row in cursor:
                # print('Adding 4 quad points for feature {0}:'.format(row[0]))

                # Step through each part of the feature
                #
                for part in row[1]:
                    re_MinX = [100000, 100000]
                    re_MaxX = [-100000, 100000]
                    re_MinY = [100000, 100000]
                    re_MaxY = [100000, -100000]

                    for pnt in part:
                        if pnt:
                            if pnt.X < re_MinX[0]: re_MinX = [pnt.X, pnt.Y]
                            if pnt.X > re_MaxX[0]: re_MaxX = [pnt.X, pnt.Y]
                            if pnt.Y < re_MinY[1]: re_MinY = [pnt.X, pnt.Y]
                            if pnt.Y > re_MaxY[1]: re_MaxY = [pnt.X, pnt.Y]
                        else:
                            # If pnt is None, this represents an interior ring
                            #
                            arcpy.AddMessage('This reentrant has an unexpected interior ring:')

                    row[2] = re_MinX[0]
                    row[3] = re_MinX[1]
                    row[4] = re_MaxX[0]
                    row[5] = re_MaxX[1]
                    row[6] = re_MinY[0]
                    row[7] = re_MinY[1]
                    row[8] = re_MaxY[0]
                    row[9] = re_MaxY[1]

                    cursor.updateRow(row)

                    # compute mid-points between re_MinX and re_MaxX
                    #midXofX = (re_MinX[0] + re_MaxX[0]) / 2.
                    #midYofX = (re_MinX[1] + re_MaxX[1]) / 2.
                    #re_MidX = [midXofX, midYofX]

                    # compute mid-point between re_MinY and re_MaxY
                    #midXofY = (re_MinY[0] + re_MaxY[0]) / 2.
                    #midYofY = (re_MinY[1] + re_MaxY[1]) / 2.
                    #re_MidY = [midXofY, midYofY]
                    re_Centroid = [self.centroid_x, self.centroid_y]

                    with arcpy.da.InsertCursor(outFC, ['SHAPE@XY', 'Ref', 'Type_', 'GKey']) as in_cursor:
                        in_cursor.insertRow([re_MinX, 'W','Outline', self.gkey])
                        in_cursor.insertRow([re_MaxX, 'E','Outline', self.gkey])
                        in_cursor.insertRow([re_MinY, 'S','Outline', self.gkey])
                        in_cursor.insertRow([re_MaxY, 'N','Outline', self.gkey])

                        #in_cursor.insertRow([re_MidX, 'X_Mid', 'Outline', self.gkey])
                        #in_cursor.insertRow([re_MidY, 'Y_Mid', 'Outline', self.gkey])
                        in_cursor.insertRow([re_Centroid, 'Centroid', 'Outline', self.gkey])

        edit.stopEditing(True)
        del cursor
        del in_cursor
        del row

        arcpy.AddXY_management(outFC)
        return True


    def create_mxd_and_images(self):
        ''' Creates and links an MXD file based on Template_<name>.MXD in Templates directory and creates jpg image

        The template MXD file has to get its data source from the Template.GDB for this routine to work properly.
        For every 'Template_<name>.mxd', creates a jpg in Images\ScreenGrabs
        Only preserves one MXD (Outlines) because they are all copies with just different layers turned on and off.
        '''
        arcpy.env.workspace = self.template_dir
        templates = arcpy.ListFiles('Template_*.mxd')

        if len(templates) == 0:
            arcpy.AddMessage('Nothing to do. No MXD template files found in ' + self.template_dir)
            exit()

        image_dir = self.image_dir + '\\ScreenGrabs'
        if not arcpy.Exists(image_dir):
            os.mkdir(image_dir)

        oldWorkspacePath = self.template_dir + '\\Template.gdb'  # path used in template mxd (search target)
        newWorkSpacePath = self.gdb_dir + "\\G" + self.gkey + ".gdb"
        NewMXDRoot = self.mxd_dir + "\\G" + self.gkey

        arcpy.env.workspace = newWorkSpacePath

        for template in templates:
            suffix = template[9:-4]  # strips 'Template_' and '.mxd' from template mxd file name
            NewMXDFile = NewMXDRoot + '_' + suffix + '.mxd'

            arcpy.AddMessage('     Creating image for ' + suffix)

            mxd = arcpy.mapping.MapDocument(self.template_dir + '\\' + template)
            mxd.findAndReplaceWorkspacePaths(oldWorkspacePath, newWorkSpacePath)

            mxd.saveACopy(NewMXDFile)

            # Reclassify some layers for newly linked data
            #
            mxd = arcpy.mapping.MapDocument(NewMXDFile)

            for lyr in arcpy.mapping.ListLayers(mxd, 'Flake scar outlines colored'):
                if arcpy.mapping.Layer.supports(lyr, 'SYMBOLOGYTYPE'):
                    if lyr.symbologyType == "GRADUATED_COLORS":  # Check range and reclassify
                        lyr.symbology.reclassify()

                arcpy.env.workspace = newWorkSpacePath  #DEBUG

            mxd.save()

            # Make jpg images of some Layers

            df = arcpy.mapping.ListDataFrames(mxd, "Layers")[0]

            lyr = arcpy.mapping.ListLayers(mxd, "Outline", df)[0]
            arcpy.SelectLayerByAttribute_management(lyr, "NEW_SELECTION")
            df.extent = lyr.getSelectedExtent()
            # df.zoomToSelectedFeatures()
            arcpy.SelectLayerByAttribute_management(lyr, "CLEAR_SELECTION")
            arcpy.RefreshActiveView()
            mxd.save()

            image_name = "G" + self.gkey + '^' + suffix

            if arcpy.Exists(image_dir + "\\" + image_name + '.jpg'):
                arcpy.Delete_management(image_dir + "\\" + image_name + '.jpg')

            arcpy.mapping.ExportToJPEG(mxd, image_dir + "\\" + image_name, df,
                                       df_export_width=1200,
                                       df_export_height=1000,
                                       world_file=False)

            if self.verbose:
                arcpy.AddMessage('     For: G' + self.gkey + ', created MXD ' + NewMXDFile +
                                 ' and saved image in ScreenGrabs: '  + image_name + '.jpg')

            '''
            #  df= "PAGE_LAYOUT"
            #Page Layout export
            arcpy.mapping.ExportToJPEG(mxd,  image_dir + "\\" + image_name, df,
                                       df_export_width=11,
                                       df_export_height=8.5)
            '''
            del mxd

            if suffix != 'Outlines':
                try:
                    if arcpy.Exists(NewMXDFile):
                        arcpy.Delete_management(NewMXDFile)
                except:
                    arcpy.AddMessage('   Exception: Could not delete ' + NewMXDFile + '   Continuing')
                    continue

        return True


    def create_all_mxd_and_images(self):
        # Only used in Custom Batch to iterate on entire directory

        arcpy.env.workspace = self.gdb_dir

        gdb_list = arcpy.ListFiles("*.gdb")
        if len(gdb_list) == 0:
            arcpy.AddMessage('No Geodatabases found in ' + self.gdb_dir + '\nCheck your GLiMR environment paths')

        for gdb_file in arcpy.ListFiles("*.gdb"):  # replace with global gdb list
            if gdb_file != 'a_GLiMR_Summary.gdb':
                ptName = str(gdb_file.split(".")[0])
                self.parse_pt_name(ptName)

                arcpy.AddMessage('Processing ' + ptName + ', gkey = ' + self.gkey)
                self.create_mxd_and_images()
        return True

    def print_environments(self):
        environments = arcpy.ListEnvironments()
        for environment in environments:
            envSetting = eval("arcpy.env." + environment)
            print "%-30s: %s" % (environment, envSetting)
        exit()

    def build_domain_poly(self, inTIN_near_fc, domain, angle_from, angle_to, dist):  # not used (yet)
        # Create domain polygons with centroids for domain(NE, NW and south) 'bearing to edge' directions
        # Selection is near_angle between from angle_from and angle_to and > dist from edge of point

        '''
        self.build_domain_poly('TIN_Tri_Near', 'NE', 0, 120, 0.5)
        self.build_domain_poly('TIN_Tri_Near', 'NW', 0, 120, 0.5)
        self.build_domain_poly('TIN_Tri_Near', 'South', 0, 120, 0.5)
        '''

        in_lyr = 'inTIN_Near_lyr'
        arcpy.MakeFeatureLayer_management(inTIN_near_fc, in_lyr)
        domain_fc = 'SubDomain_' + domain

        whereclause = "NEAR_ANGLE >= " + str(angle_from) + " AND NEAR_ANGLE <= " + str(
            angle_to) + " AND NEAR_DIST > " + str(dist)

        arcpy.SelectLayerByAttribute_management(in_lyr, 'NEW_SELECTION', whereclause)
        arcpy.AggregatePolygons_cartography(in_lyr, domain_fc, 0.1)

        arcpy.AddGeometryAttributes_management(domain_fc, 'CENTROID')
        arcpy.DeleteField_management(domain_fc, 'CENTROID_Z')
        arcpy.AddField_management(domain_fc, 'GKey', 'TEXT', '', '', 25)
        arcpy.AddField_management(domain_fc, 'Domain', 'TEXT', '', '', 10)
        arcpy.AddField_management(domain_fc, 'Name', 'TEXT', '', '', 40)

        with arcpy.da.UpdateCursor(domain_fc, ['GKey', 'Domain', 'Name', 'CENTROID_Y', 'SHAPE@AREA']) as cursor:
            for row in cursor:
                row[0] = self.gkey
                row[1] = domain
                row[2] = 'Undefined'
                if row[4] < 4:  # area < 4 (tiny)
                    row[2] = 'Tiny'
                elif row[3] > 0:  # area > 4 and centroid_y > 0
                    row[2] = 'Upper'
                elif row[3] <= 0:  # area > 4 and centroid_y <= 0
                    row[2] = 'Lower'
                cursor.updateRow(row)
        del cursor
        del row
        arcpy.SelectLayerByAttribute_management(in_lyr, "CLEAR_SELECTION")
        return True

    def raster_to_ASCII_Pnts(self, image_file, pt_file, actual_height=100.):
        ''' Convert a BW (2-bit) TIFF image file to an ASCII xyz file with black pixels as points
        Sequentially
        - Load raster into numpy via arcpy
        - filter for black pixels
        - shift to XY mean and scale XY to match actual height to actual pixel image height
        - Export to ASCII with Z = 1.
        '''

        arcpy.env.workspace = self.final_gdb

        # Convert Raster to numpy array
        arr = arcpy.RasterToNumPyArray(arcpy.Raster(image_file))

        dots = arr[arr < 1]                         # black dots in BW tiff
        xyz = np.zeros((len(dots),2))               # create appropriately-sized xyz numpy array

        ijz = []
        count = 0
        min_y = 100000.
        max_y = -1.
        min_x = 100000.
        max_x = -1.

        for i in range(0, arr.shape[0]):
            for j in range(0, arr.shape[1]):
                if arr[i,j] < 1:                    # fill x and y with i, j indices from array for black points
                    xyz[count,0] = j
                    xyz[count,1] = (arr.shape[0] - i)

                    if xyz[count,0] < min_x:        # find min/max X and Y indices of actual black image
                        min_x = xyz[count, 0]
                    if xyz[count,0] > max_x:
                        max_x = xyz[count, 0]

                    if xyz[count,1] < min_y:
                        min_y = xyz[count,1]
                    if xyz[count,1] > max_y:
                        max_y = xyz[count,1]

                    count +=1

        mean_y = (max_y - min_y) / 2.                   # position of data within image
        mean_x = (max_x - min_x) / 2.

        cell_size = actual_height / (max_y - min_y)     # scale to actual height using numpy magic
        xyz *= cell_size

        shift = [mean_x + min_x, mean_y + min_y]        # shift to mean center of range (this is not outline centroid)
        #xyz -= shift       # DEBUG Stand alone raster2pts and GLiMR yield different shift values

        if arcpy.Exists(pt_file):                       # delete pre-existing point file
            arcpy.Delete_management(pt_file)

        outfile = open(pt_file, 'w')                    # write ASCII file of (X, Y, 1.)
        for i in range(0, xyz.shape[0]):
            print  >> outfile, xyz[i][0], xyz[i][1], 1.

        outfile.close()
        del arr
        del xyz

        arcpy.AddMessage('     ASCII xyz file written: ' + pt_file)

        return True
